/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.geotools;

import java.awt.geom.Rectangle2D;
import java.text.DecimalFormat;
import java.util.Set;
import java.util.SortedMap;
import java.util.TreeMap;

import org.apache.log4j.Logger;
import org.geotools.geometry.Envelope2D;
import org.geotools.geometry.jts.JTS;
import org.geotools.referencing.CRS;
import org.opengis.geometry.Envelope;
import org.opengis.referencing.crs.CoordinateReferenceSystem;

/**
 * Diese Klasse enthaelt allgemeine Funktionen fuer die Arbeit mit Geotools.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class GTUtil {
  private static Logger LOGGER = Logger.getLogger( GTUtil.class.getName() );

  /** Konstante fuer das CRS "WGS84" (erzeugt als "EPSG:4326")*/
  public static CoordinateReferenceSystem WGS84 = null;

  // Initialisierung der CRS-Konstanten
  static {
    try {
      WGS84 = CRS.decode("EPSG:4326",true); // lat/lon (WGS84)
    } catch (Exception err) {
    	LOGGER.warn("Exception while creating default WGS84 from code 'EPSG:4326'");
    }
  }

  /**
   * Erzeugt ein {@link CoordinateReferenceSystem} aus einer String-Definition.
   * Akzeptiert wird ein EPSG-Code "EPSG:..." oder eine WKT-Definition des CRS
   * @param crsDef Definition fuer das CRS.
   * @return {@code null}, falls der String nicht zu einem CRS dekodiert werden
   *         kann
   */
  public static CoordinateReferenceSystem createCRS(String crsDef) {
    if ( crsDef == null || crsDef.trim().equals("") )
      return null;

    CoordinateReferenceSystem crs = null;
    try {
        // Akzeptiert wird: EPSG-Code
        if (crsDef.startsWith("EPSG:"))
          crs = createCRS_EPSG( crsDef );
        else 
          crs = CRS.parseWKT( crsDef );
    } catch (Exception err) {
      LOGGER.error("Error while creating CRS",err);
    }
    return crs;
  }

  /**
   * Erzeugt ein {@link CoordinateReferenceSystem} aus einem (EPSG-)Code.
   * Entspricht {@link CRS#decode(String,boolean) CRS#decode(String,true)}.
   * Exceptions werden jedoch abgefangen und stattdessen {@code null} zurueckgegeben.
   * @param code Code fuer das CRS.
   */
  public static CoordinateReferenceSystem createCRS_EPSG(String code) {
    try {
      return CRS.decode(code,true);
    } catch (Exception err) {
      LOGGER.error("Error while creating CRS",err);
      return null;
    }
  }

  /**
   * Erzeugt ein UTM-CoordinateReferenceSystem.
   * @param zone UTM-Zone
   */
  public static CoordinateReferenceSystem createCRS_UTM(int zone) {
    DecimalFormat format_99 = new DecimalFormat("00");
    return createCRS( "EPSG:326"+format_99.format(zone) );
  }

  /**
   * Erzeugt einen {@link Envelope2D} aus einem {@link Rectangle2D} .
   * @param env Georeferenz und Ausdehnung
   * @param crs CoordinateReferenceSystem
   */
  public static Envelope2D createEnvelope2D(Rectangle2D env, CoordinateReferenceSystem crs) {
    return new Envelope2D(crs, env.getX(), env.getY(), env.getWidth(),
                          env.getHeight());
  }

  /**
   * Liefert alle zur Verfuegung stehenden {@linkplain CoordinateReferenceSystem CRS} fuer eine
   * Authority.
   * @param authority      Authority fuer die die CRS geliefert werden (z.B. {@code "EPSG"})
   * @param longitudeFirst {@code true} erzwingt die Achsenordnung (longitude, latitude). Siehe
   *                       {@link CRS#decode(String, boolean)} (Bemerkung: {@code false} bedeutet
   *                       System-Default, <b>nicht</b> (latitude, longitude)!)
   * @param suppressWarnings wenn {@code true} werden Warnmeldungen unterdrueckt
   * @return eine nach dem CRS-Namen geordnete Map
   */
  public static final SortedMap<String,CoordinateReferenceSystem> getAvailableCRSByName(String authority, boolean longitudeFirst, boolean suppressWarnings) {
    SortedMap<String,CoordinateReferenceSystem> mapByCode = getAvailableCRSByCode(authority, longitudeFirst, suppressWarnings);
    TreeMap<String,CoordinateReferenceSystem>   mapByName = new TreeMap<String, CoordinateReferenceSystem>();
    for (String code : mapByCode.keySet()) {
      CoordinateReferenceSystem crs = mapByCode.get(code);
      mapByName.put( crs.getName().toString()+" ["+code+"]", crs );
    }
    return mapByName;
  }

  /**
   * Liefert alle zur Verfuegung stehenden {@linkplain CoordinateReferenceSystem CRS} fuer eine
   * Authority.
   * @param authority      Authority fuer die die CRS geliefert werden (z.B. {@code "EPSG"})
   * @param longitudeFirst {@code true} erzwingt die Achsenordnung (longitude, latitude). Siehe
   *                       {@link CRS#decode(String, boolean)} (Bemerkung: {@code false} bedeutet
   *                       System-Default, <b>nicht</b> (latitude, longitude)!)
   * @param suppressWarnings wenn {@code true} werden Warnmeldungen unterdrueckt
   * @return eine nach dem CRS-Code geordnete Map
   */
  public static final SortedMap<String,CoordinateReferenceSystem> getAvailableCRSByCode(String authority, boolean longitudeFirst, boolean suppressWarnings) {
    TreeMap<String,CoordinateReferenceSystem> map = new TreeMap<String, CoordinateReferenceSystem>();
    for (String code : (Set<String>)CRS.getSupportedCodes(authority)) {
      if ( !code.startsWith(authority) )
        code = authority + ":" + code;
      try {
        CoordinateReferenceSystem crs = CRS.decode(code, longitudeFirst);
        map.put(code, crs);
      } catch (Exception err) {
        if ( !suppressWarnings )
          LOGGER.warn("CRS could not be decoded: "+code);
      }
    }
    return map;
  }

  /**
   * Berechnet den Schnitt zweier {@link Envelope Envelopes}.
   * @param env1 erster Envelope
   * @param env2 zweiter Envelope
   * @param crs  {@link CoordinateReferenceSystem} fuer den Schnitt-Envelope
   */
  public static Envelope intersectEnvelope(Envelope env1, Envelope env2, CoordinateReferenceSystem crs) {
    // Envelopes in JTS umwandeln
    com.vividsolutions.jts.geom.Envelope env1JTS = new com.vividsolutions.jts.geom.Envelope(
      env1.getMinimum(0),env1.getMaximum(0),env1.getMinimum(1),env1.getMaximum(1));
    com.vividsolutions.jts.geom.Envelope env2JTS = new com.vividsolutions.jts.geom.Envelope(
      env2.getMinimum(0),env2.getMaximum(0),env2.getMinimum(1),env2.getMaximum(1));
    com.vividsolutions.jts.geom.Envelope intersetionEnvJTS = env1JTS.intersection(env2JTS);
    // Subset nur bzgl. des Bereichs in dem auch das Raster liegt
    return JTS.getEnvelope2D(intersetionEnvJTS,crs);
  }

}
