/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.geotools;

import java.awt.Rectangle;

import org.apache.log4j.Logger;
import org.geotools.geometry.jts.JTS;
import org.geotools.referencing.CRS;
import org.opengis.referencing.FactoryException;
import org.opengis.referencing.crs.CoordinateReferenceSystem;
import org.opengis.referencing.operation.MathTransform;
import org.opengis.referencing.operation.TransformException;

import com.vividsolutions.jts.geom.Coordinate;
import com.vividsolutions.jts.geom.Envelope;

/**
 * Diese Klasse enthaelt allgemeine Funktionen fuer die Arbeit mit den
 * in Geotools verwendeten JTS-Komponenten vereinfacht.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.1
 */
public class JTSUtil {
  private static Logger LOGGER = Logger.getLogger( JTSUtil.class.getName() );

  
  

  /**
   * Returns an {@link Envelope} that has the same aspect ratio as the given rectangle
   * @param rect
   *             defines the aspect ratio the map area is fixed with (e.g. a
   *             gui components size)
   * @param mapArea
   *             the map area to apply the aspect ratio of "r" to
   * @param grow
   *            If <code>true</code>, than the area will be enlarged to match
   *            the aspect ratio. If <code>false</code>, it will only shrink.
   */
  public static Envelope fixAspectRatio(final Rectangle rect, final Envelope mapArea,
          boolean grow) {
      // no map area to fix the aspect ratio for
      if (mapArea == null) {
          LOGGER.warn("mapArea has been null in method fixAspectRatio, returning also NULL");
          return null;
      }
      // no component size to fix the aspect ratio with
      if ( rect == null || rect.width == 0 || rect.height == 0 ) {
        LOGGER.warn("Empty rectangle in method fixAspectRatio, returning an unmodified mapArea!");
        return mapArea;
      }

      final double mapWidth = mapArea.getWidth(); /* get the extent of the map */
      final double mapHeight = mapArea.getHeight();
      final double scaleX = rect.getWidth() / mapArea.getWidth(); /*
                                                               * calculate the
                                                               * new scale
                                                               */

      final double scaleY = rect.getHeight() / mapArea.getHeight();
      double scale = 1.0; // stupid compiler!

      if ((grow && scaleX < scaleY) || (!grow && scaleX > scaleY)) {
          scale = scaleX;
      } else {
          scale = scaleY;
      }

      /* calculate the difference in width and height of the new extent */
      final double deltaX = /* Math.abs */((rect.getWidth() / scale) - mapWidth);
      final double deltaY = /* Math.abs */((rect.getHeight() / scale) - mapHeight);

      /*
       * System.out.println("delta x " + deltaX);
       * System.out.println("delta y " + deltaY);
       */

      /* create the new extent */
      final Coordinate ll = new Coordinate(
              mapArea.getMinX() - (deltaX / 2.0), mapArea.getMinY()
                      - (deltaY / 2.0));
      final Coordinate ur = new Coordinate(
              mapArea.getMaxX() + (deltaX / 2.0), mapArea.getMaxY()
                      + (deltaY / 2.0));

      return new Envelope(ll, ur);
  }

  
  /**
   * Created an (CRS-less) JTS-Envelope from an OpenGIS-Envelope.
   * @param envelope an OpenGIS-Envelope
   * @return an JTS-Envelope
   */
  public static Envelope createEnvelope(org.opengis.geometry.Envelope envelope) { // gt2-2.4.2
//  public static Envelope createEnvelope(org.opengis.spatialschema.geometry.Envelope envelope) { // gt2-2.3.4
    return new Envelope(
      envelope.getMinimum(0), // X1
      envelope.getMaximum(0), // X2
      envelope.getMinimum(1), // Y1
      envelope.getMaximum(1)  // Y2
    );
  }

  /**
   * Transformiert einen JTS-Envelope von einem CRS in ein anderes.
   * @param sourceEnv JTS-Envelope
   * @param sourceCRS CRS von {@code sourceEnv}
   * @param destCRS   CRS in das umgerechnet werden soll
   * @see CRS#findMathTransform(CoordinateReferenceSystem,CoordinateReferenceSystem)
   * @see JTS#transform(Envelope,MathTransform)
   */
  public static Envelope transformEnvelope(Envelope sourceEnv, CoordinateReferenceSystem sourceCRS, CoordinateReferenceSystem destCRS) {
    Envelope destEnv = null;
      MathTransform transform;
	try {
		transform = CRS.findMathTransform(sourceCRS, destCRS, true);
		destEnv = JTS.transform(sourceEnv,transform);
	} catch (FactoryException e) {
		LOGGER.warn("CRS tranformation for JTS envelope not successfully",e);
	} catch (TransformException e) {
		LOGGER.warn("CRS tranformation for JTS envelope not successfully",e);
	}
    return destEnv;
  }

  /**
   * Transformiert eine Koordinate von einem CRS in ein anderes.
   * @param sourceCoord Koordinate
   * @param sourceCRS   CRS von {@code sourceCoord}
   * @param destCRS     CRS in das umgerechnet werden soll
   * @see CRS#findMathTransform(CoordinateReferenceSystem,CoordinateReferenceSystem)
   * @see JTS#transform(Coordinate, Coordinate, MathTransform)
   */
  public static Coordinate transformCoordinate(Coordinate sourceCoord, CoordinateReferenceSystem sourceCRS, CoordinateReferenceSystem destCRS) {
    Coordinate destCoord = null;
    MathTransform transform;
    try {
        transform = CRS.findMathTransform(sourceCRS, destCRS, true);
        destCoord = JTS.transform(sourceCoord,null,transform);
    } catch (FactoryException e) {
        LOGGER.warn("CRS tranformation for JTS coordinate not successful",e);
    } catch (TransformException e) {
        LOGGER.warn("CRS tranformation for JTS coordinate not successful",e);
    }
    return destCoord;
  }
}
