/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.geotools.feature;

import java.io.IOException;

import org.geotools.data.FeatureReader;
import org.geotools.feature.Feature;
import org.geotools.feature.FeatureCollection;
import org.geotools.feature.FeatureIterator;
import org.geotools.feature.FeatureType;
import org.geotools.feature.collection.FeatureIteratorImpl;

/**
 * Diese Klasse implementiert einen {@link FeatureReader} ueber den
 * {@link FeatureIterator} einer {@link FeatureCollection}.<br>
 * <b>Beachte:</b><br>
 * Die <code>FeatureCollection</code> muss mindestens ein Element enthalten! Ansonsten
 * kann der {@link FeatureType} nicht ermittelt werden!
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class FeatureCollectionReader implements FeatureReader {
  private FeatureIterator   iterator    = null;
  private boolean           closed      = false;
  private FeatureType       featureType = null;
  private FeatureCollection fc          = null;

  /**
   * Erzeugt einen neuen FeatureReader.
   * @param fc FeatureCollection aus der gelesen wird.
   * @throws java.lang.UnsupportedOperationException falls die FeatureCollection
   *         kein Element enthaelt.
   */
  public FeatureCollectionReader(FeatureCollection fc) {
    if ( fc.isEmpty() )
      throw new UnsupportedOperationException("FeatureCollection must contain at least one Feature!");
    this.fc          = fc;
    this.featureType = new FeatureIteratorImpl(fc).next().getFeatureType();
    reset();
  }

  /**
   * Liefert die Art der Features.
   */
  public FeatureType getFeatureType() {
    return featureType;
  }

  /**
   * Prueft, ob ein weiteres Feature gelesen werden kann.
   */
  public boolean hasNext() {
    return !closed && iterator.hasNext();
  }

  /**
   * Liefert das naechste Feature (aus der FeatureCollection).
   * @exception IOException falls der Reader bereits geschlossen ist.
   */
  public Feature next() throws IOException {
    if (closed)
      throw new IOException("FeatureReader already closed!");
    return iterator.next();
  }

  /**
   * Schliesst den Reader. Danach koennen keine weiteren Features mehr
   * gelesen werden. Ueber {@link #reset()} kann der Reader wieder geoeffnet
   * werden.
   */
  public void close() {
    closed = true;
  }

  /**
   * Setzt den Reader auf den Anfang der {@link FeatureCollection} zurueck.
   */
  public void reset() {
    this.closed = false;
    this.iterator = new FeatureIteratorImpl(this.fc);
  }
}
