/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.geotools.feature;

import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.Vector;

import javax.swing.CellEditor;
import javax.swing.DefaultCellEditor;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JTable;
import javax.swing.JTextField;
import javax.swing.table.TableModel;

import org.geotools.feature.AttributeType;
import org.geotools.feature.AttributeTypeFactory;
import org.geotools.feature.DefaultAttributeType;
import org.geotools.feature.FeatureType;
import org.geotools.feature.FeatureTypeBuilder;
import org.geotools.feature.SchemaException;
import org.geotools.feature.type.GeometricAttributeType;
import org.geotools.gui.swing.referencing.CoordinateTableModel.CellRenderer;

import schmitzm.geotools.gui.GeotoolsGUIUtil;
import schmitzm.swing.BooleanInputOption;
import schmitzm.swing.ManualInputOption;
import schmitzm.swing.SelectionInputOption;
import schmitzm.swing.table.AbstractMutableTableModel;
import schmitzm.swing.table.ComponentRenderer;
import schmitzm.temp.BaseTypeUtil;

import com.vividsolutions.jts.geom.LineString;
import com.vividsolutions.jts.geom.Point;
import com.vividsolutions.jts.geom.Polygon;

/**
 * Diese Klasse stellt ein {@link TableModel} auf einem "unfertigen" {@link FeatureType}
 * dar. Dieses definiert vier Spalten:
 * <ol>
 *   <li>Attribut-Name als Text-Eingabe</li>
 *   <li>Attribut-Typ als Auswahl-Feld</li>
 *   <li>"Nillable"-Eigenschaft als Check-Box</li>
 *   <li>"AutoValue"-Eigenschaft als Check-Box</li>
 *   <li>Standard-Wert als Text-Eingabe</li>
 * </ol>
 * Da das {@link TableModel} auf einem {@link FeatureTypeBuilder} basiert,
 * bietet es die Moeglichkeit, Attribute hinzuzufuegen oder zu loeschen.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class FeatureTypeBuilderTableModel extends AbstractMutableTableModel {
  /** Zur Verfuegung stehende Attribut-Typen. */
  public static final Class[] ATTR_TYPES = new Class[] {
    String.class,
    Integer.class,
    Long.class,
    Byte.class,
    Boolean.class,
    Double.class,
    Float.class,
    BigInteger.class,
    BigDecimal.class,
    Point.class,
    LineString.class,
    Polygon.class
  };

  public static final String[] ATTR_TYPES_DESC = new String[ATTR_TYPES.length];
  static {
    for (int i=0; i<ATTR_TYPES_DESC.length; i++)
      ATTR_TYPES_DESC[i] = ATTR_TYPES[i].getSimpleName();
  }


  /** Die in der Tabelle dargestellten Attribute. */
  protected Vector<AttributeDefinition> attrDefinitions = new Vector<AttributeDefinition>();

  /** {@link FeatureTypeBuilder} der in der Tabelle dargestellt wird. */
  protected FeatureTypeBuilder featureTypeBuilder = null;

  /**
   * Erstellt ein leeres Tabellen-Modell.
   */
  public FeatureTypeBuilderTableModel() {
    this((FeatureTypeBuilder)null);
  }

  /**
   * Erstellt ein neues Tabellen-Modell.
   * @param type ein FeatureType, der editiert wird
   */
  public FeatureTypeBuilderTableModel(FeatureType type) {
    super();
    setFeatureType(type);
  }

  /**
   * Erstellt ein neues Tabellen-Modell.
   * @param type ein FeatureType, der editiert wird
   */
  public FeatureTypeBuilderTableModel(FeatureTypeBuilder builder) {
    super();
    setFeatureTypeBuilder(builder);
  }

  /**
   * Liefert die Spaltennamen der Tabelle.
   */
  public String[] createColumnNames() {
    return new String[] {
        GeotoolsGUIUtil.RESOURCE.getString("schmitzm.geotools.gui.FeatureTypeBuilderTableModel.AttrName"),
        GeotoolsGUIUtil.RESOURCE.getString("schmitzm.geotools.gui.FeatureTypeBuilderTableModel.AttrType"),
        GeotoolsGUIUtil.RESOURCE.getString("schmitzm.geotools.gui.FeatureTypeBuilderTableModel.Nillable"),
        GeotoolsGUIUtil.RESOURCE.getString("schmitzm.geotools.gui.FeatureTypeBuilderTableModel.AutoValue"),
        GeotoolsGUIUtil.RESOURCE.getString("schmitzm.geotools.gui.FeatureTypeBuilderTableModel.DefValue")
    };
  }

  /**
   * Setzt {@link CellRenderer} und {@link CellEditor} fuer die Tabelle, sowie
   * eine adaequate Zeilen-Hoehe.
   */
  @Override
  public void initTable(JTable table) {
    table.setRowHeight( new JComboBox(new String[] {"Dummy"}).getPreferredSize().height );
    table.getColumnModel().getColumn(0).setCellRenderer( new ComponentRenderer.JTextField() );
    table.getColumnModel().getColumn(1).setCellRenderer( new ComponentRenderer.JComboBox() );
    table.getColumnModel().getColumn(2).setCellRenderer( new ComponentRenderer.JCheckBox() );
    table.getColumnModel().getColumn(3).setCellRenderer( new ComponentRenderer.JCheckBox() );
    table.getColumnModel().getColumn(4).setCellRenderer( new ComponentRenderer.JTextField() );
    table.getColumnModel().getColumn(0).setCellEditor( new DefaultCellEditor( new JTextField() ) );
    table.getColumnModel().getColumn(1).setCellEditor( new DefaultCellEditor( new JComboBox( ATTR_TYPES_DESC ) ) );
    table.getColumnModel().getColumn(2).setCellEditor( new DefaultCellEditor( new JCheckBox() {
      public int getHorizontalAlignment() {
        return this.CENTER;
      }
      public int getVerticalAlignment() {
        return this.CENTER;
      }
    }));
    table.getColumnModel().getColumn(3).setCellEditor( new DefaultCellEditor( new JCheckBox() {
      public int getHorizontalAlignment() {
        return this.CENTER;
      }
      public int getVerticalAlignment() {
        return this.CENTER;
      }
    }));
    table.getColumnModel().getColumn(4).setCellEditor( new DefaultCellEditor( new JTextField() ) );
  }

  /**
   * Liefert die Anzahl an Attributen in der Tabelle.
   */
  public int getRowCount() {
    return attrDefinitions.size();
  }

  /**
   * Setzt den {@link FeatureTypeBuilder}, der in der Tabelle dargestellt wird.
   * @param builder {@link FeatureTypeBuilder}
   */
  public void setFeatureTypeBuilder(FeatureTypeBuilder builder) {
    if ( builder == null )
      builder = FeatureTypeBuilder.newInstance("New Type");

    this.featureTypeBuilder = builder;
    // Attribut-Typen in Array speichern
    attrDefinitions.clear();
    if ( featureTypeBuilder != null )
      for (int i=0; i<featureTypeBuilder.getAttributeCount(); i++) {
        AttributeType type = featureTypeBuilder.get(i);
        attrDefinitions.add( new AttributeDefinition(
            type.getLocalName(),
            type.getBinding(),
            type.isNillable(),
            false,
            type.createDefaultValue()
        ) );
      }
    this.fireTableDataChanged();
  }

  /**
   * Setzt den {@link FeatureTypeBuilder}, der in der Tabelle dargestellt wird.
   * Dieser wird mit den Attributen des angegebenen {@link FeatureType}
   * initialisiert
   * @param ftype {@link FeatureType}
   */
  public void setFeatureType(FeatureType ftype) {
    if ( ftype != null ) {
      FeatureTypeBuilder builder = FeatureTypeBuilder.newInstance(ftype.getTypeName());
      builder.addTypes(ftype.getAttributeTypes());
      setFeatureTypeBuilder(builder);
    } else
      setFeatureTypeBuilder(null);
  }

  /**
   * Erzeugt einen neuen {@link FeatureType}, aus den in der Tabelle dargestellen
   * Attributen.
   */
  public FeatureType createFeatureType() {
    try {
//      this.featureTypeBuilder.removeAll(); // does not work because of BUG in gt2-2.4.4
      for (;featureTypeBuilder.getAttributeCount() >0;)
        this.featureTypeBuilder.removeType(0);

      for (AttributeDefinition aDef : attrDefinitions) {
        AttributeType aType = aDef.createAttributeType();
        featureTypeBuilder.addType( aType );
        if ( featureTypeBuilder.getDefaultGeometry() == null &&
             aType instanceof GeometricAttributeType )
          featureTypeBuilder.setDefaultGeometry( (GeometricAttributeType)aType );
      }
      return this.featureTypeBuilder.getFeatureType();
    } catch (SchemaException err) {
      throw new RuntimeException(err);
    }
  }

  /**
   * Prueft, ob eine Tabellen-Zelle editierbar ist.
   * @return {@code true} fuer jede Zelle
   */
  public boolean isCellEditable(int rowIndex, int columnIndex) {
    return true;
  }

  /**
   * Liefert einen Wert der Tabelle.
   * @param rowIndex  Zeilen-Index (beginnend bei 0)
   * @param columnIndex Spalten-Index (beginnend bei 0)
   */
  public Object getValueAt(int rowIndex, int columnIndex) {
    AttributeDefinition aDef = attrDefinitions.elementAt(rowIndex);
    switch ( columnIndex ) {
      case 0: return aDef.name.getValue();
      case 1: return aDef.type.getSelectedDisplayItem();
      case 2: return aDef.nillable.getValue();
      case 3: return aDef.autoValue.getValue();
      case 4: return aDef.defaultValue.getValue();
    }
    return null;
  }

  /**
   * Setzt einen Wert der Tabelle.
   * @param value neuer Wert
   * @param rowIndex  Zeilen-Index (beginnend bei 0)
   * @param columnIndex Spalten-Index (beginnend bei 0)
   */
  public void setValueAt(Object value, int rowIndex, int columnIndex) {
    AttributeDefinition aDef = attrDefinitions.elementAt(rowIndex);
    switch ( columnIndex ) {
      case 0: aDef.name.setValue(value); break;
      case 1: aDef.type.setSelectedDisplayItem(value); break;
      case 2: aDef.nillable.setValue(value); break;
      case 3: aDef.autoValue.setValue(value); break;
      case 4: aDef.defaultValue.setValue(value); break;
    }
  }

  /**
   * Fuegt der Tabelle ein neues Standard-Attribut hinzu.
   */
  public void performAddRow() {
    AttributeDefinition newAttrDef = new AttributeDefinition();
    newAttrDef.name.setValue( GeotoolsGUIUtil.RESOURCE.getString("schmitzm.geotools.gui.FeatureTypeBuilderTableModel.NewAttr",getRowCount() ) );
    attrDefinitions.add( newAttrDef );
    fireTableDataChanged();
  }

  /**
   * Macht nichts.
   */
  public void performChangeData(int row, int col) {
    // does nothing
  }

  /**
   * Entfernt ein Attribut aus der Tabelle.
   */
  public void performRemoveRow(int row) {
    attrDefinitions.removeElementAt(row);
    fireTableDataChanged();
  }

  /**
   * Defines an Attribute displayed in the table. The advantage of this helper
   * class is that it does <b>not</b> check the consistency of the inputs
   * immediately on the input! This is importend richtig
   *
   * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
   *
   */
  protected class AttributeDefinition {
    /** Manual input field for the attribute name. */
    protected ManualInputOption.Text name;
    /** Choice input field for the attribute type.
     *  @see FeatureTypeBuilderTableModel.ATTR_TYPES */
    protected SelectionInputOption.Combo<Class> type = null;
    /** Manual input field for the attribute's default value. */
    protected ManualInputOption.Text defaultValue;
    /** Checkbox for the attribute's nillable property. */
    protected BooleanInputOption nillable = null;
    /** Checkbox to indicate that the attribute value is generated
     *  automatically. */
    protected BooleanInputOption autoValue = null;

    /**
     * Creates an empty attribute definition.
     */
    public AttributeDefinition() {
      name         = new ManualInputOption.Text(null, true); // Attribute name is mandatory
      type         = new SelectionInputOption.Combo<Class>(null,true,ATTR_TYPES,0,ATTR_TYPES_DESC);
      defaultValue = new ManualInputOption.Text(null,false); // Default value is not mandatory
      nillable     = new BooleanInputOption(null,true);
      autoValue    = new BooleanInputOption(null,false);
    }

    /**
     * Creates a new attribute definition. The consistency of the inputs (e.g. type and default
     * value) is not checked before {@link #createAttributeType()} is called.
     * @param name name of the attribute
     * @param type value type of the attribute
     * @param nillable indicates whether the attribute is mandatory
     * @param defaultValue default value for the attribute
     */
    public AttributeDefinition(String name, Class type, boolean nillable, boolean autoValue, Object defaultValue) {
      this();
      this.name.setValue(name);
      this.type.setValue(type);
      this.nillable.setValue(nillable);
      this.autoValue.setValue(autoValue);
      this.defaultValue.setValue(defaultValue);
    }

    /**
     * Creates an {@link AttributeType} from the definition.
     * @return {@link DefaultAttributeType} or {@link GeometricAttributeType}
     */
    public AttributeType createAttributeType() {
      Class   type         = (Class)this.type.getValue();
      boolean nillable     = (Boolean)this.nillable.getValue();
      Object  defaultValue = this.defaultValue.getValue();
      if ( "".equals(defaultValue) )
        defaultValue = null;
      if ( defaultValue != null &&  Number.class.isAssignableFrom(type) )
        defaultValue = BaseTypeUtil.convertFromString(defaultValue.toString(), type);
      if ( defaultValue == null && !nillable )
        defaultValue = FeatureUtil.getDefaultAttributeValue(type);

      AttributeType aType = AttributeTypeFactory.newAttributeType(
          (String)name.getValue(),
          type,
          nillable,
          null,
          defaultValue,
          null
      );

      AutoValueGenerator valueGenerator = null;
      if ( autoValue.getValue() ) {
        if ( Number.class.isAssignableFrom(type) )
          valueGenerator = new NumberValueGenerator((Number)defaultValue);
        else
          throw new UnsupportedOperationException( GeotoolsGUIUtil.RESOURCE.getString("schmitzm.geotools.gui.JEditorToolBar.NewLayer.Err.AutoVal", type.getSimpleName()));
        FeatureUtil.registerAutoValueGenerator(aType, valueGenerator);
      } else
        FeatureUtil.unregisterAutoValueGenerator(aType);


      return aType;
    }
  }
}
