/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.geotools.feature;


import java.awt.Color;
import java.io.IOException;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Locale;
import java.util.Map;
import java.util.Set;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.Vector;

import org.apache.log4j.Logger;
import org.geotools.data.DataUtilities;
import org.geotools.data.FeatureSource;
import org.geotools.data.memory.MemoryDataStore;
import org.geotools.factory.CommonFactoryFinder;
import org.geotools.factory.GeoTools;
import org.geotools.feature.AttributeType;
import org.geotools.feature.AttributeTypeFactory;
import org.geotools.feature.DefaultFeatureCollection;
import org.geotools.feature.DefaultFeatureCollections;
import org.geotools.feature.Feature;
import org.geotools.feature.FeatureCollection;
import org.geotools.feature.FeatureIterator;
import org.geotools.feature.FeatureType;
import org.geotools.feature.FeatureTypeBuilder;
import org.geotools.feature.GeometryAttributeType;
import org.geotools.feature.IllegalAttributeException;
import org.geotools.feature.SchemaException;
import org.geotools.feature.type.GeometricAttributeType;
import org.geotools.filter.FilterFactory;
import org.geotools.map.MapLayer;
import org.geotools.styling.Graphic;
import org.geotools.styling.Mark;
import org.geotools.styling.Style;
import org.geotools.styling.StyleBuilder;
import org.geotools.styling.Symbolizer;
import org.opengis.filter.FilterFactory2;
import org.opengis.referencing.crs.CoordinateReferenceSystem;

import schmitzm.lang.LangUtil;
import schmitzm.lang.ResourceProvider;

import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.GeometryFactory;
import com.vividsolutions.jts.geom.LineString;
import com.vividsolutions.jts.geom.MultiLineString;
import com.vividsolutions.jts.geom.MultiPoint;
import com.vividsolutions.jts.geom.MultiPolygon;
import com.vividsolutions.jts.geom.Point;
import com.vividsolutions.jts.geom.Polygon;


/**
 * Diese Klasse beinhaltet statische Methoden zum Arbeiten mit Features.
 *
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons Kr&uuml;ger</a>
 * @version 1.1
 */
public class FeatureUtil {
  /** {@link ResourceProvider}, der die Lokalisation fuer Komponenten
   *  des Package {@code schmitzm.geotools.feature} zur Verfuegung stellt. Diese sind
   *  in properties-Dateien unter {@code schmitzm.geotools.feature.resource.locales}
   *  hinterlegt. */
  public static ResourceProvider RESOURCE = new ResourceProvider( LangUtil.extendPackagePath(FeatureUtil.class,"resource.locales.FeatureResourceBundle"), Locale.ENGLISH );

	/**
	 * Convenience method to access the {@link ResourceProvider}.
	*/
	public static String R(String key, Object... values) {
		return RESOURCE.getString(key, values);
	}

  private static final Logger LOGGER = LangUtil.createLogger(FeatureUtil.class);

  private static final String DEFAULT_VECTOR_STYLE_NAME = "default vector style";

  private static final String DEFAULT_JOIN_ATTR = "OID";

  private static StyleBuilder STYLE_BUILDER = new StyleBuilder();

  /** Instance of JTS-GeometryFactory. */
  public static final GeometryFactory GEOMETRY_FACTORY = new GeometryFactory();

  /** Instance of AttributeTypeFactory. */
  public static final AttributeTypeFactory ATTRTYPE_FACTORY = AttributeTypeFactory.defaultInstance();

  /** Instance of {@link FilterFactory}. */
  public static final org.opengis.filter.FilterFactory FILTER_FACTORY = CommonFactoryFinder.getFilterFactory(GeoTools.getDefaultHints());
  
  /** Instance of {@link FilterFactory2}. */
  public static final FilterFactory2 FILTER_FACTORY2 = CommonFactoryFinder.getFilterFactory2(GeoTools.getDefaultHints());

  
  /** Join-Types. */
  public static enum JoinType {
    /** The feature of the left collection is taken into the result, even
     *  there is no join feature in the right collection. */
    LEFT_OUTER,
    /** The feature of the right collection is taken into the result, even
     *  there is no join feature in the left collection. */
    RIGHT_OUTER,
    /** Only features with a join feature in the other collection are
     *  taken into the result (Default). */
    FULL
  }

  /** The geometry type of a {@link FeatureCollection} */
  public static enum GeometryForm {
    /** Point */
    POINT,
    /** Line or LineString */
    LINE,
    /** Polygon */
    POLYGON
  }

  /**
   * Enthaelt die {@link AttributeType AttributeTypes}, fuer die
   * automatisch Werte generiert werden koennen.
   */
  private static HashMap<AttributeType, AutoValueGenerator> autoAttrValueGenerators = new HashMap<AttributeType, AutoValueGenerator>();

  /**
   * Determines the kind of geometry of a {@link FeatureType}.
   * @param fType a feature type
   */
  public static GeometryForm getGeometryForm(FeatureType fType) {
    if ( fType.getDefaultGeometry() == null )
      return null;
    
    GeometryAttributeType geometryType = fType.getDefaultGeometry();
	return getGeometryForm(geometryType);
  }

  /**
	 * Determines the kind of geometry of a {@link GeometryAttributeType}.
	 * 
	 * @param geometryType
	 *            a {@link GeometryAttributeType}
	 */
	public static GeometryForm getGeometryForm(
			GeometryAttributeType geometryType) {
		final Class geomType = geometryType.getType();

		if (Point.class.isAssignableFrom(geomType)
				|| MultiPoint.class.isAssignableFrom(geomType))
			return GeometryForm.POINT;
		else if (LineString.class.isAssignableFrom(geomType)
				|| MultiLineString.class.isAssignableFrom(geomType))
			return GeometryForm.LINE;
		else if (Polygon.class.isAssignableFrom(geomType)
				|| MultiPolygon.class.isAssignableFrom(geomType))
			return GeometryForm.POLYGON;

		throw new UnsupportedOperationException("Unknown geometry type: "
				+ geomType.getName());
	}

  /**
   * Determines the kind of geometry of a {@link FeatureCollection}.
   * @param fc a feature collection
   */
  public static GeometryForm getGeometryForm(FeatureCollection fc) {
    return getGeometryForm(fc.getSchema());
  }

  /**
   * Determines the kind of geometry of a {@link FeatureSource}.
   * @param fs a feature source
   */
  public static GeometryForm getGeometryForm(FeatureSource fs) {
    return getGeometryForm(fs.getSchema());
  }

  /**
   * Determines the kind of geometry of a {@link MapLayer}.
   * @param layer a map layer
   */
  public static GeometryForm getGeometryForm(MapLayer layer) {
    return getGeometryForm(layer.getFeatureSource());
  }

    /**
   * Extrahiert alle Geometrien aus einem Feature.
   * @param f Feature
   * @return Array aller Geometrien im Feature.
   */
  public static Geometry[] extractGeometriesToGeometry(Feature f) {
    Vector<Object> geomVec = new Vector<Object>();

    for (int j=0; j<f.getFeatureType().getAttributeCount(); j++)
//      if ( f.getFeatureType().getAttributeType(j).isGeometry() )
      if ( f.getFeatureType() instanceof GeometryAttributeType )
        geomVec.add(f.getAttribute(j));

    Geometry[] geomArr = new Geometry[geomVec.size()];
    for (int i=0; i<geomVec.size(); i++)
      geomArr[i] = (Geometry)geomVec.elementAt(i);
    return geomArr;
  }

  /**
   * Extrahiert alle Default-Geometrien aus einer FeatureCollection.
   * @param fc FeatureCollection
   * @return Array aller Geometrien in der FeatureCollection.
   */
  public static Vector<Geometry> extractGeometries(FeatureCollection fc, Vector<Geometry> result) {
    if ( result == null )
      result = new Vector<Geometry>();

    FeatureIterator fi = fc.features();
    for (;fi.hasNext();) {
      Feature f = fi.next();
      result.add( f.getDefaultGeometry() );
    }
    return result;
  }

  /**
   * Erzeugt einen Standard-Style fuer eine {@link FeatureCollection}
   * Und setzt eine default namen
   * @param fc FeatureCollection
   */
  public static Style createDefaultStyle(FeatureCollection fc) {
    GeometryAttributeType geometryAttrib = fc.getSchema().getDefaultGeometry();
    return createDefaultStyle( geometryAttrib );
  }

  /**
   * Erzeugt einen Standard-Style fuer einen {@link GeometryAttributeType}
   * Und setzt eine default Namen.
   * @param geometryAttrib GeometryAttributeType
   */
  public static Style createDefaultStyle(GeometryAttributeType geometryAttrib) {
    Style style = null;

    if ( geometryAttrib != null &&
        (com.vividsolutions.jts.geom.Polygon.class.isAssignableFrom(geometryAttrib.getType()) ||
         com.vividsolutions.jts.geom.MultiPolygon.class.isAssignableFrom(geometryAttrib.getType())) )
     style = createPolygonStyle(Color.ORANGE, Color.BLACK, 1);
    else if (geometryAttrib != null &&
             (com.vividsolutions.jts.geom.Point.class.isAssignableFrom(geometryAttrib.getType()) ||
              com.vividsolutions.jts.geom.MultiPoint.class.isAssignableFrom(geometryAttrib.getType())) )
      style = createPointStyle(Color.RED);
    else
      style = createLineStyle(Color.BLUE, 1);

    style.setName(DEFAULT_VECTOR_STYLE_NAME);
    style.setTitle(DEFAULT_VECTOR_STYLE_NAME);
    return style;
  }

  /**
   * Erzeugt einen Polygon-Style.
   * @param fillColor Fuell-Farbe
   * @param borderColor Rand-Farbe
   * @param borderWidth Breite des Rands
   */
  public static Style createPolygonStyle(Color fillColor, Color borderColor, double borderWidth) {
    final Symbolizer symb = STYLE_BUILDER.createPolygonSymbolizer(fillColor, borderColor, borderWidth);
    return STYLE_BUILDER.createStyle( symb );
  }

  /**
   * Erzeugt einen (randlosen) Punkt-Style in Form eines Kreises.
   * @param color Farbe des Punkts
   */
  public static Style createPointStyle(Color fillColor) {
    return createPointStyle(fillColor, fillColor, 5);
  }

  /**
   * Erzeugt einen (randlosen) Punkt-Style in Form eines Kreises.
   * @param fillColor Fuell-Farbe des Punkts
   * @param borderColor Rand-Farbe des Punkts
   * @param width Groesse des Punkts
   */
  public static Style createPointStyle(Color fillColor, Color borderColor, double width) {
    return createPointStyle(STYLE_BUILDER.MARK_CIRCLE, fillColor, borderColor, 1, 1, width, 0);
  }

  /**
   * Erzeugt einen Punkt-Style.
   * @param color Farbe des Punkts
   * @param markStyle Darstellungsweise des Punkts
   * @param opacity
   * @see StyleBuilder#MARK_ARROW;
   * @see StyleBuilder#MARK_CIRCLE;
   * @see StyleBuilder#MARK_CROSS;
   * @see StyleBuilder#MARK_SQUARE;
   * @see StyleBuilder#MARK_STAR;
   * @see StyleBuilder#MARK_TRIANGLE;
   * @see StyleBuilder#MARK_X;
   */
  public static Style createPointStyle(String markStyle, Color fillColor, Color borderColor, double borderWidth, double opacity, double size, double rotation) {
    final Mark       mark = STYLE_BUILDER.createMark(markStyle, fillColor, borderColor, borderWidth);
    final Graphic    g    = STYLE_BUILDER.createGraphic(null, mark, null, opacity, size, rotation);
    final Symbolizer symb = STYLE_BUILDER.createPointSymbolizer(g);
    return STYLE_BUILDER.createStyle( symb );
  }

  /**
   * Erzeugt einen Linien-Style.
   * @param lineColor Farbe der Linie
   */
  public static Style createLineStyle(Color lineColor) {
    return createLineStyle(lineColor, 1);
  }

  /**
   * Erzeugt einen Linien-Style.
   * @param lineColor Farbe der Linie
   * @param lineWidth Breite der Linie
   */
  public static Style createLineStyle(Color lineColor, double lineWidth) {
	  // TODO I don't like StyleBuilder!
    final Symbolizer symb = STYLE_BUILDER.createLineSymbolizer(lineColor, lineWidth);
    return STYLE_BUILDER.createStyle( symb );
  }


 /**
  * Erzeugt einen Array von {@link Feature Features} auf einer
  * {@link FeatureCollection}.<br>
  * Aufgrund von Problemen bei der Verwendung von Filtern ({@code .size()} liefert
  * nicht die Anzahl der gefilterten Features; {@code .toArray} liefert
  * manchmal {@code null}), wird der Array "naiv" ueber Durchlaufen
  * eines {@link FeatureIterator} kopiert.
  * @param fc eine Feature Collection
  * @return einen leeren Array, wenn die Collection leer oder {@code null} ist
  */
  public static Feature[] featuresToArray(FeatureCollection fc) {
//    if ( fc == null || fc.size() == 0 ) // bei leerer FC klappt .toArray nicht!!
//      return new Feature[0];
//    Feature[] featureArray = (Feature[])fc.toArray(new Feature[0]);
//    if ( featureArray != null )
//      return featureArray;
//    // Sonst "naiv" kopieren
//    featureArray = new Feature[fc.size()];
//    FeatureIterator fi = fc.features();
//    for (int i=0; fi.hasNext(); i++)
//      featureArray[i] = fi.next();
//    return featureArray;
      return featuresToArray(fc,true);
  }

  /**
   * Erzeugt einen Array von {@link Feature Features} auf einer
   * {@link FeatureCollection}.
   * @param fc
   *            eine Feature Collection
   * @param
   *        includeNullFeatures
   *            wenn {@code false} werden NULL-Features in der Collection nicht in den
   *            Array kopiert
   * @return einen leeren Array, wenn die Collection leer oder {@code null} ist
   */
   public static Feature[] featuresToArray(FeatureCollection fc, boolean includeNullFeatures) {
//     if ( includeNullFeatures )
//       return featuresToArray(fc);
     if ( fc == null )
       return new Feature[0];
     
     // Naiv alle Features durchgehen.
     Vector<Feature> fv = new Vector<Feature>();
     FeatureIterator fi = fc.features();
     for (int i=0; fi.hasNext(); i++) {
       Feature f = fi.next();
       if ( includeNullFeatures || f != null )
         fv.add(f);
     }
     return fv.toArray(new Feature[0]);
   }

   /**
   * Clones an {@link FeatureType}.
   * @param fType     type to clone
   * @param fTypeName the name for the clone (if {@code null} the name is
   *                  taken from {@code fType})
   * @param nillable  nillable property for <b>all</b> attributes (if {@code null} the
   *                  nillable property is taken individually from the source attributes)
   */
  public static FeatureType cloneFeatureType(FeatureType fType, String fTypeName, Boolean nillable) throws SchemaException {
    if ( fTypeName == null )
      fTypeName = fType.getTypeName();

    FeatureTypeBuilder builder = FeatureTypeBuilder.newInstance(fType.getTypeName());
    for (AttributeType aType : fType.getAttributeTypes() ) {
      AttributeType aTypeClone = cloneAttributeType(aType, nillable);
      if ( aTypeClone instanceof GeometryAttributeType && builder.getDefaultGeometry() == null )
        builder.setDefaultGeometry( (GeometryAttributeType)aTypeClone );
      else
        builder.addType(aTypeClone);
    }
    return builder.getFeatureType();
  }

  /**
   * Clones an {@link AttributeType}.
   * @param aType type to clone
   * @param nillable the clone gets this value for its nillable property independently
   *                 of {@code aType.isNillable()} (if {@code null} the nillable property
   *                 is taken from {@code aType})
   *
   */
  public static AttributeType cloneAttributeType(AttributeType aType, Boolean nillable) {
    // If no special nillable property is specified, take it from the
    // given AttributeType
    if ( nillable == null )
      nillable = aType.isNillable();

    // Create a default value for the attribute
    Object defaultValue = nillable ? null : getDefaultAttributeValue(aType);

    Object metaData = null;
    // if it is a GeometryAttributeType, the CRS must be stored
    // in the meta data
    if ( aType instanceof GeometryAttributeType )
      metaData = ((GeometryAttributeType)aType).getCoordinateSystem();

    // Create the clone
    return AttributeTypeFactory.newAttributeType(
            aType.getLocalName(),
            aType.getBinding(),
            nillable,
            aType.getRestriction(),
            defaultValue,
            metaData
           );
  }

  /**
   * Erweitert das Schema einer {@link FeatureCollection} um eine Reihe von
   * Attributen.
   * @param fType  zu erweiterndes Feature-Schema
   * @param aTypes Typen der neuen Attribute
   * @return eine <b>neue</b> Instanz von {@link FeatureType}
   * @throws SchemaException falls das Erweitern des Feature-Schemas scheitert
   * @throws NullPointerException falls {@code fType = null}
   */
  public static FeatureType extendFeatureType(FeatureType fType, AttributeType... aTypes) throws SchemaException {
//BEMERKUNG: Folgende Methode funktioniert, fuegt die abgeleiteten (alten)
//           Attribute aber hinten an die neuen Attribute an. Dies ist
//           nicht gewuenscht und erschwert auch den Aufbau des fValues-Array
//    FeatureType resultType = FeatureTypeBuilder.newFeatureType(
//        aTypes,                    // new Attributes
//        fType.getTypeName(),       // new type name
//        null,                      // namespace
//        false,                     // abstact?
//        new FeatureType[] { fType },
//        null
//    );
    FeatureTypeBuilder builder = FeatureTypeBuilder.newInstance(fType.getTypeName());
    builder.addTypes( fType.getAttributeTypes() ); // old attributes
    for (AttributeType aType : aTypes )
      try {
        if ( aType instanceof GeometryAttributeType && builder.getDefaultGeometry() == null )
          builder.setDefaultGeometry( (GeometryAttributeType)aType );
        else
          builder.addType(aType);
      } catch (IllegalArgumentException err) {
        builder.addType( AttributeTypeFactory.newAttributeType(
            aType.getLocalName()+"_2",
            aType.getBinding(),
            aType.isNillable(),
            aType.getRestriction(),
            aType.createDefaultValue(),
            null
        ));
      }
    FeatureType resultType = builder.getFeatureType();

    // As long as DataUtilities.defaultValue(AttributeType) does
    // not handle the default value correctly for not-nillable
    // attributes, create a new FeatureType with all attributes
    // nillable
// not necessary as long a correct default value can be created
// in createFeatureType(ResultSetMetaData, ...)
//    resultType = cloneFeatureType(resultType, null, true);

    return resultType;
  }
  
  /**
   * Orderes {@link Feature Features} according to an attribute (ascending). 
   * @param fi       the features to order
   * @param attrName the attribute to order by
   * @return a list of ordered features
   */
  public static Vector<Feature> sortFeatures(FeatureCollection fc, String attrName) {
    return sortFeatures(fc.features(), attrName);
  }

  /**
   * Orderes {@link Feature Features} according to an attribute (ascending). 
   * @param fi       an iterator for the features
   * @param attrName the attribute to order by
   * @return a list of ordered features
   */
  public static Vector<Feature> sortFeatures(FeatureIterator fi, String attrName) {
    // First create a SortedMap with CollisionList
    SortedMap<Comparable<?>, Vector<Feature>> sortedFeatureLists = new TreeMap<Comparable<?>, Vector<Feature>>();
    for (;fi.hasNext();) {
      Feature f = fi.next();
      // Check whether attribute value is Comparable
      Object attrValue  = f.getAttribute(attrName);
      if ( attrValue != null && !(attrValue instanceof Comparable) )
        throw new UnsupportedOperationException("Feature sort only supported for Comparable attributes: "+LangUtil.getSimpleClassName(attrValue));
      // Determine X value as sort attribute (NULL not permitted for XYDateset!)
      Comparable<?> xValue = (Comparable<?>)attrValue;
      if ( xValue == null )
        continue;
      
      // Check whether collision list exists
      Vector<Feature> collList = sortedFeatureLists.get(xValue);
      if ( collList == null ) {
        collList = new Vector<Feature>();
        sortedFeatureLists.put(xValue, collList);
      }
      // Add feature to collision list
      collList.add(f);
    }
    
    // Put the (now ordered) features from the collision lists
    // into an temporary list to create a list of Features
    Vector<Feature> sortedFeatures = new Vector<Feature>();
    for (Vector<Feature> collList : sortedFeatureLists.values())
      for (Feature f : collList)
        sortedFeatures.add(f);

    return sortedFeatures;
  }

  /**
   * Erweitert das Schema einer {@link FeatureCollection} um eine Reihe von
   * Attributen.
   * @param fc     zu erweiternde FeatureCollection
   * @param aTypes Typen der neuen Attribute
   * @return eine <b>neue</b> Instanz von {@link FeatureCollection}
   * @throws SchemaException falls das Erweitern des Feature-Schemas scheitert
   * @throws IllegalAttributeException falls das Einfuegen der Features in die
   *         neue {@link FeatureCollection} scheitert
   */
  public static FeatureCollection extendFeatureCollection(FeatureCollection fc, AttributeType... aTypes) throws SchemaException, IllegalAttributeException  {
    // Schema erstellen
    FeatureType fType = fc.getSchema();
    FeatureType resultType = extendFeatureType( fType, aTypes );

    // Neue Collection erstellen
    FeatureCollection resultFc = DefaultFeatureCollections.newCollection();
    FeatureIterator   fi       = fc.features();
    // Array fuer die Attribut-Werte eines Features
    Object[]          fValues  = new Object[resultType.getAttributeCount()];
    for ( ; fi.hasNext(); ) {
      // Werte der alten Attribute in Array schreiben
      fi.next().getAttributes(fValues);
      // Default-Werte der neuen Attribute in Array schreiben
      for ( int i=fType.getAttributeCount(); i<fValues.length; i++)
        fValues[i] = resultType.getAttributeType(i).createDefaultValue();
      // Erweitertes Feature erzeugen und FeatureCollection fuellen
      resultFc.add( resultType.create( fValues ) );
    }

    return resultFc;
  }

  /**
   * Fuehrt eine Join-Operation zwischen zwei {@link FeatureCollection} durch.
   * Hierfuer wird eine einfacher Nested-Loop-Join durchgefuehrt.
   * In den optionalen Parametern koennen Attribut-Namen angegeben werden, die
   * in die Ergebnis-Colletion uebernommen werden (Projektion).<br>
   * <b>Achtung:</b> Wird diese Option verwendet, sollte sichergestellt sein,
   * dass {@code fc1} und {@code fc2} KEINE gleichen Attributnamen hat! Ansonsten
   * wird (aus technischen Gruenden) das Attribut von {@code fc2} uebernommen.
   * @param fc1       erste FeatureCollection
   * @param joinAttr1 Attribut-Name in der ersten FeatureCollection
   * @param compareOp Operation, mit der der JOIN-Vergleich durchgefuehrt wird
   * @param fc2       zweiten FeatureCollection
   * @param joinAttr2 Attribut-Name in der zweiten FeatureCollection
   * @param projAttr  Attribute (von {@code fc1} und {@code fc2}), die in das
   *                  Ergebnis übernommen werden (Projektion)
   * @return eine <b>neue</b> Instanz von {@link FeatureCollection}
   * @throws SchemaException falls das Erweitern des Feature-Schemas scheitert
   * @throws IllegalAttributeException falls das Einfuegen der Features in die
   *         neue {@link FeatureCollection} scheitert
   */
  public static FeatureCollection joinFeatureCollection(JoinType joinType, FeatureCollection fc1, String joinAttr1, AttributeFilter compareOp, FeatureCollection fc2, String joinAttr2, String... projAttr) throws SchemaException, IllegalAttributeException  {
    // RIGHT-OUTER-JOIN reaslisiert als "inverser" LEFT-OUTER-JOIN
    // Bemerkung: nur Uebergangsloesung, da nicht ganz sauber!!
    //            Attribut-Reihenfolge wird vertauscht...
    if ( joinType == JoinType.RIGHT_OUTER )
//      throw new IllegalArgumentException("Join type RIGHT_OUTER not yet supported.");
      return joinFeatureCollection(
          JoinType.LEFT_OUTER,
          fc2,
          joinAttr2,
          compareOp.inverse(),
          fc1,
          joinAttr1,
          projAttr
      );

    // Flag, ob Projektion erfolgen soll
    boolean projection = projAttr.length > 0;

    // Schema erstellen
    FeatureType fType1 = fc1.getSchema();
    FeatureType fType2 = fc2.getSchema();
    FeatureType resultType = null;
    if ( fType1 == null )
      throw new UnsupportedOperationException("JOIN can not be performed. FeatureCollection 1 seems to be empty.");
    if ( fType2 == null )
      throw new UnsupportedOperationException("JOIN can not be performed. FeatureCollection 2 seems to be empty.");
    if ( fType1.getAttributeType(joinAttr1) == null )
      throw new UnsupportedOperationException("JOIN can not be performed. FeatureCollection 1 does not contain the JOIN-Attribute: "+joinAttr1);
    if ( fType2.getAttributeType(joinAttr2) == null )
      throw new UnsupportedOperationException("JOIN can not be performed. FeatureCollection 2 does not contain the JOIN-Attribute: "+joinAttr2);

    // Wenn keine Projektion, dann alle Attribute von fc1 mit denen
    // von fc2 vereinen; sonst neuen FeatureType aufbauen
    if ( !projection )
        resultType = extendFeatureType(fType1, fType2.getAttributeTypes());
    else {
      // Leeren Feature-Type erzeugen
      FeatureTypeBuilder builder = FeatureTypeBuilder.newInstance(fType1.getTypeName());
      // Projektions-Attribute hinzufuegen
      for (String attrName : projAttr) {
        // erste Wahl: Attribut von fc2 uebernehmen (da beim Belegen
        //             von "fValues" der Attributwert aus fc1 in der inneren
        //             FOR-Schleife ggf. durch den Attributwert von fc2
        //             ueberschrieben wird!)
        AttributeType aType = fType2.getAttributeType(attrName);
        // zweite Wahl: Attribut von fc1 uebernehmen
        if ( aType == null )
          aType = fType1.getAttributeType(attrName);
        // Attribut in den FeatureType einfuegen
        if ( aType != null )
          builder.addType( aType );
      }
      // FeatureType erzeugen
      resultType = builder.getFeatureType();
    }

    // Neue Collection erstellen
    DefaultFeatureCollection resultFc = (DefaultFeatureCollection)DefaultFeatureCollections.newCollection();
    FeatureIterator fi = fc1.features();
    // Array fuer die Attribut-Werte eines Features
    Object[] fValues = new Object[resultType.getAttributeCount()];

    // NestedLoopJoin: Alle Features der ersten Collection durchgehen
    //                 und die passenden in der zweiten Collection suchen
    AttributeFilter filter = compareOp.clone();
    for (Feature f1 = null; fi.hasNext(); ) {
      f1 = fi.next();
      // Werte der 1. Features in Array schreiben
      if ( !projection )
        // Alle Attribut-Werte aus fc1
        f1.getAttributes(fValues);
      else
        // Nur Werte der Projektions-Attribute aus fc1
        getAttributeValues(f1,fValues,0,projAttr);

      // Passende Features in 2. Colletion suchen
      filter.setAttributeName(joinAttr2);
      filter.setCompareValue(f1.getAttribute(joinAttr1));
      FeatureIterator joinedFeatures = fc2.subCollection(filter).features();

      // Wenn LEFT-OUTER-JOIN und JOIN nicht erfolgreich, ein Dummy
      // Feature mit Default-Werten in die Collection einfuegen
      // Sonst: JOIN-Paare "normal" bilden
      if ( !joinedFeatures.hasNext() && joinType == JoinType.LEFT_OUTER ) {
        if ( !projection )
          getDefaultAttributeValues(fType2,fValues,fType1.getAttributeCount());
        else
          getDefaultAttributeValues(fType2,fValues,0,projAttr);
        // Erweitertes Feature erzeugen und FeatureCollection fuellen
        resultFc.add( resultType.create( fValues ) );
      } else {
        for (Feature f2 = null; joinedFeatures.hasNext(); ) {
          f2 = joinedFeatures.next();
          // Werte der 2. Features in Array schreiben
          if ( !projection )
            getAttributeValues(f2,fValues,fType1.getAttributeCount());
          else
            getAttributeValues(f2,fValues,0,projAttr);
          // Erweitertes Feature erzeugen und FeatureCollection fuellen
          Feature feature = resultType.create( fValues );
          resultFc.add( feature );
        }
      }
    }

    return resultFc;
  }

  /**
   * Fuehrt eine (Full-Outer-) Join-Operation zwischen zwei {@link FeatureCollection}
   * durch.
   * Hierfuer wird eine einfacher Nested-Loop-Join durchgefuehrt.
   * In den optionalen Parametern koennen Attribut-Namen angegeben werden, die
   * in die Ergebnis-Colletion uebernommen werden (Projektion).<br>
   * <b>Achtung:</b> Wird diese Option verwendet, sollte sichergestellt sein,
   * dass {@code fc1} und {@code fc2} KEINE gleichen Attributnamen hat! Ansonsten
   * wird (aus technischen Gruenden) das Attribut von {@code fc2} uebernommen.
   * @param fc1       erste FeatureCollection
   * @param joinAttr1 Attribut-Name in der ersten FeatureCollection
   * @param compareOp Operation, mit der der JOIN-Vergleich durchgefuehrt wird
   * @param fc2       zweiten FeatureCollection
   * @param joinAttr2 Attribut-Name in der zweiten FeatureCollection
   * @param projAttr  Attribute (von {@code fc1} und {@code fc2}), die in das
   *                  Ergebnis übernommen werden (Projektion)
   * @return eine <b>neue</b> Instanz von {@link FeatureCollection}
   * @throws SchemaException falls das Erweitern des Feature-Schemas scheitert
   * @throws IllegalAttributeException falls das Einfuegen der Features in die
   *         neue {@link FeatureCollection} scheitert
   */
  public static FeatureCollection joinFeatureCollection(FeatureCollection fc1, String joinAttr1, AttributeFilter compareOp, FeatureCollection fc2, String joinAttr2, String... projAttr) throws SchemaException, IllegalAttributeException  {
    return joinFeatureCollection(JoinType.FULL, fc1, joinAttr1, compareOp, fc2, joinAttr2, projAttr);
  }

  /**
   * Fuehrt eine (Full-Outer-) Equi-Join-Operation zwischen zwei
   * {@link FeatureCollection} durch.
   * @param fc1       erste FeatureCollection
   * @param joinAttr1 Attribut-Name in der ersten FeatureCollection
   * @param fc2       zweiten FeatureCollection
   * @param joinAttr2 Attribut-Name in der zweiten FeatureCollection
   * @param projAttr  Attribute (von {@code fc1} und {@code fc2}), die in das
   *                  Ergebnis übernommen werden (Projektion)
   * @return eine <b>neue</b> Instanz von {@link FeatureCollection}
   * @throws SchemaException falls das Erweitern des Feature-Schemas scheitert
   * @throws IllegalAttributeException falls das Einfuegen der Features in die
   *         neue {@link FeatureCollection} scheitert
   * @see #joinFeatureCollection(FeatureCollection, String, AttributeFilter, FeatureCollection, String)
   * @see #AttributeFilter.EQUALS
   */
  public static FeatureCollection joinFeatureCollection(FeatureCollection fc1, String joinAttr1, FeatureCollection fc2, String joinAttr2, String... projAttr) throws SchemaException, IllegalAttributeException  {
    return joinFeatureCollection(fc1,joinAttr1,AttributeFilter.EQUALS,fc2,joinAttr2,projAttr);
  }

  /**
   * Liefert die Default-Werte fuer alle Attribute eines {@link FeatureType}.
   * @param ft ein Feature-Type
   */
  public static Object[] getDefaultAttributeValues(FeatureType ft) {
    return getDefaultAttributeValues(ft,null,0);
  }

  /**
   * Liefert die Default-Werte fuer eine Reihen von Attributwerten
   * eines FeatureTypes.
   * @param ft ein Feature-Type
   * @param destArray Array, in dem die Objekte abgelegt werden (wenn {@code null},
   *                 wird ein neuer Array erzeugt)
   * @param destStartIdx Start-Index in {@code destArray} ab dem die Attributwerte
   *                     abgelegt werden
   * @param attrName Attribute aus {@link f}
   * @return {@code destArray} oder einen neuen Array {@code Object[destStartIdx + attrName.length]}
   */
  public static Object[] getDefaultAttributeValues(FeatureType ft, Object[] destArray, int destStartIdx, String... attrName) {
    // Wenn keine Attribute angegeben sind, alle verwenden
    int attrCount = attrName.length > 0 ? attrName.length : ft.getAttributeCount();

    if ( destArray == null )
      destArray = new Object[destStartIdx + attrCount];
    if ( destArray.length < destStartIdx + attrCount )
      throw new IllegalArgumentException("Destination array "+destArray.length+" with start index "+destStartIdx+" too small for "+attrName.length+" attribute values!");

    // Wenn keine Attribute angegeben sind, alle verwenden
    for (int i=0; i<attrCount; i++) {
      AttributeType attrType = attrName.length > 0 ? ft.getAttributeType(attrName[i]) : ft.getAttributeType(i);
      // wenn attrType = null, dann gibt es das Attribut im FeatureType nicht
      // -> destArray nicht ueberschreiben
      if ( attrType != null )
        destArray[ destStartIdx+i ] = getDefaultAttributeValue(attrType);
    }
    return destArray;
  }

  /**
   * Liefert den Default-Wert fuer ein Attribut.
   * @param attrType Attribut-Typ
   * @return {@code null}, wenn das Attribut NILLABLE ist
   */
  public static Object getDefaultAttributeValue(AttributeType attrType) {
    return( getDefaultAttributeValue(attrType, true) );
  }

  /**
   * Liefert den Default-Wert fuer ein Attribut.
   * @param attrType Attribut-Typ
   * @param allowNull wenn {@code false}, wird in jedem Fall ein Wert ungleich
   *                  {@code null} geliefert, auch wenn das Attribut NILLABLE ist
   */
  public static Object getDefaultAttributeValue(AttributeType attrType, boolean allowNull) {
//    AutoValueGenerator avg = getAutoValueGenerator(attrType);
//    Object attrValue = (avg == null) ? attrType.createDefaultValue() : avg.getNextValue();
    Object attrValue = attrType.createDefaultValue();
    if ( attrValue == null && ( !attrType.isNillable() || !allowNull ) ) {
      attrValue = getDefaultAttributeValue( attrType.getBinding() );
      if ( attrValue == null )
        LOGGER.warn("Could not create default value for not-null attribute '"+attrType.getLocalName()+"': "+attrType.getBinding().getSimpleName());
    }
    return attrValue;
  }

  /**
   * Liefert den Default-Wert fuer ein Attribut.
   * @param attrType Attribut-Typ
   * @return {@code null}, falls {@link DataUtilities#defaultValue(Class)} keinen
   *         Defaul-Wert ermitteln kann (anstelle einer Exception!)
   */
  public static Object getDefaultAttributeValue(Class attrType) {
    Object attrValue = null;
    try {
      // DataUtilities.defaultValue(.) kann BigDecimal und BigInteger
      // nicht verarbeiten (gibt null zurueck!)
      if ( BigDecimal.class.isAssignableFrom( attrType ) )
        return new BigDecimal(0.0);
      if ( BigInteger.class.isAssignableFrom( attrType ) )
        return new BigInteger("0");
      attrValue = DataUtilities.defaultValue(attrType);
    } catch (IllegalArgumentException err) {
      // no default value could be generated for this type
      // --> ignore that
    }
    return attrValue;
  }

  /**
   * Liefert eine Reihen von Attributwerten eines Features.
   * @param f ein Feature
   * @param destArray Array, in dem die Objekte abgelegt werden (wenn {@code null},
   *                 wird ein neuer Array erzeugt)
   * @param destStartIdx Start-Index in {@code destArray} ab dem die Attributwerte
   *                     abgelegt werden
   * @param attrName Attribute des Features, deren Werte geliefert werden (wenn
   *                 {@code null} werden alle Attribute geliefert)
   * @return {@code destArray} oder einen neuen Array {@code Object[destStartIdx + attrName.length]}
   */
  public static Object[] getAttributeValues(Feature f, Object[] destArray, int destStartIdx, String... attrName) {
    if ( destArray == null )
      destArray = new Object[destStartIdx + attrName.length];
    if ( destArray.length < destStartIdx + attrName.length )
      throw new IllegalArgumentException("Destination array "+destArray.length+" with start index "+destStartIdx+" too small for "+attrName.length+" attribute values!");

    // Wenn keine Attribute angegeben sind, alle verwenden
    int attrCount = attrName.length > 0 ? attrName.length : f.getNumberOfAttributes();
    for (int i=0; i<attrCount; i++) {
      Object attr = attrName.length > 0 ? f.getAttribute(attrName[i]) : f.getAttribute(i);
      // wenn attr = null, dann gibt es das Attribut im Feature nicht
      // -> destArray nicht ueberschreiben
      if ( attr != null )
        destArray[ destStartIdx+i ] = attr;
    }
    return destArray;
  }

  /**
   * Liefert eine Reihen von Attributwerten eines Features.
   * @param f ein Feature
   * @param attrName Attribute des Features, deren Werte geliefert werden (wenn
   *                 {@code null} werden alle Attribute geliefert)
   * @return neuer Array {@code Object[attrName.length]}
   */
  public static Object[] getAttributeValues(Feature f, String... attrName) {
    return getAttributeValues(f,null,0,attrName);
  }

  /**
   * Liefert eine Reihen von Attributwerten eines Features.
   * @param f ein Feature
   * @param destMap Map, in dem die Objekte abgelegt werden (wenn {@code null},
   *                 wird eine neue {@link HashMap} erzeugt)
   * @param attrName Attribute des Features, deren Werte geliefert werden (wenn
   *                 {@code null} werden alle Attribute geliefert)
   * @return {@code destArray} oder einen neuen Array {@code Object[destStartIdx + attrName.length]}
   */
  public static Map<String,Object> getAttributeValues(Feature f, Map<String,Object> destMap, String... attrName) {
    if ( destMap == null )
      destMap = new HashMap<String,Object>();

    if ( attrName.length > 0 )
      for (int i=0; i<attrName.length; i++)
        destMap.put( attrName[i], f.getAttribute(attrName[i]) );
    else
      for (int i=0; i<f.getNumberOfAttributes(); i++)
        destMap.put( f.getFeatureType().getAttributeType(i).getName(), f.getAttribute(i) );

    return destMap;
  }
  
  /**
   * Fuehrt eine Join-Operation zwischen einer {@link FeatureCollection} und
   * einem {@link ResultSet} durch, in dem Zusatz-Informationen hinterlegt.
   * Hierzu wird das ResultSet zunaecht in eine {@link FeatureCollection}
   * umgewandelt und anschliessend ein Left-Outer-EquiJoin zwischen
   * den beiden Collections ausgewertet.
   * @param fc          eine FeatureCollection
   * @param fcJoinAttr  JOIN-Attribut in der FeatureCollection
   * @param rs          ResultSet mit Zusatz-Informationen
   * @param rsJoinAttr  JOIN-Attribut im ResultSet
   * @return eine <b>neue</b> Instanz von {@link FeatureCollection}
   * @throws SchemaException falls das Erweitern des Feature-Schemas scheitert
   * @throws IllegalAttributeException falls das Einfuegen der Features in die
   *         neue {@link FeatureCollection} scheitert
   */
  public static FeatureCollection joinFeatureCollection(FeatureCollection fc, String fcJoinAttr, ResultSet rs, String rsJoinAttr, String... projAttr) throws SchemaException, IllegalAttributeException, SQLException  {
    return joinFeatureCollection(fc, fcJoinAttr, AttributeFilter.EQUALS, rs, rsJoinAttr);
  }

  /**
   * Fuehrt eine Join-Operation zwischen einer {@link FeatureCollection} und
   * einem {@link ResultSet} durch, in dem Zusatz-Informationen hinterlegt.
   * Hierzu wird das ResultSet zunaecht in eine {@link FeatureCollection}
   * umgewandelt und anschliessend ein Left-Outer-Join zwischen
   * den beiden Collections ausgewertet.
   * @param fc          eine FeatureCollection
   * @param fcJoinAttr  JOIN-Attribut in der FeatureCollection
   * @param compareOp   Operation, mit der der JOIN-Vergleich durchgefuehrt wird
   * @param rs          ResultSet mit Zusatz-Informationen
   * @param rsJoinAttr  JOIN-Attribut im ResultSet
   * @return eine <b>neue</b> Instanz von {@link FeatureCollection}
   * @throws SchemaException falls das Erweitern des Feature-Schemas scheitert
   * @throws IllegalAttributeException falls das Einfuegen der Features in die
   *         neue {@link FeatureCollection} scheitert
   */
  public static FeatureCollection joinFeatureCollection(FeatureCollection fc, String fcJoinAttr, AttributeFilter compareOp, ResultSet rs, String rsJoinAttr, String... projAttr) throws SchemaException, IllegalAttributeException, SQLException  {
    return joinFeatureCollection(JoinType.LEFT_OUTER, fc, fcJoinAttr, compareOp, rs, rsJoinAttr, projAttr);

  }

  /**
   * Fuehrt eine Join-Operation zwischen einer {@link FeatureCollection} und
   * einem {@link ResultSet} durch, in dem Zusatz-Informationen hinterlegt.
   * Hierzu wird das ResultSet zunaecht in eine {@link FeatureCollection}
   * umgewandelt und anschliessend ein Join zwischen
   * den beiden Collections ausgewertet.
   * @param joinType    Join-Typ
   * @param fc          eine FeatureCollection
   * @param fcJoinAttr  JOIN-Attribut in der FeatureCollection
   * @param compareOp   Operation, mit der der JOIN-Vergleich durchgefuehrt wird
   * @param rs          ResultSet mit Zusatz-Informationen
   * @param rsJoinAttr  JOIN-Attribut im ResultSet
   * @return eine <b>neue</b> Instanz von {@link FeatureCollection}
   * @throws SchemaException falls das Erweitern des Feature-Schemas scheitert
   * @throws IllegalAttributeException falls das Einfuegen der Features in die
   *         neue {@link FeatureCollection} scheitert
   */
  public static FeatureCollection joinFeatureCollection(JoinType joinType, FeatureCollection fc, String fcJoinAttr, AttributeFilter compareOp, ResultSet rs, String rsJoinAttr, String... projAttr) throws SchemaException, IllegalAttributeException, SQLException  {
    if ( fcJoinAttr == null || fcJoinAttr.trim().equals("") ) {
      // Erste Nicht-Geometrie-Spalte suchen
      for (AttributeType aType : fc.getSchema().getAttributeTypes()) {
        if ( aType != fc.getSchema().getDefaultGeometry() ) {
          fcJoinAttr = aType.getName();
          break;
        }
      }
      LOGGER.warn("Join-Attribut for FeatureCollection not set. First attribute used: "+fcJoinAttr);
    }
    if ( rsJoinAttr == null || rsJoinAttr.trim().equals("") ) {
      rsJoinAttr = rs.getMetaData().getColumnName(1);
      LOGGER.warn("Join-Attribut for ResultSet not set. First column used: "+rsJoinAttr);
    }

    // FeatureCollection aus ResultSet erzeugen
    FeatureCollection rsFeatureCollection = createFeatureCollection(
        rs,
        // fuer den JOIN muss das Join-Attribut vorhanden sein!
        projAttr.length > 0 ? LangUtil.extendArray(projAttr,rsJoinAttr) : projAttr
    );

    // JOIN durchfuehren
    return joinFeatureCollection(
        joinType,              // Join-Typ
        fc,                    // FeatureCollection
        fcJoinAttr,            // Join-Attr. in FeatureCollection
        compareOp,             // Join-Operator
        rsFeatureCollection,   // ResultSet als FeatureCollection
        rsJoinAttr,            // Join-Attr. in ResultSet
        projAttr               // Projektions-Attribute
    );
  }


  /**
   * Erzeugt eine {@link FeatureCollection} aus einem {@linkplain ResultSet SQL-Result}.
   * @param  rs        SQL-Result
   * @param  projAttr  nur diese Attribute des ResultSet werden in die FeatureCollection
   *                   uebernommen (wenn nicht angegeben, werden ALLE Attribute
   *                   uebernommen)
   * @throws SQLException falls der Zugriff auf die SQL-Metadaten scheitert
   * @throws SchemaException falls das Erzeugen des FeatureType scheitert
   * @throws IllegalAttributeException falls das Erzeugen eines Features scheitert
   */
  public static FeatureCollection createFeatureCollection(ResultSet rs, String... projAttr) throws SchemaException, IllegalAttributeException, SQLException {
    // Schema der FeatureCollection erzeugen
    FeatureType rsFeatureType = createFeatureType(rs.getMetaData(),projAttr);

    // Array fuer die Attribut-Werte eines Features
    Object[] rsFeatureValues = new Object[rsFeatureType.getAttributeCount()];

    // ResultSet in FeatureCollection uebertragen
    FeatureCollection rsFeatureCollection = DefaultFeatureCollections.newCollection();
    for ( ;rs.next(); ) {
      for (int i=0; i<rsFeatureType.getAttributeCount(); i++)
        rsFeatureValues[i] = rs.getObject( rsFeatureType.getAttributeType(i).getName() );
      rsFeatureCollection.add( rsFeatureType.create(rsFeatureValues) );
    }

    return rsFeatureCollection;
  }

  /**
   * Erzeugt einen {@link FeatureType} aus einem {@linkplain ResultSetMetaData SQL-Result}.
   * @param  metaData         SQL-Metadaten des {@link ResultSet}
   * @param  projAttr         nur diese Attribute des ResultSet werden in den FeatureType
   *                          uebernommen (wenn nicht angegeben, werden ALLE
   *                          Attribute uebernommen)
   * @throws SQLException falls der Zugriff auf die SQL-Metadaten scheitert
   * @throws SchemaException falls das Erzeugen des FeatureType scheitert
   */
  public static FeatureType createFeatureType(ResultSetMetaData metaData, String... projAttr) throws SQLException, SchemaException {
    try {
      return createFeatureType(metaData, null, Object.class, projAttr);
    } catch (ClassNotFoundException err) {
      // KANN NICHT VORKOMMEN, da Object.class als Standard-Klasse angegeben
      // ist. Exception wird nur abgefangen, damit sie nicht in throw-clause
      // deklariert (und in uebergeordneter Methode) abgefangen werden muss
      return null;
    }
  }

  /**
   * Erzeugt einen {@link FeatureType} aus einem {@linkplain ResultSet SQL-Result}.
   * @param  metaData         SQL-Metadaten des {@link ResultSet}
   * @param  featureTypeName  Name fuer den FeatureType (kann {@code null} sein)
   * @param  defaultAttrClass Klasse, die verwendet als Attribut-Typ wird, wenn
   *                          die entsprechende Klasse des ResultSet nicht zur
   *                          Verfuegung steht (wenn {@code null}, wird eine
   *                          Exception geworfen)
   * @param  projAttr         nur diese Attribute des ResultSet werden in den FeatureType
   *                          uebernommen (wenn nicht angegeben, werden ALLE
   *                          Attribute uebernommen)
   * @throws SQLException falls der Zugriff auf die SQL-Metadaten scheitert
   * @throws SchemaException falls das Erzeugen des FeatureType scheitert
   * @throws ClassNotFoundException falls eine Attribut-Klasse aus dem ResultSet
   *         nicht erzeugt werden kann und keine Default-Klasse angegeben ist
   */
  public static FeatureType createFeatureType(ResultSetMetaData metaData, String featureTypeName, Class defaultAttrClass, String... projAttr) throws SQLException, SchemaException, ClassNotFoundException  {
    if ( featureTypeName == null )
      featureTypeName = "ResultSetFeatureType";

    // Projektions-Attribute in ArrayList uebertragen
    Set<String> projAttrList = new HashSet<String>();
    for (String attrName : projAttr)
      projAttrList.add(attrName);

    // Fuer alle Attribute aus Metadaten einen AttributeType erzeugen
    // und diesen in FeatureType uebernehmen
    FeatureTypeBuilder   featureTypeFac = FeatureTypeBuilder.newInstance(featureTypeName);
    for (int i=1; i<=metaData.getColumnCount(); i++) {
      String columnName = metaData.getColumnName(i);
      // Attribut nur uebernehmen, wenn als Projektions-Attribut angegeben
      if ( projAttr.length > 0 && !projAttrList.contains(columnName) )
        continue;

      Class columnClass = defaultAttrClass;
      try {
        columnClass = Class.forName( metaData.getColumnClassName(i) );
      } catch (ClassNotFoundException err) {
        if ( defaultAttrClass == null )
          throw err;
        // Fehler Ignorieren und Object als Attribut-Klasse verwenden
        LOGGER.warn("Column '"+columnName+"' ("+i+"): Class '"+metaData.getColumnClassName(i)+ "' not found. '"+columnClass.getName()+"' used.");
      }
      boolean nillable     = metaData.isNullable(i) != ResultSetMetaData.columnNoNulls;
      Object  defaultValue = getDefaultAttributeValue(columnClass);
      if ( defaultValue == null && !nillable)
        LOGGER.warn("Could not create default value for not-null attribute '"+columnName+"': "+columnClass.getSimpleName());

      // Add an attribute for the column to the FeatureType
      featureTypeFac.addType( AttributeTypeFactory.newAttributeType(
        columnName,
        columnClass,
        nillable,
        null, // no restrictions
        defaultValue,
        null  // no meta data
      ));
    }
    return featureTypeFac.getFeatureType();
  }

  /**
   * Erzeugt einen {@link FeatureType} aus Beispiel-Attribut-Werten.
   * @param  featureTypeName  Name fuer den FeatureType (kann {@code null} sein)
   * @param  attrValues       Attribut-Werte
   * @param  defaultAttrClass Standard-Typ, der fuer ein Attribut verwendet wird,
   *                          wenn der angegebene Attribut-Wert {@code null} ist.
   * @throws SchemaException falls das Erzeugen des FeatureType scheitert
   */
  public static FeatureType createFeatureType(String featureTypeName, Map<String,Object> attrValues,  Class defaultAttrClass) throws SchemaException  {
    if ( featureTypeName == null )
      featureTypeName = "FeatureType";

    FeatureTypeBuilder featureTypeFac = FeatureTypeBuilder.newInstance(featureTypeName);
    for (String attrName : attrValues.keySet()) {
      Object attrValue = attrValues.get( attrName );
      Class  attrClass = attrValue == null ? defaultAttrClass : attrValue.getClass();
      featureTypeFac.addType( AttributeTypeFactory.newAttributeType(
        attrName,
        attrClass
      ));
    }
    return featureTypeFac.getFeatureType();
  }

  /**
   * Erzeugt einen Geometrie-Attribut-Typ.
   * @param geomType  Art der Geometrie
   * @param defGeom   Standard-Wert
   * @param crs       Koordinatensystem
   */
  public static GeometricAttributeType createGeometryAttributeType(Class<? extends Geometry> geomType, Geometry defGeom, CoordinateReferenceSystem crs) {
    return new GeometricAttributeType(
      "the_geom",
      geomType,
      false,
      defGeom,
      crs,
      null
    );
  }

  /**
   * Registriert einen {@link AutoValueGenerator} fuer einen {@link AttributeType}, so
   * dass {@link #getNextAutoValue(AttributeType)} fuer diesen {@link AttributeType}
   * verwendet werden kann.
   * @param aType ein {@link AttributeType}
   * @param generator generiert automatische Attribut-Werte
   */
  public static void registerAutoValueGenerator(AttributeType aType, AutoValueGenerator generator) {
    autoAttrValueGenerators.put(aType, generator);
  }

  /**
   * Entfernt einen {@link AutoValueGenerator} fuer einen {@link AttributeType}.
   * @param aType ein {@link AttributeType}
   */
  public static void unregisterAutoValueGenerator(AttributeType aType) {
    if ( autoAttrValueGenerators.containsKey( aType ) )
      autoAttrValueGenerators.remove(aType);
  }

  /**
   * Liefert den {@link AutoValueGenerator} fuer einen {@link AttributeType}.
   * @param aType ein {@link AttributeType}
   * @return {@code null}, wenn fuer den {@link AttributeType} noch kein
   *         {@link AutoValueGenerator} registriert worden ist
   */
  public static AutoValueGenerator getAutoValueGenerator(AttributeType aType) {
    return autoAttrValueGenerators.get(aType);
  }

  /**
   * Liefert den naechsten Auto-Wert fuer einen {@link AttributeType}.
   * @param aType ein {@link AttributeType}
   * @exception UnsupportedOperationException falls fuer das Attribut noch
   *            kein {@link AutoValueGenerator} registriert ist
   */
  public static Object getNextAutoValue(AttributeType aType) {
    AutoValueGenerator avg = getAutoValueGenerator(aType);
    if ( avg == null )
      throw new UnsupportedOperationException("No AutoValueGenerator registered for attribute type: "+aType.getLocalName());
    return avg.getNextValue();
  }
  

	/**
	 * Creates a sample Feature instance of the given {@link FeatureType}
	 * 
	 * @param schema
	 *            the schema for which to create a sample Feature instance
	 *            
	 * @author SK
	 * 
	 * @throws IllegalAttributeException
	 * 
	 */
	public static Feature createSampleFeature(FeatureType schema) {
		Feature sampleFeature;

		Object[] attributes = getDefaultAttributeValues(schema);

		try {
			sampleFeature = schema.create(attributes);
		} catch (IllegalAttributeException e) {
			// Thrown if the FeatureType does not validate the attributes
			// Not very likely if we just create a sample feature, so we make it
			// a runtime exception.
			throw new RuntimeException(e);
		}

		return sampleFeature;
	}

	/**
	 * Erstellt eine {@link FeatureSource} basierend auf einem {@link MemoryDataStore} für eine übergebene {@link FeatureCollection}
	 * @param features 
	 * @return
	 */
	public static FeatureSource createMemoryFeatureSource(
			FeatureCollection features) {
		try {
			return  new MemoryDataStore(
					features).getFeatureSource(features
					.getSchema().getTypeName());
		} catch (IOException e) {
			LOGGER.error("Creating a memory datastore for the allFeaturesCollection", e);
			throw new RuntimeException("Creating a memory datastore for the allFeaturesCollection",e);
		}
	}


//  /**
//   * Extrahiert alle Geometrien aus einer FeatureCollection. Fuer jedes
//   * Geometry-Attribut der FeatureCollection wird eine GeometrieCollection
//   * (<code>JTSGeometries</code>) angelegt, in der die Geometrien des jeweiligen
//   * Attributs zusammengefasst werden.
//   * @param fc FeatureCollection
//   * @return Array aller Geometrien im Feature.
//   * @see org.geotools.feature.FeatureCollection
//   * @see org.geotools.renderer.geom.JTSGeometries
//   */
//  public static GeometryCollection[] extractGeometriesToGeometryCollection(FeatureCollection fc) throws org.opengis.referencing.operation.TransformException{
//    FeatureIterator it        = fc.features(); // Zeiger durch die einzelnen Features
//    Vector          gcVec     = new Vector();  // Speichert die Collections
//    int             nextGCIdx = 0;             // verweist innerhaln eines Features
//                                               // auf die naechste zu verwendende
//                                               // Collection
//
//    // alle Features nach Geometrien durchsuchen
//    for (int i=0; it.hasNext(); i++) {
//      Feature f = it.next();
//
//      // Das erste Geometry-Attribut des Features wird in GeometryCollection
//      // eins gespeichert, das zweite in der zweiten, usw.
//      // Bei jedem Feature wird also wieder von vorne begonnen.
//      nextGCIdx = 0;
//
//      // Alle Attribute auf Geometrien checken
//      for (int j=0; j<f.getFeatureType().getAttributeCount(); j++) {
//        // Handelt es sich um eine Geometrie, wird sie in einer
//        // GeometryCollection gespeichert (fuer jedes Geometry-Attr.
//        // gibt es eine eigene Collection!!)
////        if ( f.getFeatureType().getAttributeType(j).isGeometry() ){
//        if ( f.getFeatureType() instanceof GeometryAttributeType ) {
//          // gibt es noch keine Collection fuer das Attr., wird eine
//          // neue erstellt
//          if ( nextGCIdx >= gcVec.size() )
//            gcVec.add( new JTSGeometries() );
//          // Geometrie in Collection einfuegen
//          ((JTSGeometries)gcVec.elementAt(nextGCIdx++)).add((Geometry)f.getAttribute(j));
//        }
//      }
//    }
//
//    GeometryCollection[] gcArr = new GeometryCollection[gcVec.size()];
//    for (int i=0; i<gcVec.size(); i++)
//      gcArr[i] = (GeometryCollection)gcVec.elementAt(i);
//    return gcArr;
//  }
  
}

