/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.geotools.gui;

import java.util.HashMap;
import java.util.Vector;

import javax.swing.table.AbstractTableModel;

import org.geotools.feature.AttributeType;
import org.geotools.feature.Feature;
import org.geotools.feature.FeatureCollection;
import org.geotools.feature.FeatureCollections;
import org.geotools.feature.FeatureIterator;
import org.geotools.feature.FeatureType;

import schmitzm.geotools.feature.AttributeTypeFilter;
import schmitzm.geotools.feature.FeatureUtil;

/**
 * Tabellen-Modell fuer eine FeatureCollection. Aus Effizienzgruenden
 * (um wahlfreien Zugriff zu erreichen) werden die Features beim Aufruf von
 * {@link #setFeatureCollection(FeatureCollection)} in ein Array kopiert.
 * Veraenderungen an der zugrunde liegenden {@link FeatureCollection} werden
 * somit nicht automatisch in das Tabellen-Modell uebernommen, sondern es ist
 * ein expliziter Aufruf von {@link #reorganize()} notwendig.<br>
 * Bis zu diesem Punkt arbeitet diese Tabellen-Modell identisch zum
 * {@code FeatureTableModel} von Geotools. Im Gegensatz zu
 * {@link org.geotools.gui.swing.table.FeatureTableModel org.geotools.gui.swing.table.FeatureTableModel}.
 * werden die Tabellennamen und die Anzahl an Spalten jedoch nicht bei jedem Zugriff
 * neu aus dem ersten Feature ermittelt, sondern global gespeichert. Dies ist
 * erffizienter, da das haeufige Oeffnen eines {@link FeatureIterator FeatureIterators}
 * vermieden wird.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class FeatureCollectionTableModel extends AbstractTableModel {
  /** Holds the feature table that will be represented by this model. */
  protected FeatureCollection featureTable = null;
  /** Array mit den Daten von {@link #featureTable}. Wird nur beim Aufruf von
   *  {@link #reorganize} befuellt. */
  protected Feature[] featureArray;
  /** Spaltennamen. Wird nur beim Aufruf von {@link #reorganize} befuellt. */
  protected String[] colNames = null;
  /** Spalten-Typen. Wird nur beim Aufruf von {@link #reorganize} befuellt. */
  protected Class<?>[] colClass = null;
  /** Bestimmt die angezeigten Features */
  protected AttributeTypeFilter attrFilter = AttributeTypeFilter.ALL;

  /** Speichert die angezeigten Attribut-Typen. */
  protected Vector<AttributeType> attrTypes = new Vector<AttributeType>();
  /** Speichert den Attribut-Index fuer jede Spalte (wichtig, wenn nicht
   *  alle Spalten angezeigt werden!) */
  protected int[] attrIdxForCol = null;
  /** Speichert die Indexe der Features im TableModel um die Funktion
   *  {@link #findFeature(Feature)} moeglichst effizient zugestalten. 
   *  Der Key ist nicht das Feature, sondern die FeatureID, weil 
   *  {@link Feature} die equals methode nicht überschreibt.*/
  protected HashMap<String,Integer> featureIdx = null;

  /**
   * Erzeugt ein neues (leeres) Tabellen-Modell.
   */
  public FeatureCollectionTableModel() {
    this(null);
  }

  /**
   * Erzeugt ein neues Tabellen-Modell.
   * @param features Daten fuer die Tabelle
   */
  public FeatureCollectionTableModel(final FeatureCollection features) {
    super();
    setFeatureCollection(features);
  }

  /**
   * Baut die interne Datenbasis (Array) des Tabellen-Modells neu auf.
   * Muss aufgerufen werden, damit nachtraegliche Aenderungen an der
   * zugrunde liegenden {@link FeatureCollection} in das Tabellenmodell
   * uebernommen werden.
   */
  public void reorganize() {
    reorganize(true);
  }
  
  /**
   * Baut die interne Datenbasis (Array) des Tabellen-Modells neu auf.
   * Muss aufgerufen werden, damit nachtraegliche Aenderungen an der
   * zugrunde liegenden {@link FeatureCollection} in das Tabellenmodell
   * uebernommen werden.
   * @param fireTableStructureChanged bestimmt, ob nach dem Neu-Aufbau
   *        ein entsprechendes Event initiiert wird
   */
  protected void reorganize(boolean fireTableStructureChanged) {
    featureArray = FeatureUtil.featuresToArray(featureTable);
    if ( featureArray == null || featureArray.length == 0 ) {
      colNames = new String[0];
      colClass = new Class[0];
    } else {
      // Struktur der Tabelle von erstem Feature uebernehmen
      FeatureType ft = featureArray[0].getFeatureType();
      // Pruefen, welche Attribute angezeigt werden
      attrTypes.clear();
      for (int i=0; i<ft.getAttributeCount(); i++) {
        AttributeType type = ft.getAttributeType(i);
        if ( attrFilter == null || attrFilter.accept( type, i ) )
          attrTypes.add( type );
      }
      // Namen und Attribut-Indizes der angezeigten Spalten ermitteln
      colNames      = new String[ attrTypes.size() ];
      colClass      = new Class[ attrTypes.size() ];
      attrIdxForCol = new int[ attrTypes.size() ];
      for (int i=0; i<colNames.length; i++) {
        int idx = ft.find( attrTypes.elementAt(i) );
        attrIdxForCol[i] = idx;
        colNames[i]      = ft.getAttributeType(idx).getLocalName();
        colClass[i]      = ft.getAttributeType(idx).getBinding();
      }
    }

    // store feature indexes in HashMap to optimize findFeature(.)
    featureIdx = new HashMap<String, Integer>();
    for (int i=0; i<featureArray.length; i++)
      if ( featureArray[i] != null )
        featureIdx.put(featureArray[i].getID(), i);
    
    if ( fireTableStructureChanged )
      fireTableStructureChanged();
  }

  /**
   * Setzt die FeatureCollection fuer das Tabellen-Modell.
   * Ruft {@link #reorganize} auf.<br>
   * Nachtraegliche Aenderungen an der {@link FeatureCollection} werden
   * erst nach erneutem {@link #reorganize()} in das Tabellen-Modell uebernommen!
   * @param features Daten-Basis fuer die Tabelle
   */
  public void setFeatureCollection(final FeatureCollection features) {
      featureTable = features;
      reorganize();
   }

  /**
   * Liefert die FeatureCollection fuer das Tabellen-Modell.
   * Nachtraegliche Aenderungen an der {@link FeatureCollection} werden
   * erst nach erneutem {@link #reorganize()} in das Tabellen-Modell uebernommen!
   */
  public FeatureCollection getFeatureCollection() {
    return featureTable;
  }

  /**
   * Liefert den Filter, der die dargestellten Attribut-Spalten bestimmt.
   */
  public AttributeTypeFilter getAttributeFilter() {
    return attrFilter;
  }

  /**
   * Setzt den Filter, der die dargestellten Attribut-Spalten bestimmt.
   * @param attrFilter Filter
   */
  public void setAttributeFilter(AttributeTypeFilter attrFilter) {
    this.attrFilter = attrFilter;
    // Filter neu anwenden
    reorganize();
  }

  /**
   * Liefert die Anzahl an Spalten der Tabelle. Entspricht der Anzahl an
   * Attributen der Features.
   */
  public int getColumnCount() {
    return colNames == null ? 0 : colNames.length;
  }

  /**
   * Liefert einen Spaltennamen der Tabelle. Entspricht den Attributnamen
   * des (ersten) Features der Collection.
   * @param col Spalten-Index, beginnend bei 0
   */
  public String getColumnName(int col) {
    return colNames == null || col >= colNames.length ? "" : colNames[col];
  }

  /**
   * Liefert die Anzahl an Zeilen der Tabelle. Entspricht der Anzahl an
   * Features in der Collection.
   */
  public int getRowCount() {
    if (featureArray == null)
      featureArray = FeatureUtil.featuresToArray(featureTable);
    return featureArray.length;
  }

  /**
   * Liefert eine Zellen-Wert.
   * @param row Zeilennummer (Feature), beginnend bei 0
   * @param col Spaltennummer (Attribut), beginnen bei 0
   */
  public Object getValueAt(final int row, final int col) {
      if (featureArray == null)
          featureArray = FeatureUtil.featuresToArray(featureTable);
      if (featureArray[row] == null)
        return null;
      return featureArray[row].getAttribute( attrIdxForCol[col] );
  }
  
  /**
   * Liefert die Klasse des Feature-Attributs.
   * @param col Spaltennummer (Attribut), beginnen bei 0
   */
  @Override
  public Class<?> getColumnClass(int col) {
    return colClass[col];
  }

  /**
   * Liefert Features der Tabelle als Array.
   * @param idx Indizes der FeatureCollection
   */
  public Feature[] getFeaturesAsArray(int[] idx) {
    Feature[] features = new Feature[ idx == null ? 0 : idx.length ];
    for (int i=0; i<features.length; i++)
      features[i] = featureArray[ idx[i] ];
    return features;
  }

  /**
   * Liefert Features der Tabelle als Collection.
   * @param idx Indizes der FeatureCollection
   */
  public FeatureCollection getFeaturesAsCollection(int[] idx) {
    FeatureCollection featureCol = FeatureCollections.newCollection();
    for (int i=0; idx != null && i<idx.length; i++)
      featureCol.add( featureArray[ idx[i] ] );
    return featureCol;
  }

  /**
   * Returns a Feature of the model.
   * @param row row index
   */
  public Feature getFeature(int row) {
   return featureArray[row]; 
  }
  
  /**
   * Returns the index of a {@link Feature} in the model.
   * @param f a feature
   * @return -1 if the model does not contain the feature
   */
  public int findFeature(Feature f) {
   return findFeature(f.getID());
  }
  
  /**
   * Returns the index of a {@link Feature} in the model.
   * @param fid A feature-ID
   * @return -1 if the model does not contain the feature
   */
  public int findFeature(String fId) {
   Integer idx = featureIdx.get(fId);
   if ( idx != null )
     return idx;
   return -1; 
  }
  

}
