/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.geotools.gui;

import java.awt.BorderLayout;
import java.awt.FlowLayout;
import java.awt.Frame;
import java.awt.Window;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.IOException;
import java.util.Map;

import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JPanel;

import org.geotools.feature.FeatureCollection;
import org.geotools.filter.Filter;
import org.geotools.map.MapLayer;
import org.geotools.map.event.MapLayerEvent;

import schmitzm.geotools.map.event.FeatureSelectedEvent;
import schmitzm.geotools.map.event.MapLayerAdapter;
import schmitzm.swing.CaptionsChangeable;
import schmitzm.swing.ExceptionDialog;
import schmitzm.swing.SwingUtil;


/**
 * Diese Klasse stellt einen Dialog dar, in dem eine {@link FeatureCollection}
 * ueber eine Formel gefiltert werden kann. Beim Anwenden des Filters wird
 * ein {@link FeatureSelectedEvent} ausgeloest, auf das z.B. mit dem
 * Einfuegen eines neuen Layers reagiert werden kann.
 * @see FeatureCollectionFilterPanel
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class FeatureLayerFilterDialog extends JDialog implements CaptionsChangeable {
  /** Key, um den Titel des FeatureCollection-Filter-Dialogs in der {@link CaptionsChangeable}-Map anzusprechen. */
  public static final String DIALOG_TITLE = FeatureLayerFilterDialog.class.getName()+".TITLE";
  /** Key, um den OK-Button des FeatureCollection-Filter-Dialogs in der {@link CaptionsChangeable}-Map anzusprechen. */
  public static final String OK_BUTTON = FeatureLayerFilterDialog.class.getName()+".OK";
  /** Key, um den ABBRECHEN-Button des FeatureCollection-Filter-Dialogs in der {@link CaptionsChangeable}-Map anzusprechen. */
  public static final String CANCEL_BUTTON = FeatureLayerFilterDialog.class.getName()+".CANCEL";
  /** Key, um den ÜBERNEHMEN-Button des FeatureCollection-Filter-Dialogs in der {@link CaptionsChangeable}-Map anzusprechen. */
  public static final String APPLY_BUTTON = FeatureLayerFilterDialog.class.getName()+".APPLY";

  /** Panel in dem der Filter definiert und getestet werden kann. */
  protected FeatureCollectionFilterPanel filterPanel = null;
  /** Der OK-Button. */
  protected JButton okButton = null;
  /** Der ABBRECHEN-Button. */
  protected JButton cancelButton = null;
  /** Der ANWENDEN-Button. */
  protected JButton applyButton = null;

  private String   frameTitle = GeotoolsGUIUtil.RESOURCE.getString(DIALOG_TITLE);
  private JMapPane mapPane = null;
  private MapLayer layer = null;

  /**
   * Erzeugt einen neuen Dialog.
   * @param parent Uebergeordnetes Fenster (kann {@code null} sein)
   * @param mapPane MapPane fuer welches {@link FeatureSelectedEvent} ausgeloest
   *                werden
   * @param mapLayer MapLayer, aus dem die FeatureCollection stammt, auf der der
   *                 Filter definiert wird
   * @exception IOException falls beim Ermitteln der {@link FeatureCollection} aus
   *            dem Layer ein Fehler auftritt
   *           
   * CHANGE BY SK: Takes {@link Window} instead of {@link Frame}
   * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons Kr&uuml;ger</a>
   */
  public FeatureLayerFilterDialog(Window parent, JMapPane mapPane, MapLayer mapLayer) throws IOException {
    this(parent, mapPane, mapLayer, true);
  }


  /**
   * Erzeugt einen neuen Dialog.
   * @param parent Uebergeordnetes Fenster (kann {@code null} sein)
   * @param mapPane MapPane fuer welches {@link FeatureSelectedEvent} ausgeloest
   *                werden
   * @param mapLayer MapLayer, aus dem die FeatureCollection stammt, auf der der
   *                 Filter definiert wird
   * @param initGUI bestimmt, ob {@link #initGUI()} aufgerufen werden soll. Wenn
   *                {@code false}, muss der Konstruktor der Unterklasse dafuer
   *                sorgen, dass {@link #initGUI()} aufgerufen wird
   * @exception IOException falls beim Ermitteln der {@link FeatureCollection} aus
   *            dem Layer ein Fehler auftritt
   *            
   * CHANGE BY SK: Takes {@link Window} instead of {@link Frame}     
   * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons Kr&uuml;ger</a>       
   */
  protected FeatureLayerFilterDialog(Window parent, JMapPane mapPane, MapLayer mapLayer, boolean initGUI) throws IOException {
    super(parent);
    setModal(true);
    this.setTitle( frameTitle + " ["+mapLayer.getTitle()+"]" );
    this.setLayout( new BorderLayout() );
    this.setSize(380,200);
    this.mapPane = mapPane;
    this.layer   = mapLayer;
    if ( initGUI ) {
      initGUI();
      pack();
    }
  }



  /**
   * Initalisiert die GUI.
   */
  protected void initGUI() throws IOException {
    filterPanel = new FeatureCollectionFilterPanel( layer.getFeatureSource().getFeatures(), true ) {
      @Override
      protected void resetComponentsAfterTest(Throwable err) {
        super.resetComponentsAfterTest(err);
        // OK- und Uebernehmen-Button (de)aktivieren
        okButton.setEnabled( err == null );
        applyButton.setEnabled( err == null );
      }
      @Override
      protected boolean acceptOperator(String operator) {
        // Operatoren durch Methode des Dialogs filtern
        return super.acceptOperator(operator) &&
               FeatureLayerFilterDialog.this.acceptOperator(operator);
      }
    };
    okButton = new JButton( SwingUtil.RESOURCE.getString("Ok") );
    cancelButton = new JButton( SwingUtil.RESOURCE.getString("Cancel") );
    applyButton = new JButton( SwingUtil.RESOURCE.getString("Apply") );
    add( filterPanel, BorderLayout.CENTER);
    JPanel buttonPanel = new JPanel( new FlowLayout(FlowLayout.CENTER, 10, 5) );
    buttonPanel.add( okButton);
    buttonPanel.add( cancelButton);
    buttonPanel.add( applyButton);
    add( buttonPanel, BorderLayout.SOUTH );

    // Aktionen der Filter-Fenster-Button
    ActionListener action = new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        // Ok- und Apply-Button > Filter anwenden und Event erzeugen
        if ( e.getSource() == okButton || e.getSource() == applyButton ) {
          // FeatureCollection filtern
          try {
            FeatureCollection fc = filterPanel.filterFeatureCollection();
            FeatureSelectedEvent fse = new FeatureSelectedEvent(getMapPane() , layer, fc.getBounds(), fc, FeatureLayerFilterDialog.this);
            getMapPane().fireMapPaneEvent( fse );
          } catch ( Exception err ) {
            ExceptionDialog.show(filterPanel,err);
            // Fenster nicht schliessen, bei OK!!
            return;
          }
        }
        // Ok- und Cancel-Button > Dialog schliessen
        if ( e.getSource() == okButton || e.getSource() == cancelButton ) {
          setVisible( false );
        }
      }
    };
    okButton.addActionListener( action );
    cancelButton.addActionListener( action );
    applyButton.addActionListener( action );

    // Wenn sich die die Bezeichnung des Layers aendert, soll auch der
    // Titel des Filter-Fensters geaendert werden
    layer.addMapLayerListener( new MapLayerAdapter() {
      public void layerChanged(MapLayerEvent e) {
        // Bezeichnung der Checkbox aendern
        setTitle(frameTitle + " ["+layer.getTitle()+"]");
      }
    });
  }

  /**
   * Kann von Sub-Klassen ueberschrieben werden, um bestimmte Operatoren
   * in der Auswahl-Liste auszublenden.
   * @param operator ein Operator
   * @return immer {@code true}
   */
  protected boolean acceptOperator(String operator) {
    return true;
  }

  /**
   * Liefert das MapPane, fuer das {@link FeatureSelectedEvent FeatureSelectedEvents}
   * ausgeloest werden.
   */
  public JMapPane getMapPane() {
    return mapPane;
  }

  /**
   * Liefert das Layer, aus dem die {@link FeatureCollection} stammt auf der
   * der Filter definiert wird.
   */
  public MapLayer getMapLayer() {
    return layer;
  }

  /**
   * Liefert das Panel, in dem der Filter definiert und die Vorschau angezeigt
   * wird.
   */
  public FeatureCollectionFilterPanel getFeatureCollectionFilterPanel() {
    return filterPanel;
  }

  /**
   * Liefert die aktuell im Dialog eingegebene Formel.
   */
  public String getFilterRule() {
    return filterPanel.getRule();
  }

  /**
   * Setzt die im Dialog eingegebene Formel und aktualisiert entsprechend die
   * Vorschau
   * @param rule Formel als String
   */
  public void setFilterRule(String rule) {
    filterPanel.setRule(rule);
  }

  /**
   * Liefert den Filter der aktuell im Dialog eingegebenen Formel.
   */
  public Filter getFilter() {
    return filterPanel.createFilter();
  }

  /**
   * Belegt die Labels und Buttons im Dialog neu
   * @param captionMap Map in der die neuen Labels hinterlegt sind.
   * @see FeatureLayerFilterDialog
   */
  public void resetCaptions(Map<String,Object> captionMap) {
    SwingUtil.resetCaption ( okButton, captionMap.get(OK_BUTTON) );
    SwingUtil.resetCaption ( applyButton, captionMap.get(APPLY_BUTTON) );
    SwingUtil.resetCaption ( cancelButton, captionMap.get(CANCEL_BUTTON) );
    Object caption = captionMap.get( DIALOG_TITLE );
    if ( caption != null ) {
      frameTitle = caption.toString();
      // Damit Layer-Bezeichnung in Titel uebernommen wird, ein MapLayerEvent
      // ausloesen
      layer.setTitle( layer.getTitle() );
    }
    filterPanel.resetCaptions( captionMap );
  }
}
