/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.geotools.gui;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.util.HashMap;

import javax.swing.BorderFactory;

import org.geotools.map.MapContext;
import org.geotools.renderer.GTRenderer;
import org.geotools.renderer.lite.StreamingRenderer;
import org.geotools.renderer.shape.ShapefileRenderer;

import schmitzm.geotools.map.event.JMapPaneEvent;
import schmitzm.geotools.map.event.JMapPaneListener;
import schmitzm.geotools.map.event.MapAreaChangedEvent;
import schmitzm.geotools.map.event.ScaleChangedEvent;
import schmitzm.swing.JPanel;
import schmitzm.swing.SwingUtil;

/**
 * Das {@code GeoMapPane} erweitert das {@link JMapPane} um einen
 * Massstab-Balken, sowie ein horizontales und vertikales Koordinaten-Raster
 * (Grid), in dem die Geo-Referenz des angezeigten Karten-Bereichs in
 * Welt-Koordinaten (Grad, Minuten) dargestellt ist.
 * 
 * @see ScalePane
 * @see GridPanel
 * @author Martin Schmitz
 * @version 1.0
 */
public class GeoMapPane extends JPanel {


	/**
	 * Konstante fuer das vertikale Koordinaten-Grid.
	 * 
	 * @see #layoutConstraints
	 */
	public static final String DLC_VGRID = "vertGrid";
	/**
	 * Konstante fuer das horizontale Koordinaten-Grid.
	 * 
	 * @see #layoutConstraints
	 */
	public static final String DLC_HGRID = "horGrid";
	/**
	 * Konstante fuer die Karte.
	 * 
	 * @see #layoutConstraints
	 */
	public static final String DLC_MAP = "mapPane";
	/**
	 * Konstante fuer den Massstab-Balken.
	 * 
	 * @see #layoutConstraints
	 */
	public static final String DLC_SCALE = "scalePane";
	/**
	 * Werte fuer die grafische Anordnung der Layout-Komponenten. Ueber die
	 * Konstanten {@link #DLC_MAP}, {@link #DLC_VGRID}, {@link #DLC_HGRID} und
	 * {@link #DLC_SCALE} koennen die Constraints (am Anfang von
	 * {@link #guiInit()}) veraendert oder erweitert werden.
	 */
	protected HashMap<String, GridBagConstraints> layoutConstraints = new HashMap<String, GridBagConstraints>();

	/** Karten-Bereich des {@code GeoMapPane}. */
	protected JMapPane mapPane = null;
	/** Massstab-Balken */
	protected ScalePane scalePane = null;
	/**
	 * Vertikale Koordinaten-Leiste (Grid). Zeigt die Latitude-Koordinate in
	 * Grad und Minutenm an.
	 */
	protected GridPanel vertGrid = null;
	/**
	 * Horizontale Koordinaten-Leiste (Grid). Zeigt die Longitude-Koordinate in
	 * Grad und Minutenm an.
	 */
	protected GridPanel horGrid = null;

	/**
	 * A flag indicating if dispose() was already called. If true, then further
	 * use of this {@link GeoMapPane} is undefined.
	 */
	private boolean disposed = false;

	/**
	 * Erzeugt ein neues {@code GeoMapPane}.
	 */
	public GeoMapPane() {
		this(null, null, null, null, null);
	}

	/**
	 * Erzeugt ein neues {@code GeoMapPane}. Dieser Konstruktor bietet die
	 * Moeglichkeit, alternative Implementierungen der einzelnen Komponenten
	 * darzustellen. <b>Die uebergebenen Parameter koennen {@code null} sein! In
	 * diesem Fall wird die entsprechende Standard-Implementierung
	 * verwendet.</b>
	 * 
	 * @param mapPane
	 *            Karten-Bereich
	 * @param vGrid
	 *            vertikales Koordinaten-Raster
	 * @param hGrid
	 *            horizontales Koordinaten-Raster
	 * @param scalePane
	 *            Massstab-Balken
	 * @param renderer
	 *            Allows you to choose between the {@link StreamingRenderer}
	 *            which is cool for creating PDF, SVG.. And the much faster
	 *            {@link ShapefileRenderer}.
	 */
	public GeoMapPane(JMapPane mapPane, GridPanel vGrid, GridPanel hGrid,
			ScalePane scalePane, GTRenderer renderer) {
		super();

		if (vGrid != null && !vGrid.isVertical())
			throw new IllegalArgumentException(
					"GridPanel for vertical grid must be of type GridPanel.VERTICAL!!");
		if (hGrid != null && !hGrid.isHorizontal())
			throw new IllegalArgumentException(
					"GridPanel for horizontal grid must be of type GridPanel.HORIZONTAL!!");

		// Karte
		this.mapPane = (mapPane != null) ? mapPane : new JMapPane(null,null,renderer, null);
		// Koordinaten-Leisten
		this.vertGrid = (vGrid != null) ? vGrid : new GridPanel(
				GridPanel.VERTICAL, this.mapPane);
		this.horGrid = (hGrid != null) ? hGrid : new GridPanel(
				GridPanel.HORIZONTAL, this.mapPane);
		// Massstab-Balken
		this.scalePane = (scalePane != null) ? scalePane : new ScalePane();

		// Standard-Werte fuer GUI-Anordnung setzen
		layoutConstraints.put(DLC_MAP, new GridBagConstraints(1, 0, 1, 1, 1.0,
				1.0, GridBagConstraints.CENTER, GridBagConstraints.BOTH,
				new Insets(0, 0, 0, 0), 0, 0));
		layoutConstraints.put(DLC_VGRID, new GridBagConstraints(0, 0, 1, 1,
				0.0, 1.0, GridBagConstraints.CENTER, GridBagConstraints.BOTH,
				new Insets(0, 0, 0, 0), 0, 0));
		layoutConstraints.put(DLC_HGRID, new GridBagConstraints(1, 1, 1, 1,
				1.0, 0.0, GridBagConstraints.CENTER, GridBagConstraints.BOTH,
				new Insets(0, 0, 0, 0), 0, 0));
		layoutConstraints.put(DLC_SCALE, new GridBagConstraints(1, 2, 1, 1,
				0.0, 0.0, GridBagConstraints.WEST, GridBagConstraints.BOTH,
				new Insets(2, 5, 2, 5), 0, 0));
		// GUI initialisieren
		guiInit();
	}

	/**
	 * Wird vom Konstruktor aufgerufen und initialisiert die grafische
	 * Darstellung/Anordnung der einzelnen GUI-Komponenten in einem
	 * {@link GridBagLayout}. Die Constraints fuer die Anordnung der Komponenten
	 * sind in {@link #layoutConstraints} hinterlegt und koennen zu beginn
	 * dieser Methode ueberschrieben oder erweitert werden.
	 */
	protected void guiInit() {
		this.setLayout(new GridBagLayout());

		// Karten-Darestellung initialisieren
		this.mapPane.setBorder(BorderFactory.createLoweredBevelBorder());
		SwingUtil.setPreferredWidth(this.mapPane, 200);

		// MapListener that listens to Scale and MapArea changes
		this.mapPane.addMapPaneListener(new JMapPaneListener() {
			public void performMapPaneEvent(JMapPaneEvent e) {
				if (e instanceof ScaleChangedEvent) {
					ScaleChangedEvent sce = (ScaleChangedEvent) e;
					scalePane.setScale(sce.getNewScale());
				}
				if (e instanceof MapAreaChangedEvent) {
					vertGrid.repaint();
					horGrid.repaint();
				}
			}
		});

		// Komponenten in Layout anordnen
		this.add(vertGrid, layoutConstraints.get(DLC_VGRID));
		this.add(horGrid, layoutConstraints.get(DLC_HGRID));
		this.add(scalePane, layoutConstraints.get(DLC_SCALE));
		// WICHTIG: mapPane als LETZTES einfuegen, damit es beim Re-Paint ALS
		// ERSTES
		// aktualisiert wird und die Aenderungen (Area, Scale) im
		// Update des Massstab-Balken und der der Koordinaten-Leiste
		// bereits zur Verfuegung stehen!
		this.add(mapPane, layoutConstraints.get(DLC_MAP));

		// // Hintergrundfarbe aller enthaltenen Komponenten auf Weiss setzen
		// setBackground(BGCOLOR, true);
	}

	/**
	 * Aktualisiert die Karten-Darstellung.
	 * 
	 * @see JMapPane#refresh()
	 */
	public void refreshMap() {
		mapPane.refresh();
	}

	/**
	 * Nach dem {@code super}-Aufruf, wird der Massstab neu gesetzt (und somit
	 * neu angezeigt), falls sich der Massstab der Karte geaendert hat. Dies ist
	 * ein Workaround, damit der Massstab auch beim allerersten anzeigen korrekt
	 * angezeigt wird (ohne {@link ScaleChangedEvent}).
	 */
	public void paint(Graphics g) {
		super.paint(g);
		if (scalePane.getScaleInMeters() != mapPane.getScale()
				&& mapPane.getScale() > 0)
			scalePane.setScale(mapPane.getScale());
	}

	/**
	 * Liefert das {@link JMapPane}, in dem die Karten dargestellt werden.
	 */
	public final JMapPane getMapPane() {
		return mapPane;
	}

	/**
	 * Liefert den {@link MapContext} der die einzelnen Karten-Layer verwaltet.
	 */
	public final MapContext getMapContext() {
		return mapPane.getContext();
	}

	/**
	 * Setzt die Hintergrundfarbe des kompletten {@link GeoMapPane}.
	 * 
	 * @param color
	 *            Hintergrundfarbe
	 */
	public void setBackground(Color color) {
		setBackground(color, true);
	}

	/**
	 * @return If dispose() has been called. If
	 *         <code>true<(code>, further use of this {@link GeoMapPane} is undefined
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public boolean isDisposed() {
		return disposed;
	}

	/**
	 * Should be called when the {@link GeoMapPane} is not needed no more to
	 * help the GarbageCollector
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public void dispose() {
		if (isDisposed())
			return;
		mapPane.dispose();
		removeAll();
		disposed = true;
	}
	
	/**
	 * Nuetzlich wenn die Componente gedruckt (z.B. wenn ein Screenshot gemacht
	 * wird) wird. Dann werden wird der Hintergrund auf WEISS gesetzt.
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	@Override
	public void print(Graphics g) {
		final Color orig = getBackground();
		setBackground(Color.WHITE);
		// wrap in try/finally so that we always restore the state
		try {
			super.print(g);
		} finally {
			setBackground(orig);
		}
	}

	// /**
	// * Trying to reproduce a bug #37
	// *
	// * for explanation and screenshots see the xulu trac
	// *
	// * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	// Kr&uuml;ger</a>
	// */
	// public static void main(String[] args) throws Exception {
	// JFrame f = new JFrame();
	// f.pack();
	// f.setDefaultCloseOperation( JFrame.EXIT_ON_CLOSE );
	// GeoMapPane gmp = new GeoMapPane();
	// // FeatureCollection shape = GeoImportUtil.readFeaturesFromShapeFile( new
	// File("/home/stefan/Desktop/AtlasData/benin_karte_UTM31/admin_boundary_benin_utm.shp")
	// );
	// FeatureCollection shape = GeoImportUtil.readFeaturesFromShapeFile( new
	// File("F:\\Arbeit\\Impetus - ZFL\\Daten\\Neuer Ordner\\admin_boundary_benin_utm.shp")
	// );
	// gmp.getMapContext().addLayer( shape,
	// FeatureUtil.createDefaultStyle(shape) );
	// f.add( gmp);
	// gmp.getMapPane().setPreferredSize( new Dimension(400,400));
	// gmp.refreshMap();
	// f.pack();
	// f.setVisible(true);
	// }

}
