/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.geotools.gui;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.geom.AffineTransform;
import java.text.DecimalFormat;

import org.apache.log4j.Logger;
import org.geotools.referencing.CRS;
import org.geotools.referencing.operation.projection.PointOutsideEnvelopeException;
import org.opengis.referencing.FactoryException;
import org.opengis.referencing.crs.CoordinateReferenceSystem;
import org.opengis.referencing.operation.MathTransform;

import schmitzm.lang.LangUtil;
import schmitzm.lang.ResourceProvider;
import schmitzm.swing.JPanel;
import schmitzm.swing.SwingUtil;

/**
 * Diese Klasse stellt eine horizontale oder vertikale Koordinaten-Leiste (Grid)
 * dar, die an ein {@link JMapPane} gekoppelt ist.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class GridPanel extends JPanel {
  private final Logger LOGGER = LangUtil.createLogger(this);
  private static final ResourceProvider RES = GeotoolsGUIUtil.RESOURCE;

  private static final DecimalFormat numFormat = new DecimalFormat("###,###,##0");

  /** Laenge der Grid-Abschnitts-Linien.  */
  private static final int GRID_LINE_SIZE = 10;
  /** Container-Breite bei vertikalem Grid. */
  public static final int VERT_WIDTH     = 50;
  /** Container-Hoehe bei horizontalem Grid. */
  public static final int HOR_HEIGHT     = 20;

  /** Flag fuer ein horizontales Grid. */
  public static final int HORIZONTAL = 1;
  /** Flag fuer ein vertikales Grid. */
  public static final int VERTICAL = 2;

  /** Speichert die Orientierung des Grid (Horizontal oder Vertikal). */
  private int orientation = 0;

  /** Karte, an der die Koordinaten-Leiste ausgerichtet wird. */
  protected JMapPane mapPane = null;

  /** Transformation von Karten-CRS zu CRS des Koordinaten-Rasters */
  protected MathTransform mapToGrid = null;
  /** Transformation von CRS des Koordinaten-Rasters zu Karten-CRS*/
  protected MathTransform gridToMap = null;
  /** Speichert das CRS, fuer das die Transformationen erstellt wurden. */
  private CoordinateReferenceSystem transformedMapCRS = null;

  /**
   * Erzeugt eine Koordinaten-Leiste fuer ein {@link JMapPane}.
   * @param orientation Orientierung ({@link #HORIZONTAL} oder {@link #VERTICAL})
   * @param mapPane     Karte an der die Koordinaten-Leiste ausgerichtet wird
   */
  public GridPanel(int orientation, JMapPane mapPane) {
    super();

    this.orientation = orientation;
    this.mapPane     = mapPane;
    if ( isVertical() )
      SwingUtil.setPreferredWidth(this,VERT_WIDTH);
    else
      SwingUtil.setPreferredHeight(this,HOR_HEIGHT);
  }

  /**
   * Liefert die Orientierung der Koordinaten-Leiste.
   * @see #HORIZONTAL
   * @see #VERTICAL
   */
  public int getOrientation() {
    return orientation;
  }

  /**
   * Prueft, ob es sich um eine horizontale Koordinaten-Leiste handelt.
   */
  public boolean isHorizontal() {
    return this.orientation == HORIZONTAL;
  }

  /**
   * Prueft, ob es sich um eine vertikale Koordinaten-Leiste handelt.
   */
  public boolean isVertical() {
    return this.orientation == VERTICAL;
  }

  /**
   * Zeichnet die Koordinaten-Leiste.
   * @param g Graphics
   */
  public void paint(Graphics g) {
    super.paint(g);
    if ( getParent() == null || mapPane.getMapArea() == null )
      return;

    CoordinateReferenceSystem mapCRS = null;   // (aktuelles) CRS des MapPane
    CoordinateReferenceSystem gridCRS = null;  // CRS fuer die Grad/Min/Sek-Anzeige
//    try {
//      final CoordinateReferenceSystem UTM   = CRS.decode("EPSG:32631"); // utm zone 31n
//      final CoordinateReferenceSystem WGS84 = CRS.decode("EPSG:4326"); // lat/lon (WGS84)
////      final CoordinateReferenceSystem WGS84 = org.geotools.referencing.crs.DefaultGeographicCRS.WGS84;// lat/lon
////      final CoordinateReferenceSystem NAD83 = CRS.decode("EPSG:100002"); // lat/lon
////      mapCRS  = mapPane.getContext().getCoordinateReferenceSystem();
//      mapCRS = UTM;
//      gridCRS = WGS84;
//    } catch (Exception err) {
//      return;
//    }


    mapCRS  = mapPane.getContext().getCoordinateReferenceSystem();
    gridCRS = org.geotools.referencing.crs.DefaultGeographicCRS.WGS84;// lat/lon
//    LOGGER.debug("mapCRS: "+mapCRS);
//    LOGGER.debug("gridCRS: "+gridCRS);

    double     minX   = mapPane.getMapArea().getMinX();
    double     minY   = mapPane.getMapArea().getMinY();
    double     maxX   = mapPane.getMapArea().getMaxX();
    double     maxY   = mapPane.getMapArea().getMaxY();

// commented out, because it was unused
//    double     width  = mapPane.getMapArea().getWidth();
//    double     height = mapPane.getMapArea().getHeight();

    // Minimum/Maximum des Grids in MapPane-CRS (MCRS)
    double[] mapMin_MCRS = new double[] {minX,minY};
    double[] mapMax_MCRS = isVertical() ? new double[] {minX,maxY} : new double[] {maxX,minY};
    // Minimum/Maximum des Grids in Grid-CRS (GCRS)
    double[] mapMin_GCRS = new double[2];
    double[] mapMax_GCRS = new double[2];

    try {
      try {
        // Der Effizienz halber, die Transformation nur neu ermitteln
        // wenn sich das Karten-CRS geaendert hat
        if ( mapToGrid == null || gridToMap == null || transformedMapCRS == null || !transformedMapCRS.equals(mapCRS) ) {
          mapToGrid = CRS.findMathTransform(mapCRS, gridCRS, true);
          gridToMap = CRS.findMathTransform(gridCRS, mapCRS, true);
          // CRS merken, fuer das Transformation erstellt wurde
          transformedMapCRS = mapCRS;
        }
      } catch (FactoryException  err) {
        LOGGER.warn("CRS-Transformation to WGS84 failed: "+err.getMessage());
        mapToGrid = CRS.findMathTransform(mapCRS, gridCRS, true);
        gridToMap = CRS.findMathTransform(gridCRS, mapCRS, true);
        LOGGER.warn("CRS-Transformation without datum shift");
        // CRS merken, fuer das Transformation erstellt wurde
        transformedMapCRS = mapCRS;
      }
      mapToGrid.transform( mapMin_MCRS, 0, mapMin_GCRS, 0, 1  );
      mapToGrid.transform( mapMax_MCRS, 0, mapMax_GCRS, 0, 1  );
    } catch (PointOutsideEnvelopeException err) {
      return;
    } catch (Exception err) {
      LOGGER.error("CRS-Transformation to WGS84 failed: "+err.getMessage());
      LOGGER.debug("CRS-Transformation to WGS84 failed: "+err.getMessage(),err);
    }

//   // Ergebnis von Transform: {Latitude,Longitude}
//   int idx = isVertical() ? 0 : 1;
   // SEHR SEHR MERKWUERDIG: Mal liefert Transform {Latitude,Longitude}, und
   //                        manchmal {Longitude,Latitude}!!
   // WORKAROUND (klappt natuerlich nicht mit allen CRS!):
   //       Wenn  X_MCRS > Y_MCRS, dann muss auch X_GCRS > Y_GCRS
  int idx = determineCompatibleArrayIndex(mapMin_MCRS,mapMin_GCRS,mapMax_MCRS,mapMax_GCRS);
//   LOGGER.debug( isVertical() ? "V" : "H" );
//   LOGGER.debug( "Min MCRS: "+mapMin_MCRS[0]+", "+mapMin_MCRS[1]+"         GCRS:"+mapMin_GCRS[0]+", "+mapMin_GCRS[1]);
//   LOGGER.debug( "Max MCRS: "+mapMax_MCRS[0]+", "+mapMax_MCRS[1]+"         GCRS:"+mapMax_GCRS[0]+", "+mapMax_GCRS[1]);
//   LOGGER.debug( "  --> "+idx+"\n");

   double mapMin = mapMin_GCRS[idx];
   double mapMax = mapMax_GCRS[idx];
   // Maximale Groesse des Grid in Pixel
   int maxSize = isHorizontal() ? getParent().getWidth() : getParent().getHeight();

   // Passende Grid-Unterteilung in Grad
   double gridDist = determineGridDistance(mapMin,mapMax,maxSize);
   // Leiste zeichnen
   if ( isHorizontal() )
     g.drawLine(0,0,maxSize,0);
   else
     g.drawLine(VERT_WIDTH-1,0,VERT_WIDTH-1,maxSize);

   // Abschnitte zeichnen
   double   firstGrid = Math.ceil(mapMin/gridDist)*gridDist; // 1. Abschnitt in Lat bzw. Lon
   double[] grid_GCRS = mapMin_GCRS.clone(); // Ausgangspunkt
   double[] grid_MCRS = new double[2];
   AffineTransform mapToWin = mapPane.getTransform();
   for (double gridPos=firstGrid; gridPos<=mapMax; gridPos+=gridDist) {
     grid_GCRS[idx] = gridPos;
     // In Map-CRS und dann in Fenster-Koordinaten transformieren
     double[] grid_WINDOW = new double[2];
     try {
       gridToMap.transform(grid_GCRS, 0, grid_MCRS, 0, 1);
       mapToWin.inverseTransform(grid_MCRS,0,grid_WINDOW,0,1);
//     } catch (Exception err) {
     } catch (Throwable err) {
       LOGGER.warn("Error ignored during paint of coordinate grid: "+err);
       continue;
     }
     // Grid-Abschnitt zeichnen
     int x1 = isVertical() ? VERT_WIDTH-GRID_LINE_SIZE : (int)grid_WINDOW[0];
     int y1 = isVertical() ? (int)grid_WINDOW[1] : 0;
     int x2 = isVertical() ? x1+GRID_LINE_SIZE : x1;
     int y2 = isVertical() ? y1 : y1+GRID_LINE_SIZE;
     g.drawLine(x1,y1,x2,y2);

     // Koordinate anzeigen
     double coord = grid_GCRS[idx];
     int degrees = (int)coord;
     int minutes = (int)Math.round((coord-degrees)*60 );
     // Aufgrund des Minuten-Systems kann schlecht im Vorhinein gerundet
     // werden. Deshalb wird im Nachhinein ein etwaiger Rundungsfehler
     // ausgeglichen
     if ( Math.abs(minutes) >= 60 ) {
       degrees += Math.signum(minutes);
       minutes -= Math.signum(minutes) * 60;
     }
     String coordStr = Math.abs(degrees)+"°"+Math.abs(minutes);
     if ( degrees != 0 || minutes != 0 ) {
       if (isVertical())
         coordStr += RES.getString( (coord < 0) ? "schmitzm.geotools.gui.GeotoolsGUIUtil.SOUTH.Abb" : "schmitzm.geotools.gui.GeotoolsGUIUtil.NORTH.Abb" );
       else
         coordStr += RES.getString( (coord < 0) ? "schmitzm.geotools.gui.GeotoolsGUIUtil.WEST.Abb" : "schmitzm.geotools.gui.GeotoolsGUIUtil.EAST.Abb" );
     }
     x1 = isVertical() ?    5 : x1 + 5;
     y1 = isVertical() ? y1-2 : y1 + 12;
     g.drawString( coordStr, x1, y1 );
   }

/////////////   Test und Debugging /////////////////
//    System.out.println("2 UTM lat="+minY+" lon="+minX+"       --> Degree lat="+dest2[0]+" lon="+dest2[1]);

//    org.geotools.referencing.operation.DefaultMathTransformFactory.main(new String[] {});
//    if ( debug == 0) {
//      schmitzm.lang.WorkingThread w = new schmitzm.lang.WorkingThread() {
//        java.util.Set codes = null;
//        public void performInit() {
//          codes = CRS.getSupportedCodes("EPSG");
//        }
//        public void performWork() {
//          for (Object code : codes) {
//            String crsCode = ((String)code).startsWith("EPSG") ? (String)code : "EPSG:"+((String)code);
//            try {
//              CoordinateReferenceSystem crs = CRS.decode(crsCode);
//              System.out.println(code+"\t"+crsCode+"\t"+crs.getName());
//            } catch (Exception e) {
//              System.out.println(code+"\t"+crsCode+"\tError");
//              break;
//            }
//          }
//        }
//        public void performDispose() {
//        }
//      };
//      debug = 1;
//      w.start();
//    }
  }

  /**
   * Bestimmt eine "passende" Unterteilung fuer das Grid.
   */
  private static double determineGridDistance(double mapMin_LL, double mapMax_LL, int panelSize) {
    // Differenz zwischen Min/Max-Koordinaten = darzustellender Bereich in Grad
    double diff = Math.abs( mapMin_LL - mapMax_LL );
    // Distanz zwischen zwei Unterteilungen in Grad
    double dist = 0.0;
    if ( diff < 10/60.0 )      // < 10 Min. -->  1 Min.-Aufteilung
      dist = 1/60.0;
    else if ( diff < 30/60.0 ) // < 30 Min. -->  5 Min.-Aufteilung
      dist = 5/60.0;
    else if ( diff <  1 )      // <  1?     --> 10 Min.-Aufteilung
      dist = 10/60.0;
    else if ( diff <  3 )      // <  3?     --> 15 Min.-Aufteilung
      dist = 15/60.0;
    else if ( diff <  5 )      // <  5?     --> 30 Min.-Aufteilung
      dist = 30/60.0;
    else if ( diff < 11 )      // < 11?     -->  1?-Aufteilung
      dist = 1;
    else                       // sonst     -->  2?-Aufteilung
      dist = 2;

    // Solange der Abstand zwischen 2 Unterteilungen weniger als 40 Pixel
    // betraegt --> Unterteilung verdoppeln
    while ( diff != 0 && dist != 0 && panelSize / (diff / dist) < 40 )
      dist *=2;

    return dist;
  }

//  /**
//   * Diese Methode stellt einen Workaround fuer einen sehr sehr merkwuerdigen
//   * {@link MathTransform}-Effekt dar. Mal liefert Transform die Koordinaten
//   * als {Latitude,Longitude}-Array, und manchmal als {Longitude,Latitude}!
//   * Diese Methode versucht zu ermitteln, in welchem Array-Index die
//   * benoetigte Koordinate (je nachdem ob es sich um ein horizontales oder
//   * vertikales Grid handelt) tatsaechlich befindet.
//   * @param coord_SCRS Koordinate (in Quell-CRS)
//   * @param coord_DCRS transformierte Koordinate (in Ziel-CRS)
//   */
//  private int determineCompatibleArrayIndex(double[] coord_SCRS, double[] coord_DCRS) {
//    // SEHR SEHR MERKWUERDIG: Mal liefert Transform {Latitude,Longitude}, und
//    //                        manchmal {Longitude,Latitude}!!
//    // WORKAROUND (klappt natuerlich nicht mit allen CRS!):
//    //       Wenn  X_sourceCRS > Y_sourceCRS, dann muss auch X_destCRS > Y_destCRS
//    int idx = 0;
//    if (coord_SCRS[0] > coord_SCRS[1] && coord_DCRS[0] > coord_DCRS[1] ||
//        coord_SCRS[0] < coord_SCRS[1] && coord_DCRS[0] < coord_DCRS[1] ||
//        coord_SCRS[0] == coord_SCRS[1] && coord_DCRS[0] == coord_DCRS[1])
//      idx = isHorizontal() ? 0 : 1;
//    else
//      idx = isHorizontal() ? 1 : 0;
//    return idx;
//  }

  /**
   * Diese Methode stellt einen Workaround fuer einen sehr sehr merkwuerdigen
   * {@link MathTransform}-Effekt dar. Mal liefert Transform die Koordinaten
   * als {Latitude,Longitude}-Array, und manchmal als {Longitude,Latitude}!
   * Diese Methode versucht zu ermitteln, in welchem Array-Index die
   * benoetigte Koordinate (je nachdem ob es sich um ein horizontales oder
   * vertikales Grid handelt) tatsaechlich befindet.
   */
  private int determineCompatibleArrayIndex(double[] mapMin_MCRS, double[] mapMin_GCRS,double[] mapMax_MCRS, double[] mapMax_GCRS) {
    // SEHR SEHR MERKWUERDIG: Mal liefert Transform {Latitude,Longitude}, und
    //                        manchmal {Longitude,Latitude}!!
    // WORKAROUND (klappt natuerlich nicht mit allen CRS!):
    //       Wenn  X_sourceCRS > Y_sourceCRS, dann muss auch X_destCRS > Y_destCRS


    double diff0_GCRS = mapMax_GCRS[0] - mapMin_GCRS[0];
    double diff1_GCRS = mapMax_GCRS[1] - mapMin_GCRS[1];

    return diff0_GCRS > diff1_GCRS ? 0 : 1;
  }
  
  /**
   * Nuetzlich wenn die Componente gedruckt (z.B. wenn ein Screenshot gemacht wird) wird. Dann werden wird der
   * Hintergrund auf WEISS gesetzt.
   *
   * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
   *         Kr&uuml;ger</a>
   */
  @Override
  public void print(Graphics g) {
      Color orig = getBackground();
      setBackground(Color.WHITE);

      // wrap in try/finally so that we always restore the state
      try {
          super.print(g);
      } finally {
          setBackground(orig);
      }
  }

}
