/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.geotools.gui;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.LayoutManager;
import java.awt.Rectangle;
import java.awt.event.MouseEvent;
import java.awt.geom.Point2D;
import java.util.HashMap;
import java.util.Map;
import java.util.Stack;

import org.geotools.feature.AttributeType;
import org.geotools.feature.DefaultFeatureCollection;
import org.geotools.feature.DefaultFeatureCollections;
import org.geotools.feature.Feature;
import org.geotools.feature.FeatureCollection;
import org.geotools.feature.FeatureType;
import org.geotools.feature.FeatureTypeBuilder;
import org.geotools.feature.GeometryAttributeType;
import org.geotools.feature.SchemaException;
import org.geotools.feature.type.GeometricAttributeType;
import org.geotools.map.DefaultMapContext;
import org.geotools.map.DefaultMapLayer;
import org.geotools.map.MapContext;
import org.geotools.map.MapLayer;
import org.geotools.referencing.crs.DefaultGeographicCRS;
import org.geotools.renderer.GTRenderer;
import org.geotools.renderer.shape.TransitionShapefileRenderer;
import org.geotools.styling.Style;

import schmitzm.geotools.feature.FeatureUtil;
import schmitzm.geotools.feature.FeatureUtil.GeometryForm;
import schmitzm.geotools.map.event.FeatureModifiedEvent;
import schmitzm.geotools.map.event.LayerEditCanceledEvent;
import schmitzm.geotools.map.event.LayerEditFinishedEvent;
import schmitzm.geotools.map.event.LayerEditStartedEvent;
import schmitzm.geotools.map.event.MapContextSynchronizer;
import schmitzm.swing.InputOption;
import schmitzm.swing.MultipleOptionPane;

import com.vividsolutions.jts.geom.Coordinate;
import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.LineString;
import com.vividsolutions.jts.geom.LinearRing;
import com.vividsolutions.jts.geom.MultiPoint;
import com.vividsolutions.jts.geom.Point;
import com.vividsolutions.jts.geom.Polygon;


/**
 * The {@code GeoEditorPane} extends the {@link JMapPane} with functionalities
 * to create new vector layers by successively click points via mouse.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class JEditorPane extends JMapPane {
  /** Modes, the editor can be perform.
   *  @see JEditorPane#startEditing(EditorMode, String, Style) */
  public static enum EditorMode {
    /** Create a new point layer. */
    New_Point,
    /** Create a new linestring layer. */
    New_Line,
    /** Create a new polygon layer. */
    New_Polygon
  }

  /** Attribute name used for the geometry of new layers. */
  public static final String GEOMETRY_ATTR = "the_geom";

  /** Default-Style for point layers (blue dots). */
  public static final Style DEFAULT_POINT_STYLE = FeatureUtil.createPointStyle(Color.BLUE);
  /** Default-Style for line layers (blue lines). */
  public static final Style DEFAULT_LINE_STYLE = FeatureUtil.createLineStyle(Color.BLUE,2);
  /** Default-Style for polygon layers (orange with black borders). */
  public static final Style DEFAULT_POLYGON_STYLE = FeatureUtil.createPolygonStyle(Color.ORANGE, Color.BLUE, 2);

  /** If the edited layer is empty, a dummy feature is inserted, because
   *  the StreamingRenderer can not handle empty FeatureCollections yet. */
  private Feature DUMMY_EDITOR_FEATURE = null;
  private Feature DUMMY_LINE_FEATURE = null;
  private Feature DUMMY_POINT_FEATURE = null;

  /** The map context of the displayed layer. */
  protected MapContext mapContext = null;

  /** Contains the additional attributes for new FeatureCollections. */
  protected FeatureType additionalAttr = null;
  /** Contains the {@link InputOption} to specify the additional
   *  attribute values.  */
  protected FeatureInputOption attrInputOption = null;

  /** The map context where the edited layers are displayed. This
   *  context is displayed on top of the other layers. */
  protected MapContext editorMapContext = null;
  /** The renderer the edited layers are rendered with. */
  protected GTRenderer editorRenderer = null;
  /** Holds the {@link Style styles} to display edited layers. */
  protected Map<GeometryForm,Style> editorStyles = null;
  /** Holds the operation the editor currently performs. */
  protected EditorMode editorMode = null;

  /** Holds the {@link FeatureType} of the edited layer. */
  protected FeatureType editorFeatureType = null;
  /** Holds the kind of geometry of the edited layer. */
  protected GeometryForm editorGeometryForm = null;
  /** Holds the {@link FeatureCollection} of the edited layer. */
  protected DefaultFeatureCollection editorFeatureCollection = null;
  /** Holds the edited layer. */
  protected MapLayer editorLayer = null;

  /** Holds the {@link FeatureType} of the new segment (incomplete Feature)
   *  displayed as line. */
  protected FeatureType segmLineFeatureType = null;
  /** Holds the {@link FeatureCollection} which holds the new segment
   *  (incomplete Feature) displayed as line. */
  protected DefaultFeatureCollection segmLineFeatureCollection = null;
  /** Holds the layer, the new segment (incomplete Feature) is displayed in
   *  as line. */
  protected MapLayer segmLineLayer = null;

  /** Holds the {@link FeatureType} of the new segment (incomplete Feature)
   *  displayed as points. */
  protected FeatureType segmPointFeatureType = null;
  /** Holds the {@link FeatureCollection} which holds the new segment
   *  (incomplete Feature) displayed as points. */
  protected DefaultFeatureCollection segmPointFeatureCollection = null;
  /** Holds the layer, the new segment (incomplete Feature) is displayed in
   *  as points. */
  protected MapLayer segmPointLayer = null;


  /** Holds the points of the edited segment (incomplete Feature) which
   *  can be undone. */
  protected Stack<Coordinate> segmUndoPoints = new Stack<Coordinate>();
  /** Holds the undone points of the edited segment (incomplete Feature) which
   *  can be redone. */
  protected Stack<Coordinate> segmRedoPoints = new Stack<Coordinate>();
  /** Holds the points of former segments (complete Features) which
   *  can be undone. */
  protected Stack<Stack<Coordinate>> globalUndoPoints = new Stack<Stack<Coordinate>>();
  /** Holds the undone points of former segments (complete Features) which
   *  can be redone. */
  protected Stack<Stack<Coordinate>> globalRedoPoints = new Stack<Stack<Coordinate>>();
  /** Holds the former segments (complete Features) which can be undone. */
  protected Stack<Feature> globalUndoFeatures = new Stack<Feature>();
  /** Holds the undone segments (complete Features) which can be redone. */
  protected Stack<Feature> globalRedoFeatures = new Stack<Feature>();

  /**
   * Creates a new {@code GeoEditorPane}.
   */
  public JEditorPane() {
    this( new BorderLayout(),
          true,
          new TransitionShapefileRenderer(),
          new DefaultMapContext(DefaultGeographicCRS.WGS84)
    );
  }

  /**
   * Creates a new {@code GeoEditorPane}. This constructor provides the possibility
   * to use alternative implementations of the internal components.
   * <b>The specified parameter all can be {@code null}! In this case
   * the respective default component is used.</b>
   */
  public JEditorPane(LayoutManager layout, boolean isDoubleBuffered, GTRenderer renderer, MapContext context) {
    super(layout,isDoubleBuffered,renderer,context);
    this.mapContext = getContext();
    // initalize the editor styles
    this.editorStyles = new HashMap<GeometryForm, Style>();
    for (GeometryForm geomForm : GeometryForm.values())
      this.setEditorStyle(geomForm, null);

    // no standard action on left mouse button, instead: adding points
    setState( NONE );
    setWindowSelectionState( ZOOM_IN );
    // special map context for the editor layers, so the
    // editor layers are not shown in "normal" layer list
    this.editorRenderer   = new TransitionShapefileRenderer();
    this.editorMapContext = new DefaultMapContext( mapContext.getCoordinateReferenceSystem() );
    // listen to CRS/Area changes on the "normal" Context to
    // synchronize the editor context
    this.mapContext.addMapBoundsListener( new MapContextSynchronizer(editorMapContext) );

    // create the input option to specify additional attribute values
    attrInputOption = new FeatureInputOption(null,true,(Feature)null);
  }

  /**
   * After the actions of the super method, this method paints the
   * special editor layers in {@link #editorMapContext}.
   */
  protected void paintComponent(Graphics g) {
    resetEditorLayerVisibility();
    super.paintComponent(g);
    Rectangle r = getBounds();
    Rectangle dr = new Rectangle(r.width, r.height);
    editorRenderer.setContext(editorMapContext);
    editorRenderer.paint((Graphics2D) g, dr, mapArea);
  }

  /**
   * Ignores the actions of the super class for left-clicks.
   * Instead this method reacts according to the current
   * editor mode.
   */
  public void mouseClicked(MouseEvent e) {
    Point2D geoCoord = getTransform().transform(e.getPoint(), null);
    // Zunaechst Modus auf NULL pruefen, da ansonsten
    // NullPointerException in switch-Statement
    if ( editorMode == null )
      return;

    switch( editorMode ) {
      case New_Point:
      case New_Line:
      case New_Polygon: addSegment( new Coordinate(geoCoord.getX(), geoCoord.getY()) );
                        break;
    }
  }

  //**********************************************************************
  //***** General public methods
  //**********************************************************************

  /**
   * Checks, whether a layer is edited.
   */
  public boolean isEditorEnabled() {
    return editorMode != null;
  }

  /**
   * Sets the style for edited layers.
   * @param geomForm kind of layers the style is set for
   * @param style a Style (if {@code null} a default style is set, so
   *              {@link #getEditorStyle(GeometryForm)} never returns null)
   */
  public void setEditorStyle(GeometryForm geomForm, Style style) {
    // If no style is specified, set a default style
    if ( style == null )
      switch ( geomForm ) {
        case POINT:   style = DEFAULT_POINT_STYLE; break;
        case LINE:    style = DEFAULT_LINE_STYLE; break;
        case POLYGON: style = DEFAULT_POLYGON_STYLE; break;
      }

    // Set the style
    editorStyles.put(geomForm, style);

    // Apply the style also to the current edited layer
    if ( editorLayer != null && FeatureUtil.getGeometryForm(editorLayer) == geomForm )
      editorLayer.setStyle(style);
    if ( segmLineLayer != null && FeatureUtil.getGeometryForm(segmLineLayer) == geomForm )
      segmLineLayer.setStyle(style);
    if ( segmPointLayer != null && FeatureUtil.getGeometryForm(segmPointLayer) == geomForm )
      segmPointLayer.setStyle(style);
  }

  /**
   * Returns the style for edited layers.
   * @param geomForm type of layers the style is returned for
   */
  public Style getEditorStyle(GeometryForm geomForm) {
    return this.editorStyles.get(geomForm);
  }

  /**
   * Returns the style for the current edited layer.
   */
  public Style getEditorStyle() {
    if ( editorGeometryForm == null )
      return null;
    return this.editorStyles.get(editorGeometryForm);
  }

  /**
   * Sets the additional attributes for new FeatureCollections (besides
   * the geometric attribute, which is added automatically).
   * @param fType defines the attributes (If {@code null}, new FeatureCollections
   *              only contain the geometric attribute)
   */
  public void setAdditionalAttributes(FeatureType fType) {
    this.additionalAttr = fType;
  }

  /**
   * Returns the additional attributes for new FeatureCollections (besides
   * the geometric attribute, which is added automatically).
   */
  public FeatureType getAdditionalAttributes() {
    return this.additionalAttr;
  }

  /**
   * Returns the current editor mode.
   */
  public EditorMode getEditorMode() {
    return this.editorMode;
  }

  /**
   * Returns the edited layer.
   */
  public MapLayer getEditorLayer() {
    return this.editorLayer;
  }

  /**
   * Returns the {@link FeatureCollection} of the edited layer.
   */
  public FeatureCollection getEditorFeatureCollection() {
    return this.editorFeatureCollection;
  }

  //**********************************************************************
  //*****  Public Methods controlling the editor operations
  //**********************************************************************

  /**
   * Starts a new layer.
   * @param mode type of layer
   * @param layerTitle Title for the new layer
   * @exception UnsupportedOperationException if no map area is defined yet
   */
  public void startEditing(EditorMode mode, String layerTitle) {
    // to create a new layer, first an existing layer must be
    // displayed, so that a CRS and geo-position is available
    // for the new layer
    if ( getMapArea() == null )
      throw new UnsupportedOperationException( GeotoolsGUIUtil.RESOURCE.getString("schmitzm.geotools.gui.JEditorPane.Err.MissingMap") );
    // cancel active editings
    cancelEditing();
    editorMode = mode;

    // Create new Layers
    initEditorFeatureCollection(layerTitle);
    initSegmentFeatureCollection();
    editorMapContext.addLayer(0,editorLayer);
    fireMapPaneEvent( new LayerEditStartedEvent(this,editorLayer,this) );
  }

  /**
   * Called when the mouse is clicked on map during creating a new layer.
   * @param coord World coordinates of click position
   * @param refresh indicates whether the visualisation will be refreshed
   * @param aClearRedo indicates whether the REDO-Stack is cleared (normally {@code true},
   *                   but {@code false} during REDO operation!)
   */
  protected void addSegment(Coordinate coord, boolean refresh, boolean clearRedo) {
    if ( editorMode != EditorMode.New_Point &&
         editorMode != EditorMode.New_Line &&
         editorMode != EditorMode.New_Polygon )
      return;

    // store the new point for undoing
    segmUndoPoints.push(coord);
    // delete all former redo possibilities
    if ( clearRedo )
      segmRedoPoints.clear();
    // recreate the segment FeatureCollection to show the extended line
    generateSegmentFeatureCollection();

    // finish the feature automatically
    // - when editing points, because there are no segments for point layer,
    //   because every point is a new feature
    // - when editing polygons and reaching the start point
    if ( editorGeometryForm == GeometryForm.POINT ||
         editorGeometryForm == GeometryForm.POLYGON && segmUndoPoints.size() > 1 && segmUndoPoints.lastElement().equals(segmUndoPoints.firstElement()) )
      finishFeature(refresh, clearRedo); // refresh is done by finishFeature()
    else
      if ( refresh )
        refresh();
  }

  /**
   * Called when the mouse is clicked on map during creating a new layer.
   * @param coord World coordinates of click position
   */
  protected void addSegment(Coordinate coord) {
    addSegment(coord, true, true);
  }

  /**
   * Finishes the editing of the current segment (Feature).
   * If no editing operation is currently in progress, this method does nothing.
   * @param refresh indicates whether the visualisation will be refreshed
   * @param aClearRedo indicates whether the REDO-Stack is cleared (normally {@code true},
   *                   but {@code false} during REDO operation!)
   * @exception UnsupportedOperationException if a line or polygon feature can not
   *            be finished because of less specified points
   */
  protected void finishFeature(boolean refresh, boolean clearRedo) {
    if ( editorMode == null )
      return;

    if ( editorGeometryForm == GeometryForm.LINE && segmUndoPoints.size() < 2 )
      throw new UnsupportedOperationException( GeotoolsGUIUtil.RESOURCE.getString("schmitzm.geotools.gui.JEditorPane.Err.Line.LessPoints") );
    if ( editorGeometryForm == GeometryForm.POLYGON && segmUndoPoints.size() < 3 )
      throw new UnsupportedOperationException( GeotoolsGUIUtil.RESOURCE.getString("schmitzm.geotools.gui.JEditorPane.Err.Polygon.LessPoints") );

    // close a polygon automatically
    if ( editorGeometryForm == GeometryForm.POLYGON &&
        !segmUndoPoints.lastElement().equals(segmUndoPoints.firstElement()) ) {
      addSegment( segmUndoPoints.firstElement(), refresh, clearRedo );
      return; // !!!! finishFeature(.) is called again by addSegment(..) !!!!
    }
    // create a new Feature from the segment points
    Feature feature = createFeature();
    if ( feature == null )
      return;
    fireMapPaneEvent( new FeatureModifiedEvent(this,editorLayer,feature,this) );

    // update the global undo
    globalUndoPoints.push( segmUndoPoints );
    globalUndoFeatures.push( feature );
    editorFeatureCollection.add( feature );

    // after finishing a Feature, a new segment is started
    segmUndoPoints = new Stack<Coordinate>();
    if ( clearRedo )
      segmRedoPoints = globalRedoPoints.isEmpty() ? new Stack<Coordinate>() : globalRedoPoints.pop();
    initSegmentFeatureCollection();
    // refresh the display
    if ( refresh )
      refresh();
  }

  /**
   * Finishes the editing of the current segment (Feature).
   * If no editing operation is currently in progress, this method does nothing.
   */
  public void finishFeature() {
    finishFeature(true, true);
  }

  /**
   * Finishes the current editing operation. After that no more operations
   * on this layer can be performed.
   * As long as no other editing operation is started, the edited (new) layer is
   * available by {@link #getEditorLayer()} and {@link #getEditorFeatureCollection()}.
   * If no editing operation is currently in progress, this method does nothing.
   */
  public void finishEditing() {
    if ( editorMode == null )
      return;

    // if current segment is not closed, finish it automatically
    if ( !segmUndoPoints.isEmpty() )
      finishFeature();
    initUndoRedo();

    // move the edited layer to the "real" MapContext
    editorMapContext.removeLayer(editorLayer);
    if ( mapContext.indexOf(editorLayer) < 0 ) {
      mapContext.addLayer(editorLayer);
      editorLayer.setVisible(true);
    }
    fireMapPaneEvent( new LayerEditFinishedEvent(this,editorLayer,this) );

    // initialize the editing variables
    editorMode = null;
    editorLayer = null;
    editorFeatureCollection = null;
    editorGeometryForm = null;

    // refresh the display
    refresh();
  }

  /**
   * Cancels the current edititing operation. The edited layer is removed.
   */
  public void cancelEditing() {
    if ( editorLayer != null )
      this.editorMapContext.removeLayer( editorLayer );
    if ( editorFeatureCollection != null ) {
      if ( mapContext.indexOf(editorLayer) < 0 ) {
        // Layer was a new one, so clear it completly
        editorFeatureCollection.clear();
        initUndoRedo();
        fireMapPaneEvent( new LayerEditCanceledEvent(this,null,this) );
      } else {
        // Layer was an existing one, so undo all editing operations
        undoAll();
        initUndoRedo();
        fireMapPaneEvent( new LayerEditFinishedEvent(this,editorLayer,this) );
      }
    }
    if ( segmLineFeatureCollection != null )
      segmLineFeatureCollection.clear();
    if ( segmPointFeatureCollection != null )
      segmPointFeatureCollection.clear();

    // initialize the editing variables
    editorMode = null;
    editorLayer = null;
    editorFeatureCollection = null;
    editorGeometryForm = null;

    // refresh the display
    refresh();
  }

  /**
   * Makes previously made editing actions undone.
   * @param count count of operations made undone
   */
  public void undoEditing(int count) {
    for (int i=0; i<count; i++)
    {
      if ( segmUndoPoints.isEmpty() ) {
        if ( globalUndoPoints.isEmpty() )
          // No more undo operations available
          break;
        else {
          segmUndoPoints = globalUndoPoints.pop();
          editorFeatureCollection.remove( globalUndoFeatures.peek() );

          globalRedoFeatures.push( globalUndoFeatures.pop() );
          if ( !segmRedoPoints.isEmpty() )
            globalRedoPoints.push( segmRedoPoints );
          segmRedoPoints = new Stack<Coordinate>();

          // On lines the last point and finish feature operation
          // are separated, so they also must be undone
          // separately.
          if ( editorGeometryForm == GeometryForm.LINE ) {
            generateSegmentFeatureCollection();
            continue;
          }
        }
      }
      segmRedoPoints.push( segmUndoPoints.pop() );
      generateSegmentFeatureCollection();
    }
    refresh();
  }

  /**
   * Makes the last made editing action undone.
   */
  public void undoEditing() {
    undoEditing(1);
  }

  /**
   * Makes all made editing action undone.
   */
  public void undoAll() {
    undoEditing( Integer.MAX_VALUE );
  }

  /**
   * Checks wheater a undo operation can be performed.
   */
  public boolean isUndoPossible() {
    return !segmUndoPoints.isEmpty() || !globalUndoPoints.isEmpty();
  }

  /**
   * Restores previous undone editing actions.
   * @param count count of redo operations
   */
  public void redoEditing(int count) {
    for (int i=0; i<count; i++)
    {
      if ( segmRedoPoints.isEmpty() ) {
        if ( globalRedoPoints.isEmpty() )
          // No more redo operations available
          break;
        else {
          // Lines must be finished when all segment redo operations
          // are redone; Points and Polygons must not, because they
          // were already closed on last addSegment(.) (closing point
          // of polygon is also a part of the redo stack!)
          if ( editorGeometryForm == GeometryForm.LINE )
            finishFeature(false, false);
          segmRedoPoints = globalRedoPoints.pop();
        }
      }
      addSegment( segmRedoPoints.pop(), false, false );
    }
    refresh();
  }

  /**
   * Restores the last undone editing action.
   */
  public void redoEditing() {
    redoEditing(1);
  }

  /**
   * Checks wheater a redo operation can be performed.
   */
  public boolean isRedoPossible() {
    // Redo possible if
    // - some operations of the current segment can be redone
    // - some operations of former segments can be redone
    return !segmRedoPoints.isEmpty() || !globalRedoPoints.isEmpty();
  }

  //**********************************************************************
  //*****  Helper methods for the editor operations
  //**********************************************************************
  /**
   * Sets {@link #editorFeatureCollection} and {@link #editorLayer} to completely
   * new instances.
   * @param title title for the new layer
   */
  protected void initUndoRedo() {
    // initialize the Undo/Redo-Stacks
    for ( Stack s : globalRedoPoints ) s.clear();
    for ( Stack s : globalUndoPoints ) s.clear();
    globalRedoPoints.clear();
    globalUndoPoints.clear();
    globalRedoFeatures.clear();
    globalUndoFeatures.clear();
    segmRedoPoints.clear();
    segmUndoPoints.clear();
  }

  /**
   * Sets {@link #editorFeatureCollection} and {@link #editorLayer} to completely
   * new instances.
   * @param title title for the new layer
   */
  protected void initEditorFeatureCollection(String title) {
    // create the new layer
    editorFeatureCollection = (DefaultFeatureCollection)DefaultFeatureCollections.newCollection(title);
    editorFeatureType       = createFeatureType(editorMode, getAdditionalAttributes());
    DUMMY_EDITOR_FEATURE    = createDummyFeature(editorFeatureType);
    editorFeatureCollection.add(DUMMY_EDITOR_FEATURE);
    editorGeometryForm      = FeatureUtil.getGeometryForm(editorFeatureCollection);
    editorLayer             = new DefaultMapLayer(editorFeatureCollection,getEditorStyle(editorGeometryForm));
    editorLayer.setTitle(title);
  }

  /**
   * Clears the {@link FeatureCollection} used to handle the current
   * edited segment.
   */
  protected void initSegmentFeatureCollection() {
    // initialize the segment to start a new feature
    if ( segmLineFeatureCollection == null ) {
      // initialize the objects used to display the
      // current segment as line
      segmLineFeatureCollection = (DefaultFeatureCollection)DefaultFeatureCollections.newCollection("SegmentAsLine");
      segmLineFeatureType       = createFeatureType(LineString.class, null);
      DUMMY_LINE_FEATURE        = createDummyFeature(segmLineFeatureType);
      segmLineFeatureCollection.add(DUMMY_LINE_FEATURE);
      segmLineLayer             = new DefaultMapLayer(segmLineFeatureCollection,getEditorStyle(GeometryForm.LINE));
      segmLineLayer.setTitle("Line layer for current segment");
      editorMapContext.addLayer(segmLineLayer);
      // initialize the objects used to display the
      // current segment as points
      segmPointFeatureCollection = (DefaultFeatureCollection)DefaultFeatureCollections.newCollection("SegmentAsPoints");
      segmPointFeatureType       = createFeatureType(MultiPoint.class, null);
      DUMMY_POINT_FEATURE        = createDummyFeature(segmPointFeatureType);
      segmPointFeatureCollection.add(DUMMY_POINT_FEATURE);
      segmPointLayer             = new DefaultMapLayer(segmPointFeatureCollection,getEditorStyle(GeometryForm.POINT));
      segmPointLayer.setTitle("Point layer for current segment");
      editorMapContext.addLayer(segmPointLayer);
    } else {
      segmLineFeatureCollection.clear();
      segmPointFeatureCollection.clear();
    }
  }

  /**
   * Creates a new segment Feature from the currently selected points.
   */
  protected void generateSegmentFeatureCollection() {
    initSegmentFeatureCollection();
    // show the segment as single points
    Feature segmPointFeature = createSegmentPointFeature();
    if ( segmPointFeature != null )
      segmPointFeatureCollection.add( segmPointFeature );
    // show the segment as a line
    Feature segmLineFeature = createSegmentLineFeature();
    if ( segmLineFeature != null )
      segmLineFeatureCollection.add( segmLineFeature );
    // Fire event even if no (real) feature is created!
    fireMapPaneEvent( new FeatureModifiedEvent(this,segmLineLayer,segmLineFeature,this) );
  }

  /**
   * Creates a new {@link Feature Line-Feature} from the currently selected
   * segment points.
   */
  protected Feature createSegmentLineFeature() {
    // Line can only be created with at least 2 points
    // For point layer the segment is not shown
    if ( segmUndoPoints.size() < 2 ||
         editorGeometryForm == GeometryForm.POINT )
      return null;

    Coordinate[] coord      = segmUndoPoints.toArray(new Coordinate[0]);
    LineString   lineString = createGeometryFromPoints(LineString.class, coord);
    try {
      return segmLineFeatureType.create( new Object[] {lineString} );
    } catch (Exception e) {
      throw new RuntimeException(e);
    }
  }

  /**
   * Creates a new {@link Feature MultiPoint-Feature} from the currently selected
   * segment points.
   */
  protected Feature createSegmentPointFeature() {
    // Point can only be created with at least 1 point
    // For point layer the segment is not shown
    if ( segmUndoPoints.size() < 1 ||
         editorGeometryForm == GeometryForm.POINT )
      return null;

    Coordinate[] coord      = segmUndoPoints.toArray(new Coordinate[0]);
    MultiPoint   multiPoint = createGeometryFromPoints(MultiPoint.class, coord);
    try {
      return segmPointFeatureType.create( new Object[] {multiPoint} );
    } catch (Exception e) {
      throw new RuntimeException(e);
    }
  }

  /**
   * Creates a new Feature from the currently selected points. The non-geometric
   * attributes are set to default values.
   */
  private Feature createFeature() {
    Coordinate[] coord   = segmUndoPoints.toArray(new Coordinate[0]);
    Feature      feature = null;
    try {
      // if a feature was undone previously use this as
      // default for the new feature
      if ( !globalRedoFeatures.isEmpty() )
        feature = globalRedoFeatures.pop();
      else {
        // generate default attribute values
        Object[] attr = FeatureUtil.getDefaultAttributeValues(editorFeatureType);
        // replace default attribute values with auto generate
        // values (for all registered attributes)
        for (int i=0; i<attr.length; i++) {
          AttributeType aType = editorFeatureType.getAttributeType(i);
          if ( FeatureUtil.getAutoValueGenerator(aType) != null )
            attr[i] = FeatureUtil.getNextAutoValue(aType);
        }
        // create new feature
        feature = editorFeatureType.create(attr);
      }
      feature.setDefaultGeometry(
          createGeometryFromPoints(getGeometryType(editorMode), coord )
      );

      if ( feature.getNumberOfAttributes() > 1 ) {
        attrInputOption.setValue(feature);
        Object[] value = MultipleOptionPane.showMultipleInputDialog(
                            this,
                            GeotoolsGUIUtil.RESOURCE.getString("schmitzm.geotools.gui.JEditorToolBar.NewFeature.title"),
                            attrInputOption
        );
        if ( value == null )
          return null;
      }
      return feature;
    } catch (Exception e) {
      throw new RuntimeException(e);
    }

  }


  //**********************************************************************
  //*****  Helper methods to deal with geometries
  //**********************************************************************
  /**
   * Determines the geometry type to deal with according to the
   * editor mode.
   * @param mode editor mode
   */
  private static Class<? extends Geometry> getGeometryType(EditorMode mode) {
    switch( mode ) {
      case New_Point:   return Point.class;
      case New_Line:    return LineString.class;
      case New_Polygon: return Polygon.class;
    }
    return null;
  }

  /**
   * Creates a geometry (point, linestring/ring, polygon) from a set of points. If
   * no points are specified ({@code null}), a default geometry is created.
   * @param gtype type of geometry
   * @param coord set of points (can be {@code null})
   */
  private static <G extends Geometry> G createGeometryFromPoints(Class<G> gtype, Coordinate[] coord) {
    // Point
    if ( gtype.isAssignableFrom( com.vividsolutions.jts.geom.Point.class ) ) {
      if ( coord == null )
        coord = new Coordinate[] { new Coordinate() };
      return (G)FeatureUtil.GEOMETRY_FACTORY.createPoint( coord[0] );
    }
    // MultiPoint
    if ( gtype.isAssignableFrom( com.vividsolutions.jts.geom.MultiPoint.class ) ) {
      if ( coord == null )
        coord = new Coordinate[] { new Coordinate() };
      return (G)FeatureUtil.GEOMETRY_FACTORY.createMultiPoint( coord );
    }
    // LineString
    if ( gtype.isAssignableFrom( com.vividsolutions.jts.geom.LineString.class ) ) {
      if ( coord != null && coord.length > 2 && coord[0].equals(coord[coord.length-1]) )
        return (G)FeatureUtil.GEOMETRY_FACTORY.createLinearRing(coord);
      return (G)FeatureUtil.GEOMETRY_FACTORY.createLineString(coord);
    }
    // Polygon
    if ( gtype.isAssignableFrom( com.vividsolutions.jts.geom.Polygon.class ) ) {
      LinearRing shell = null;
      if ( coord != null && coord.length > 2 && coord[0].equals(coord[coord.length-1]) )
        shell = FeatureUtil.GEOMETRY_FACTORY.createLinearRing(coord);
      return (G)FeatureUtil.GEOMETRY_FACTORY.createPolygon(shell,null);
    }

    throw new UnsupportedOperationException("Can not create geometry for "+gtype.getName());
  }

  /**
   * Extends a feature type with a default geometry attribute.
   * @param gtype default geometry of the new feature type
   * @param ftype a feature type which is extended (can be {@code null})
   */
  private FeatureType createFeatureType(Class<? extends Geometry> gtype, FeatureType ftype) {
    GeometryAttributeType geomAttrType = new GeometricAttributeType(
        GEOMETRY_ATTR,
        gtype,
        false,
        createGeometryFromPoints(gtype,null),
        mapContext.getCoordinateReferenceSystem(),
        null
    );
    String ftypeName           = ftype != null ? ftype.getTypeName() : gtype.getSimpleName()+"_feature";
    FeatureTypeBuilder builder = FeatureTypeBuilder.newInstance(ftypeName);
    builder.setDefaultGeometry( geomAttrType );
    if ( ftype != null )
      builder.addTypes( ftype.getAttributeTypes() );
    try {
      return builder.getFeatureType();
    } catch ( SchemaException err ) {
      throw new RuntimeException(err);
    }
  }

  /**
   * Extends a feature type with a default geometry attribute.
   * @param mode  specifies the the default geometry used for the feature type
   * @param ftype a feature type which is extended (can be {@code null})
   */
  private FeatureType createFeatureType(EditorMode mode, FeatureType ftype) {
    Class<? extends Geometry> gtype = null;
    switch( mode ) {
      case New_Point:   gtype = Point.class; break;
      case New_Line:    gtype = LineString.class; break;
      case New_Polygon: gtype = Polygon.class; break;
    }
    return createFeatureType(gtype,ftype);
  }

  /**
   * Creates a feature with default values for a given feature type.
   * @param ftype a feature type
   */
  private static Feature createDummyFeature(FeatureType ftype) {
    try {
      return ftype.create( FeatureUtil.getDefaultAttributeValues(ftype) );
    } catch (Exception e) {
      throw new RuntimeException(e);
    }
  }

  /**
   * Checks, whether the editor layer is empty. If it is, a dummy feature is inserted
   * because the StreamingRenderer can not handle empty Feature-Layers yet
   * (also if the layer is hidden!).
   * At the moment, the first a feature is added, the dummy feature is removed, so
   * is is not displayed any time.
   */
  private void resetEditorLayerVisibility() {
    // if layer is empty, insert a Dummy-Feature (and hide the
    // layer) because the StreamingRenderer can not handle empty
    // Feature-Layers yet
    if ( segmLineFeatureCollection != null && segmLineFeatureCollection.size() == 0 )
      segmLineFeatureCollection.add( DUMMY_LINE_FEATURE );
    if ( segmPointFeatureCollection != null && segmPointFeatureCollection.size() == 0 )
      segmPointFeatureCollection.add( DUMMY_POINT_FEATURE );
    if ( editorFeatureCollection != null && editorFeatureCollection.size() == 0 )
      editorFeatureCollection.add( DUMMY_EDITOR_FEATURE );

    // if the layer is not empty anymore, remove the Dummy-Feature, so
    // it is not displayed
    if ( segmLineFeatureCollection != null && segmLineFeatureCollection.size() == 2 )
      segmLineFeatureCollection.remove( DUMMY_LINE_FEATURE );
    if ( segmPointFeatureCollection != null && segmPointFeatureCollection.size() == 2 )
      segmPointFeatureCollection.remove( DUMMY_POINT_FEATURE );
    if ( editorFeatureCollection != null && editorFeatureCollection.size() == 2 )
      editorFeatureCollection.remove( DUMMY_EDITOR_FEATURE );

    // hide the editor layer if it only contains the dummy feature
    if ( editorLayer != null )
      editorLayer.setVisible(editorFeatureCollection.size() > 1 || !editorFeatureCollection.contains(DUMMY_EDITOR_FEATURE));
    if ( segmLineLayer != null )
      segmLineLayer.setVisible(segmLineFeatureCollection.size() > 1 || !segmLineFeatureCollection.contains(DUMMY_LINE_FEATURE));
    if ( segmPointLayer != null )
      segmPointLayer.setVisible(segmPointFeatureCollection.size() > 1 || !segmPointFeatureCollection.contains(DUMMY_POINT_FEATURE));
  }

}
