/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.geotools.io;

import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;
import java.awt.image.RenderedImage;
import java.awt.image.WritableRaster;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.URL;
import java.nio.channels.FileChannel;
import java.util.HashMap;
import java.util.Map;
import java.util.StringTokenizer;
import java.util.Vector;

import javax.imageio.IIOException;
import javax.imageio.ImageIO;

import org.apache.log4j.Logger;
import org.geotools.coverage.grid.GridCoverage2D;
import org.geotools.coverage.grid.GridCoverageFactory;
import org.geotools.coverage.grid.io.AbstractGridCoverage2DReader;
import org.geotools.data.DataStore;
import org.geotools.data.shapefile.Lock;
import org.geotools.data.shapefile.ShapefileDataStore;
import org.geotools.data.shapefile.indexed.IndexedShapefileDataStore;
import org.geotools.data.shapefile.shp.ShapefileReader;
import org.geotools.factory.Hints;
import org.geotools.feature.FeatureCollection;
import org.geotools.feature.FeatureCollections;
import org.geotools.gce.arcgrid.ArcGridRaster;
import org.geotools.gce.geotiff.GeoTiffReader;
import org.geotools.gce.image.WorldImageReader;
import org.geotools.geometry.Envelope2D;
import org.geotools.referencing.CRS;
import org.geotools.referencing.crs.DefaultGeographicCRS;
import org.opengis.referencing.crs.CoordinateReferenceSystem;

import schmitzm.data.WritableGridRaster;
import schmitzm.geotools.GTUtil;
import schmitzm.io.IOUtil;

import com.vividsolutions.jts.geom.Geometry;

/**
 * In dieser Klasse sind Funktionen zum Datenimport von Geo-Daten
 * zusammengefasst.
 * 
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
 *         (University of Bonn/Germany)
 * @version 1.0
 */
public class GeoImportUtil {
	private static final Logger LOGGER = Logger.getLogger(GeoImportUtil.class
			.getName());

	/**
	 * These postfixes are associated with Arc/Info ASCII Grid files TODO .0
	 * habe ich auch schon gesehen. Nut leider kann man das nicht in einen enum
	 * packen.
	 * */
	static public enum ARCASCII_POSTFIXES {
		arc, dat, ascii, txt, asc, a00
	};

	/** These postfixes are associated with GeoTiff files */
	static public enum GEOTIFF_POSTFIXES {
		tif, tiff
	};

	/** These postfixes are associated with ESRI SHape files */
	static public enum SHP_POSTFIXES {
		shp, sbx, sbn, dbf, shx, qix, fix, // Non-ESRI QuadTree Index stuff
		cpg
		// Supported by ESRI. Contains the name of codepage used inside the DBF
	};

	/**
	 * These postfixes are associated with ordinary image files, excluding
	 * GeoTIFF endings please
	 */
	static public enum IMAGE_POSTFIXES {
		png, gif, jpg, jpeg
	};

	/** These postfixes are associated with world files */
	static public enum WORLD_POSTFIXES {
		wld, jgw, pgw, tfw
	};

	/**
	 * Standard-CRS, welches verwendet wird, wenn beim Import kein CRS ermittelt
	 * werden kann (Default: {@link DefaultGeographicCRS#WGS84}).<br>
	 * <b>Achtung:</b><br>
	 * Anwendungen koennen diese Variable gefahrlos ueberschreiben, um ein fuer
	 * die Anwendung adaequates Standard-CRS zu verwenden.
	 */
	public static CoordinateReferenceSystem DEFAULT_CRS = DefaultGeographicCRS.WGS84;

	/**
	 * Diese Methode extrahiert saemtliche Features aus einem ShapeFile-Projekt
	 * (<code><i>name</i>.shp <i>name</i>.prj <i>name</i>.dbf ...</code>) und
	 * speichert diese in einer
	 * <code>org.geotools.feature.FeatureCollection</code>.<br>
	 * 
	 * @param url
	 *            {@link URL} to Shape-File
	 * @param prjUrl
	 *            {@link URL} zu .prj Datei des Shape-File
	 * 
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 * 
	 * @return {@link FeatureCollection} that was read
	 * 
	 * @throws IOException
	 */
	public static FeatureCollection readFeaturesFromShapeURL(URL url, URL prjUrl)
			throws IOException {
		ShapefileDataStore store = new ShapefileDataStore(url);
		try {
			// Testen, ob Projektion ermittelt werden kann, um vorab das
			// Standard-CRS
			// zu setzen
			LOGGER.debug("  parseWKT on " + prjUrl);
			CRS.parseWKT(readProjectionString(prjUrl));
		} catch (Exception err) {
			LOGGER.warn(" CRS.parseWKT mit Ex\n: ", err);
			LOGGER.warn(" No projection found for file. Default is used.");
			LOGGER
					.warn(" NOT calling forceSchemaCRS now... please provide a URL to an existing .prj file!");
			// store.forceSchemaCRS(DEFAULT_CRS);
		}
		String[] typeNames = store.getTypeNames();
		FeatureCollection fc = store.getFeatureSource(typeNames[0])
				.getFeatures();

		// Create a new DefaultFeatureCollection to allow modifying
		// operations on the collection ("fc" is a DataFeatureCollection, whose
		// add(.) and remove(.) methods do nothing. Furthermore a
		// FIDFeatureReader
		// is used which returns copies of the features, so modifying the
		// attributes
		// does not effect the features in the collection!)
		FeatureCollection fc1 = FeatureCollections.newCollection(fc.getID());
		fc1.addAll(fc);
		fc = fc1;

		return fc;
	}

	/**
	 * Diese Methode extrahiert saemtliche Features aus einem ShapeFile-Projekt
	 * (<code><i>name</i>.shp <i>name</i>.prj <i>name</i>.dbf ...</code>) und
	 * speichert diese in einer
	 * <code>org.geotools.feature.FeatureCollection</code>.<br>
	 * 
	 * @param file
	 *            Shape-File
	 * @throws java.lang.Exception
	 *             bei irgendeinem Fehler
	 * 
	 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
	 *         (University of Bonn/Germany)
	 * 
	 *         TODO Wenn readFeaturesFromShapeURL getestet ist, dann sollte die
	 *         hiet auf readFeaturesFromShapeURL(file.getURL) umgeleitet werden
	 *         (SK)
	 */
	public static FeatureCollection readFeaturesFromShapeFile(File file)
			throws Exception {
		ShapefileDataStore store = new ShapefileDataStore(file.toURI().toURL());
		File prjFile = IOUtil.changeFileExt(file, "prj");
		boolean delPrjFile = false;
		try {
			// Testen, ob Projektion ermittelt werden kann, um vorab das
			// Standard-CRS
			// zu setzen
			String prjString = readProjectionString(prjFile);
			if (prjString == null || prjString.trim().equals(""))
				throw new FileNotFoundException("No proper prj-File exists: "
						+ file.getName());
			CRS.parseWKT(prjString);
		} catch (FileNotFoundException err) {
			store.forceSchemaCRS(getDEFAULT_CRS());
			LOGGER.warn("No projection found for file " + file.getName()
					+ ". Default is used.");
			delPrjFile = true; // von DataStore erzeugte Datei wieder loeschen
		}
		String[] typeNames = store.getTypeNames();
		FeatureCollection fc = store.getFeatureSource(typeNames[0])
				.getFeatures();
		if (delPrjFile)
			prjFile.delete();

		// Create a new DefaultFeatureCollection to allow modifying
		// operations on the collection ("fc" is a DataFeatureCollection, whose
		// add(.) and remove(.) methods do nothing. Furthermore a
		// FIDFeatureReader
		// is used which returns copies of the features, so modifying the
		// attributes
		// does not effect the features in the collection!)
		FeatureCollection fc1 = FeatureCollections.newCollection(fc.getID());
		fc1.addAll(fc);
		fc = fc1;

		return fc;
	}

	/**
	 * TODO DOKU
	 * 
	 * @param shpURL
	 * @param prjURL
	 * @return
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 * @throws IOException
	 */
	public static DataStore readDataStoreFromShape(URL shpURL, URL prjURL)
			throws IOException {
		Map map = new HashMap();
		map.put("url", shpURL);
		map.put("create spatial index", true);

		// DataStore dataStore = DataStoreFinder.getDataStore( map );
		DataStore dataStore = new IndexedShapefileDataStore(shpURL, null,
				false, true, IndexedShapefileDataStore.TREE_QIX);
		// DataStore dataStore = new ShapefileDataStore(shpURL);
		System.out.println("DataStore = " + dataStore.getClass().toString());
		IndexedShapefileDataStore dataStoreIndex = (IndexedShapefileDataStore) dataStore;
		System.out.println("indexed = " + dataStoreIndex.isIndexed());
		System.out.println("memory = " + dataStoreIndex.isMemoryMapped());
		return dataStore;
	}

	/**
	 * Diese Methode extrahiert saemtliche Geometrien aus einem
	 * ShapeFile-Projekt (
	 * <code><i>name</i>.shp <i>name</i>.prj <i>name</i>.dbf ...</code>) und
	 * speichert diese in einer Liste (Vector) von
	 * {@link com.vividsolutions.jts.geom.Geometry}-Objekten.<br>
	 * Baut auf folgenden Geotools-Klassen auf: <code>
	 * <ul>
	 * <li>{@link ShapefileReader org.geotools.data.shapefile.shp.ShapefileReader}</li>
   * <li>{@link Geometry        com.vividsolutions.jts.geom.Geometry}</li>
   * </ul>
   * </code>
	 * 
	 * @param file
	 *            Shape-File
	 * @throws java.lang.Exception
	 *             bei irgendeinem Fehler
	 */
	public static Vector<Geometry> readGeometriesFromShapeFile(File file)
			throws Exception {
		FileChannel in = new FileInputStream(file).getChannel();
		ShapefileReader r = new ShapefileReader(in, new Lock());
		Vector<Geometry> geomList = new Vector<Geometry>();
		for (int i = 1; r.hasNext(); i++) {
			// org.geotools.renderer.geom.Geometry shape =
			// (org.geotools.renderer.geom.Geometry) r.nextRecord().shape();
			// org.geotools.geometry.Geometry shape =
			// (org.geotools.geometry.Geometry) r.nextRecord().shape();
			Object obj = r.nextRecord().shape();
			Geometry shape = (Geometry) obj;
			geomList.add(shape);
		}
		r.close();
		return geomList;
	}

	/**
	 * Diese Methode importiert ein Raster aus einer Datei im
	 * ArcInfoASCII-Grid-Format. Das CRS wird aus einem prj-File (EPSG-Code
	 * "EPSG:..." oder WKT-Definition) gelesen. Ist dies nicht erfolgreich, wird
	 * {@link #DEFAULT_CRS} als CRS verwendet.
	 * 
	 * @param file
	 *            ASCII-File
	 * @throws java.lang.Exception
	 *             bei irgendeinem Fehler
	 * @return {@link GridCoverage2D}
	 */
	public static GridCoverage2D readGridFromArcInfoASCII(File file)
			throws Exception {
		return readGridFromArcInfoASCII(file, null);
	}

	/**
	 * Diese Methode importiert ein Raster aus einer Datei im
	 * ArcInfoASCII-Grid-Format. Wenn kein CRS angegeben wird, wird versucht das
	 * CRS aus einem prj-File (EPSG-Code "EPSG:..." oder WKT-Definition) zu
	 * lesen. Ist dies nicht erfolgreich, wird {@link #DEFAULT_CRS} als CRS
	 * verwendet.
	 * 
	 * @param file
	 *            ASCII-File
	 * @param crs
	 *            CoordinateReferenceSystem fuer das Raster (kann {@code null}
	 *            sein)
	 * @throws java.lang.Exception
	 *             bei irgendeinem Fehler TODO
	 */
	public static GridCoverage2D readGridFromArcInfoASCII(File file,
			CoordinateReferenceSystem crs) throws Exception {
		return readGridFromArcInfoASCII(file.toURI().toURL(), crs);

		// ArcGridRaster reader = new ArcGridRaster( new BufferedReader( new
		// InputStreamReader( new FileInputStream(file) ) ), false );
		// WritableRaster raster = reader.readRaster();
		// if (crs == null)
		// crs = determineProjection(file);
		//
		// //
		// System.out.println("Position  "+reader.getXlCorner()+" / "+reader.getYlCorner());
		// //
		// System.out.println("Size      "+reader.getNCols()+" / "+reader.getNRows());
		// //
		// System.out.println("Min/Max   "+reader.getMinValue()+" / "+reader.getMaxValue());
		// // System.out.println("NoData    "+reader.getNoData());
		// // System.out.println("CellSize  "+reader.getCellSize());
		//
		// float x = (float) reader.getXlCorner(); // Suedwestliche Ecke!
		// float y = (float) reader.getYlCorner(); // Suedwestliche Ecke!
		// float w = (float) (reader.getNCols() * reader.getCellSize()); //
		// reale Breite = RasterSpalten * Aufloesung
		// float h = (float) (reader.getNRows() * reader.getCellSize()); //
		// reale Hoehe = RasterZeilen * Aufloesung
		// Envelope2D envelope = new Envelope2D(crs, new Rectangle2D.Float(x, y,
		// w, h));
		//
		// // WICHTIG: Name des Rasters sollte Leer-String sein, da ansonsten
		// das
		// // Coloring des Rasters nicht klappt.
		// // --> Name der Categories muss (warum auch immer) mit dem Namen
		// // des Rasters uebereinstimmen!!!
		// return new GridCoverageFactory().create("", raster, envelope);
		// //
		// //=== OHNE ArcGridRaster ===
		// //// ArcGridReader reader = new ArcGridReader( new BufferedReader(
		// new InputStreamReader( new FileInputStream(file) ) ));
		// // ArcGridReader reader = new ArcGridReader( file );
		// // return (GridCoverage2D)reader.read(null);
	}

	/**
	 * Diese Methode importiert ein Raster aus einer Datei im
	 * ArcInfoASCII-Grid-Format. Wenn kein CRS angegeben wird, wird versucht das
	 * CRS aus einem prj-File (EPSG-Code "EPSG:..." oder WKT-Definition) zu
	 * lesen. Ist dies nicht erfolgreich, wird {@link #DEFAULT_CRS} als CRS
	 * verwendet.
	 * 
	 * @param url
	 *            Link to Arc/Info GRID ASCII file
	 * @param crs
	 *            CoordinateReferenceSystem fuer das Raster (kann {@code null}
	 *            sein)
	 * @throws java.lang.Exception
	 *             bei irgendeinem Fehler
	 */
	public static GridCoverage2D readGridFromArcInfoASCII(URL url,
			CoordinateReferenceSystem crs) throws Exception {
		ArcGridRaster reader = new ArcGridRaster(new BufferedReader(
				new InputStreamReader(url.openStream())), false);
		WritableRaster raster = reader.readRaster();
		if (crs == null)
			crs = determineProjection(IOUtil.changeUrlExt(url, "prj"));

		LOGGER.debug("Position  " + reader.getXlCorner() + " / "
				+ reader.getYlCorner());
		LOGGER.debug("Size      " + reader.getNCols() + " / "
				+ reader.getNRows());
		LOGGER.debug("Min/Max   " + reader.getMinValue() + " / "
				+ reader.getMaxValue());
		LOGGER.debug("NoData    " + reader.getNoData());
		LOGGER.debug("CellSize  " + reader.getCellSize());

		Double x = reader.getXlCorner(); // Suedwestliche Ecke!
		Double y = reader.getYlCorner(); // Suedwestliche Ecke!
		if (y.isNaN() || x.isNaN()) {
			throw new IllegalArgumentException(
					"The parameters xlCorner and/or ylCorner could not be read. Hint: Check if the file defines 'xlcenter' instead of 'xlcorner'. If so, just replace all 'center' with 'corner' and try again.");
		}

		Double w = (reader.getNCols() * reader.getCellSize()); // reale Breite =
		// RasterSpalten
		// * Aufloesung
		Double h = (reader.getNRows() * reader.getCellSize()); // reale Hoehe =
		// RasterZeilen
		// * Aufloesung
		Envelope2D envelope = new Envelope2D(crs, new Rectangle2D.Double(x, y,
				w, h));

		// WICHTIG: Name des Rasters sollte Leer-String sein, da ansonsten das
		// Coloring des Rasters nicht klappt.
		// --> Name der Categories muss (warum auch immer) mit dem Namen
		// des Rasters uebereinstimmen!!!
		return new GridCoverageFactory().create("", raster, envelope);
		//
		// === OHNE ArcGridRaster ===
		// // ArcGridReader reader = new ArcGridReader( new BufferedReader( new
		// InputStreamReader( new FileInputStream(file) ) ));
		// ArcGridReader reader = new ArcGridReader( file );
		// return (GridCoverage2D)reader.read(null);
	}

	/**
	 * Diese Methode erzeugt einen {@link AbstractGridCoverage2DReader} aus
	 * einer Datei im GeoTIFF-Format. Zunaechst wird versucht, die
	 * Geo-Informationen (Referenz+CRS) aus den TIFF-Metadaten zu ermitteln. Ist
	 * dies nicht erfolgreich, werden ein gleichnamiges World-File (.tfw) und
	 * Projection-File (.prj) herangezogen (siehe {@link WorldImageReader}).
	 * 
	 * @param file
	 *            GeoTIFF-File
	 * @param crs
	 *            erzwungenes CoordinateReferenceSystem fuer das Raster (kann
	 *            {@code null} sein)
	 * @throws java.lang.Exception
	 *             bei irgendeinem Fehler
	 * @deprecated {@link WorldImageReader} klappt noch nicht so 100%ig
	 *             (Colorisierung schlaegt fehl!)
	 */
	public static AbstractGridCoverage2DReader createGridReaderFromGeoTiff(
			File file, CoordinateReferenceSystem crs) throws Exception {
		// Versuchen Geo-Information aus Tiff zu lesen
		try {
			Hints hints = new Hints(null);
			// Wenn CRS angegeben, dieses durch Hint erzwingen
			if (crs != null)
				hints.put(Hints.DEFAULT_COORDINATE_REFERENCE_SYSTEM, crs);
			// Reader (mit Metadaten) erzeugen
			GeoTiffReader reader = new GeoTiffReader(file, hints);
			// Wenn kein Referenzsystem vorhanden, versuchen ein prj-File zu
			// verwenden
			if (reader.getOriginalEnvelope().getCoordinateReferenceSystem() == null) {
				LOGGER.warn("No projection information found in '"
						+ file.getName() + "'. Using prj-file...");
				hints.put(Hints.DEFAULT_COORDINATE_REFERENCE_SYSTEM,
						determineProjection(file));
				reader = new GeoTiffReader(file, hints);
			}
			return reader;
		} catch (UnsupportedOperationException err) {
			// erwartet, wenn keine Geo-Information im Tiff vorhanden ist
		} catch (NullPointerException err) {
			// erwartet, wenn keine Geo-Information im Tiff vorhanden ist
		}

		// keine Geo-Informationen in Tiff
		// --> Raster ueber ImageIO einlesen und WorldFile/ProjectionFile
		// verwenden
		LOGGER.warn("No geo information found in '" + file.getName()
				+ "'. Using world- and prj-file...");
		// @todo check the work of WorldImageReader with colorization!
		throw new UnsupportedOperationException(
				"WorldImageReader does not work well, yet... AbstractGridCoverage2DReader can not be created from this Tiff!");
		// Hints hints = new Hints(null);
		// // Wenn CRS angegeben, dieses durch Hint erzwingen
		// if ( crs != null )
		// hints.put( Hints.DEFAULT_COORDINATE_REFERENCE_SYSTEM, crs );
		// WorldImageReader reader = new WorldImageReader(file);
		// return reader;
	}

	/**
	 * Diese Methode erzeugt einen {@link AbstractGridCoverage2DReader} aus
	 * einer Datei im GeoTIFF-Format.
	 * 
	 * @param file
	 *            GeoTIFF-File
	 * @throws java.lang.Exception
	 *             bei irgendeinem Fehler
	 * @see #createGridReaderFromGeoTiff(File, CoordinateReferenceSystem)
	 * @deprecated {@link WorldImageReader} klappt noch nicht so 100%ig
	 *             (Colorisierung schlaegt fehl!)
	 */
	public static AbstractGridCoverage2DReader createGridReaderFromGeoTiff(
			File file) throws Exception {
		return createGridReaderFromGeoTiff(file, null);
	}

	/**
	 * Diese Methode importiert ein Raster aus einer Datei im GeoTIFF-Format.
	 * 
	 * @param file
	 *            GeoTIFF-File
	 * @param crs
	 *            erzwungenes CoordinateReferenceSystem fuer das Raster (kann
	 *            {@code null} sein)
	 * @throws java.lang.Exception
	 *             bei irgendeinem Fehler
	 * @see #createGridReaderFromGeoTiff(File, CoordinateReferenceSystem)
	 */
	public static GridCoverage2D readGridFromGeoTiff(File file,
			CoordinateReferenceSystem crs) throws Exception {
		// return (GridCoverage2D)createGridReaderFromGeoTiff(file,
		// crs).read(null);
		GridCoverage2D gc = null;
		// Versuchen Geo-Information aus Tiff zu lesen
		try {
			Hints hints = new Hints(null);

			// Wenn CRS angegeben, dieses durch Hint erzwingen
			if (crs != null)
				hints.put(Hints.DEFAULT_COORDINATE_REFERENCE_SYSTEM, crs);
			// Reader (mit Metadaten) erzeugen
			GeoTiffReader reader = new GeoTiffReader(file, hints);
			// Wenn kein Referenzsystem vorhanden, versuchen ein prj-File zu
			// verwenden
			if (reader.getOriginalEnvelope().getCoordinateReferenceSystem() == null) {
				LOGGER.warn("No projection information found in '"
						+ file.getName() + "'. Using prj-file...");
				hints.put(Hints.DEFAULT_COORDINATE_REFERENCE_SYSTEM,
						determineProjection(file));
				reader = new GeoTiffReader(file, hints);
			}
			gc = (GridCoverage2D) reader.read(null);
			return gc;
		} catch (UnsupportedOperationException err) {
			// erwartet, wenn keine Geo-Information im Tiff vorhanden ist
			// ExceptionMonitor.show(null, err,
			// "No geoinformation in TIFF, importing it as a normal file. Trying to read it as normal TIFF + Worldfile.");
		} catch (NullPointerException err) {
			// erwartet, wenn keine Geo-Information im Tiff vorhanden ist
		}

		// keine Geo-Informationen in Tiff
		// --> Raster ueber ImageIO einlesen und WorldFile/ProjectionFile
		// verwenden
		LOGGER.warn("No geo information found in '" + file.getName()
				+ "'. Using world- and prj-file...");
		BufferedImage im = ImageIO.read(file);
		if (im == null)
			throw new IIOException("No image reader found for this image type!");
		// World-File einlesen
		File tfw = IOUtil.changeFileExt(file, "tfw");
		double[] tfwInfo = readWorldFile(tfw);
		float w = (float) (im.getWidth() * tfwInfo[0]); // reale Breite =
		// RasterSpalten * hor.
		// Aufloesung
		float h = (float) (im.getHeight() * (-tfwInfo[3])); // reale Hoehe =
		// RasterZeilen *
		// vert. Aufloesung
		float x = (float) tfwInfo[4]; // Suedwestliche Ecke!
		float y = (float) tfwInfo[5] - h; // Suedwestliche Ecke (im tfw-File
		// steht die Nordwestliche!)
		// ggf. Projektion einlesen
		if (crs == null)
			crs = determineProjection(file);
		Envelope2D envelope = new Envelope2D(crs, new Rectangle2D.Float(x, y,
				w, h));
		// WICHTIG: Name des Rasters sollte Leer-String sein, da ansonsten das
		// Coloring des Rasters nicht klappt.
		// --> Name der Categories muss (warum auch immer) mit dem Namen
		// des Rasters uebereinstimmen!!!
		gc = new GridCoverageFactory().create("", im.copyData(null), envelope);
		return gc;
	}

	/**
	 * Diese Methode importiert ein Raster aus einer Datei im GeoTIFF-Format.
	 * 
	 * @param file
	 *            GeoTIFF-File
	 * @throws java.lang.Exception
	 *             bei irgendeinem Fehler
	 * @see #readGridFromGeoTiff(File, CoordinateReferenceSystem)
	 */
	public static GridCoverage2D readGridFromGeoTiff(File file)
			throws Exception {
		return readGridFromGeoTiff(file, null);
	}

	/**
	 * Diese Methode importiert ein Raster aus einer Datei im
	 * ArcInfoASCII-Grid-Format. Das CRS wird aus einem prj-File (EPSG-Code
	 * "EPSG:..." oder WKT-Definition) gelesen. Ist dies nicht erfolgreich, wird
	 * {@link #DEFAULT_CRS} als CRS verwendet.
	 * 
	 * @param file
	 *            ASCII-File
	 * @throws java.lang.Exception
	 *             bei irgendeinem Fehler
	 */
	public static WritableGridRaster readGridRasterFromArcInfoASCII(File file)
			throws Exception {
		return readGridRasterFromArcInfoASCII(file, null);
	}

	/**
	 * Diese Methode importiert ein Raster aus einer Datei im
	 * ArcInfoASCII-Grid-Format. Wenn kein CRS angegeben wird, wird versucht das
	 * CRS aus einem prj-File (EPSG-Code "EPSG:..." oder WKT-Definition) zu
	 * lesen. Ist dies nicht erfolgreich, wird {@link #DEFAULT_CRS} als CRS
	 * verwendet.
	 * 
	 * @param file
	 *            ASCII-File
	 * @param crs
	 *            CoordinateReferenceSystem fuer das Raster (kann {@code null}
	 *            sein)
	 * @throws java.lang.Exception
	 *             bei irgendeinem Fehler
	 */
	public static WritableGridRaster readGridRasterFromArcInfoASCII(File file,
			CoordinateReferenceSystem crs) throws Exception {
		ArcGridRaster reader = new ArcGridRaster(new BufferedReader(
				new InputStreamReader(new FileInputStream(file))), false);
		WritableRaster raster = reader.readRaster();
		float x = (float) reader.getXlCorner(); // Suedwestliche Ecke!
		float y = (float) reader.getYlCorner(); // Suedwestliche Ecke!
		float w = (float) (reader.getNCols() * reader.getCellSize()); // reale
		// Breite
		// =
		// RasterSpalten
		// *
		// Aufloesung
		float h = (float) (reader.getNRows() * reader.getCellSize()); // reale
		// Hoehe
		// =
		// RasterZeilen
		// *
		// Aufloesung
		// Projektion einlesen
		if (crs == null)
			crs = determineProjection(file);
		return new WritableGridRaster(raster,
				new Rectangle2D.Float(x, y, w, h), crs);

		// === OHNE ArcGridRaster ===
		// GridCoverage2D grid = readGridFromArcInfoASCII(null,input);
		// WritableRaster raster = grid.getRenderedImage().copyData(null);
		// return( new WritableGridRaster(raster,grid.getEnvelope2D()) );
	}

	/**
	 * Diese Methode importiert ein Raster aus einer Datei im GeoTIFF-Format.
	 * Zunaechst wird versucht, die Geo-Informationen (Referenz+CRS) aus den
	 * TIFF-Metadaten zu ermitteln. Ist dies nicht erfolgreich, werden ein
	 * gleichnamiges World-File (.tfw) und Projection-File (.prj) herangezogen.
	 * Als Projektion (.prj) ist sowohl ein EPSG-Code "EPSG:...", als auch eine
	 * WKT-Definition erlaubt. Kann kein CRS ermitteln werden, wird
	 * {@link #DEFAULT_CRS} als CRS verwendet.
	 * 
	 * @param file
	 *            GeoTIFF-File
	 * @throws java.lang.Exception
	 *             bei irgendeinem Fehler
	 */
	public static WritableGridRaster readGridRasterFromGeoTiff(File file)
			throws Exception {
		return readGridRasterFromGeoTiff(file, null);
	}

	/**
	 * Diese Methode importiert ein Raster aus einer Datei im GeoTIFF-Format.
	 * Zunaechst wird versucht, die Geo-Informationen (Referenz+CRS) aus den
	 * TIFF-Metadaten zu ermitteln. Ist dies nicht erfolgreich, werden ein
	 * gleichnamiges World-File (.tfw) und Projection-File (.prj) herangezogen.
	 * Als Projektion (.prj) ist sowohl ein EPSG-Code "EPSG:...", als auch eine
	 * WKT-Definition erlaubt. Kann kein CRS ermitteln werden, wird
	 * {@link #DEFAULT_CRS} als CRS verwendet.
	 * 
	 * @param file
	 *            GeoTIFF-File
	 * @param crs
	 *            erzwungenes CoordinateReferenceSystem fuer das Raster (kann
	 *            {@code null} sein)
	 * @throws java.lang.Exception
	 *             bei irgendeinem Fehler
	 */
	public static WritableGridRaster readGridRasterFromGeoTiff(File file,
			CoordinateReferenceSystem crs) throws Exception {
		// // GeoTiff-File einlesen
		// BufferedImage im = ImageIO.read( file );
		// if ( im == null )
		// throw new IIOException("No image reader found for this image type!");
		// // World-File einlesen
		// File tfw = IOUtil.changeFileExt(file,"tfw");
		// double[] tfwInfo = readWorldFile(tfw);
		// // ggf. Projektion einlesen
		// if (crs == null)
		// crs = determineProjection(file);
		//
		// float w = (float) (im.getWidth() * tfwInfo[0]); // reale Breite =
		// RasterSpalten * hor. Aufloesung
		// float h = (float) (im.getHeight() * (-tfwInfo[3])); // reale Hoehe =
		// RasterZeilen * vert. Aufloesung
		// float x = (float) tfwInfo[4]; // Suedwestliche Ecke!
		// float y = (float) tfwInfo[5] - h ; // Suedwestliche Ecke (im tfw-File
		// steht die Nordwestliche!)
		//
		// Rectangle2D env = new Rectangle2D.Float(x,y,w,h);
		// WritableGridRaster raster = new WritableGridRaster(
		// im.getData().getDataBuffer().getDataType(),
		// 0,
		// 0,
		// im.getWidth(),
		// im.getHeight(),
		// env,
		// crs
		// );
		//

		GridCoverage2D gc = readGridFromGeoTiff(file, crs);
		Rectangle2D env = gc.getEnvelope2D();
		RenderedImage im = gc.getRenderedImage();
		WritableGridRaster raster = new WritableGridRaster(im.getData()
				.getDataBuffer().getDataType(), 0, 0, im.getWidth(), im
				.getHeight(), env, crs);

		im.copyData(raster);

		return raster;
	}

	/**
	 * Liest ein World-File (.tfw) ein und liefert die darin zeilenweise
	 * gespeicherten Werte zurueck. Leer- und Kommentarzeilen werden dabei
	 * ignoriert. Kann ein Zeilen-Wert nicht in einen <code>double</code>
	 * umgewandelt werden, wird diese Zeile ignoriert, als 1.0 interpretiert und
	 * eine Meldung in die Standard-Fehler-Ausgabe geschrieben.<br>
	 * Der zurueckgegebene Array hat mindestens die Groesse 6:<br>
	 * <ol>
	 * <li>Zellengroesse in X-Richtung (in Meter)</li>
	 * <li>Koeffizient fuer Rotation in Y-Richtung</li>
	 * <li>Koeffizient fuer Rotation in X-Richtung</li>
	 * <li>negative Zellengroesse in Y-Richtung (in Meter)</li>
	 * <li>Horizontale Geo-Referenz (Longitude) der nord-westlichen Ecke</li>
	 * <li>Vertikale Geo-Referenz (Latitude) der nord-westlichen Ecke</li>
	 * </ol>
	 * 
	 * @todo WorldFile-Spec vervollstaendigen
	 * @param file
	 *            World-File
	 * @exception IOException
	 *                falls die Datei nicht existiert oder ein unerwarteter
	 *                Fehler beim Einlesen auftritt
	 * @see IOUtil#isCommentLine(String)
	 * @see System#err
	 */

	public static double[] readWorldFile(File file) throws IOException {
		return readWorldFile(new FileInputStream(file));
	}

	/**
	 * Liest ein World-File (.tfw) ein und liefert die darin zeilenweise
	 * gespeicherten Werte zurueck. Leer- und Kommentarzeilen werden dabei
	 * ignoriert. Kann ein Zeilen-Wert nicht in einen <code>double</code>
	 * umgewandelt werden, wird diese Zeile ignoriert, als 1.0 interpretiert und
	 * eine Meldung in die Standard-Fehler-Ausgabe geschrieben.<br>
	 * Der zurueckgegebene Array hat mindestens die Groesse 6:<br>
	 * <ol>
	 * <li>Zellengroesse in X-Richtung (in Meter)</li>
	 * <li>Koeffizient fuer Rotation in Y-Richtung</li>
	 * <li>Koeffizient fuer Rotation in X-Richtung</li>
	 * <li>negative Zellengroesse in Y-Richtung (in Meter)</li>
	 * <li>Horizontale Geo-Referenz (Longitude) der nord-westlichen Ecke</li>
	 * <li>Vertikale Geo-Referenz (Latitude) der nord-westlichen Ecke</li>
	 * </ol>
	 * 
	 * @todo WorldFile-Spec vervollstaendigen
	 * 
	 * @param inputStream
	 *            Stream of World-File
	 * @exception IOException
	 *                falls die Datei nicht existiert oder ein unerwarteter
	 *                Fehler beim Einlesen auftritt
	 * @see IOUtil#isCommentLine(String)
	 * @see System#err
	 * 
	 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
	 *         (University of Bonn/Germany)
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public static double[] readWorldFile(InputStream inputStream)
			throws IOException {
		Vector<Double> tfwValues = new Vector<Double>();
		BufferedReader in = new BufferedReader(new InputStreamReader(
				inputStream));
		String line = null;
		for (int lineNo = 1; in.ready()
				&& (line = in.readLine().trim()) != null; lineNo++) {
			if (IOUtil.isCommentLine(line) || line.equals(""))
				continue;
			try {
				StringTokenizer st = new StringTokenizer(line);
				tfwValues.add(Double.parseDouble(st.nextToken(" \n;,#|/")));
			} catch (Exception err) {
				System.err.println("WorldFile-Error in line " + lineNo + ": "
						+ err);

				// (SK)
				// System.err.println("WorldFile-Error in '"+file.getAbsolutePath()+"' (line "+lineNo+"): "+err);
				System.err.println("   >> expected value is set to 1.0");
				tfwValues.add(1.0);
			}
		}

		while (tfwValues.size() < 6) {
			System.err.println("WorldFile-Error: value " + tfwValues.size()
					+ " missing!");
			// (SK)
			// System.err.println("WorldFile-Error in '"+file.getAbsolutePath()+"': value "+tfwValues.size()+" missing!");
			System.err.println("   >> expected value is set to 1.0");
			tfwValues.add(1.0);
		}

		// Vector in Array umwandeln
		double[] ret = new double[tfwValues.size()];
		for (int i = 0; i < ret.length; i++)
			ret[i] = tfwValues.elementAt(i);
		return ret;
	}

	/**
	 * Liest das CRS aus einer Datei. Zunaechst wird versucht das CRS aus der
	 * uebergebene Datei zu lesen. Ist dies nicht erfolgreich wird das zur Datei
	 * korrespondierende prj-File herangezogen.
	 * 
	 * @param file
	 *            Datei
	 * @return {@code null}, wenn kein CRS gelesen werden konnte
	 */
	public static CoordinateReferenceSystem determineProjection(File file)
			throws IOException {

		// ****************************************************************************
		// schonmal teilweise migriert... das geht direkt an die URL methode
		// weiter...
		// ****************************************************************************

		// CoordinateReferenceSystem crs = null;
		// // Versuchen CRS aus angegebener Datei zu lesen
		// crs = readProjectionFile(file);
		// // Wenn nicht erfolgreich versuchen die zur Datei korrespondierende
		// // prj-Datei einzulesen
		// if ( crs == null )
		// crs = readProjectionFile( IOUtil.changeFileExt(file,"prj") );
		// // Wenn nicht erfolgreich, Default verwenden
		// if ( crs == null ) {
		// LOGGER.warn("No projection found for file '"+file.getName()+"'. Default CRS used.");
		// crs = DEFAULT_CRS;
		// }
		return determineProjection(file.toURI().toURL());
	}

	/**
	 * Liest das CRS aus einer URL. Wenn keine CRS gelesen werden kann, dann
	 * wird die Endung der URL gegen .prj gewechselt und nochmal versucht bevor
	 * das {@link #DEFAULT_CRS} benutzt wird.
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public static CoordinateReferenceSystem determineProjection(URL prjUrl) {
		CoordinateReferenceSystem crs = null;
		// Versuchen CRS aus angegebener Datei zu lesen

		LOGGER.debug("determineProjection: Try to read a projection from URL "
				+ prjUrl);
		try {
			crs = GeoImportUtil.readProjectionFile(prjUrl);
		} catch (IOException e) {
		}

		if (crs == null) {
			try {
				crs = GeoImportUtil.readProjectionFile(IOUtil.changeUrlExt(
						prjUrl, "prj"));
			} catch (IllegalArgumentException e) {
			} catch (IOException e) {
			}
		}

		// Wenn nicht erfolgreich, Default verwenden
		if (crs == null) {
			LOGGER.warn("No projection found in URL. Default CRS used.");
			crs = getDEFAULT_CRS();
		}
		return crs;
	}

	/**
	 * Liest ein CRS aus einer URL. Als Dateiinhalt wird entweder ein EPSG-Code
	 * oder eine WKT-Definition des CRS akzeptiert.
	 * 
	 * @param prjURL
	 *            prj-Datei in der die Projektion hinterlegt ist
	 * @return {@code null}, wenn kein CRS gelesen werden konnte
	 * @see GTUtil#createCRS(String)
	 * 
	 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
	 *         (University of Bonn/Germany)
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 * @throws IOException
	 *             falls die URL nicht wie erwartet gelesen werden konnte.
	 */
	public static CoordinateReferenceSystem readProjectionFile(URL prjURL)
			throws IOException {
		String crsDefinition = readProjectionString(prjURL);
		CoordinateReferenceSystem crs = GTUtil.createCRS(crsDefinition);
		return crs;
	}

	/**
	 * Liest ein CRS aus einer Datei. Als Dateiinhalt wird entweder ein
	 * EPSG-Code oder eine WKT-Definition des CRS akzeptiert.
	 * 
	 * @param prjFile
	 *            prj-Datei in der die Projektion hinterlegt ist
	 * @return {@code null}, wenn kein CRS gelesen werden konnte
	 * @see GTUtil#createCRS(String)
	 * 
	 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
	 *         (University of Bonn/Germany)
	 */
	public static CoordinateReferenceSystem readProjectionFile(File prjFile)
			throws IOException {
		String crsDefinition = readProjectionString(prjFile);
		CoordinateReferenceSystem crs = GTUtil.createCRS(crsDefinition);
		// if ( crs == null )
		// LOGGER.warn("CRS can not be read from file '"+(prjFile == null ?
		// "null" : prjFile.getName())+"'");
		return crs;
	}

	/**
	 * Liest die ersten Zeilen der angegebenen prj-Datei und fuegt sie zu einem
	 * String zusammen. Wenn die Datei zu gross ist, wird ein Leerstring
	 * zurueckgegeben, da es sich dann nicht um eine Projektionsdefinition
	 * handelt.
	 * 
	 * @param prjFile
	 *            prj-Datei
	 * @return Leerstring, falls die Datei nicht existiert oder groesser als
	 *         10KB ist
	 * 
	 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
	 *         (University of Bonn/Germany)
	 */
	private static String readProjectionString(File prjFile) throws IOException {
		// Wenn Datei groesser als 10KB, dann ist es kein prj-File
		if (!prjFile.exists() || !prjFile.isFile() || prjFile.length() == 0
				|| prjFile.length() > 10240)
			return "";

		// Alle Zeilen der Datei aneinander haengen
		BufferedReader input = new BufferedReader(new InputStreamReader(
				new FileInputStream(prjFile)));
		StringBuffer buffer = new StringBuffer();
		String line = null;
		while (input.ready() && (line = input.readLine()) != null)
			if (!IOUtil.isCommentLine(line))
				buffer.append(" ").append(line);
		input.close();
		return buffer.toString().trim();

	}

	/**
	 * Liest alle Zeilen aus der prj-Datei URL und fuegt diese zu einem String
	 * zusammen
	 * 
	 * Kommentarzeilen werden ignoriert.
	 * 
	 * Wenn mehr als 3000 Zeichen (ohne Kommentare) gelesen werden, wird davon
	 * ausgegangen, dass es sich nicht um eine .PRJ Datei handelt, und "" wird
	 * zurückgegeben.
	 * 
	 * @param url
	 *            {@link URL} auf prj-Datei
	 * 
	 * @return Leerstring, wenn url == null
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public static String readProjectionString(URL url) throws IOException {
		if (url == null) {
			LOGGER
					.debug("readProjectionString returned empty String for url==null");
			return "";
		}

		BufferedReader input = new BufferedReader(new InputStreamReader(url
				.openStream()));
		try {
			// Alle Zeilen der Datei aneinander haengen
			StringBuffer buffer = new StringBuffer();
			String line = null;
			while (input.ready() && (line = input.readLine()) != null) {
				if (!IOUtil.isCommentLine(line))
					buffer.append(" ").append(line);
				if (buffer.length() > 3000) {
					// LOGGER.warn("The URL " + url +
					// " doesn't seem point to a .prj file, because size is > 3000 characters. Returning \"\"");
					return "";
				}
			}
			final String prjString = buffer.toString().trim();
			// LOGGER.debug("This is the PRJ String from URL = \n"+prjString);
			return prjString;
		} finally {
			input.close();
		}
	}

	public static void setDEFAULT_CRS(CoordinateReferenceSystem dEFAULT_CRS) {
		DEFAULT_CRS = dEFAULT_CRS;
	}

	public static CoordinateReferenceSystem getDEFAULT_CRS() {
		return DEFAULT_CRS;
	}

}
