/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.jfree;

import java.awt.Color;
import java.util.HashMap;
import java.util.Locale;
import java.util.Map;

import org.apache.log4j.Logger;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.axis.NumberAxis;
import org.jfree.chart.event.PlotChangeEvent;
import org.jfree.chart.event.PlotChangeListener;
import org.jfree.chart.event.RendererChangeEvent;
import org.jfree.chart.event.RendererChangeListener;
import org.jfree.chart.labels.StandardXYToolTipGenerator;
import org.jfree.chart.labels.XYToolTipGenerator;
import org.jfree.chart.plot.CategoryPlot;
import org.jfree.chart.plot.Plot;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.chart.plot.XYPlot;
import org.jfree.chart.renderer.AbstractRenderer;
import org.jfree.chart.renderer.category.AreaRenderer;
import org.jfree.chart.renderer.category.BarRenderer;
import org.jfree.chart.renderer.category.CategoryItemRenderer;
import org.jfree.chart.renderer.category.CategoryStepRenderer;
import org.jfree.chart.renderer.category.LineAndShapeRenderer;
import org.jfree.chart.renderer.category.StackedAreaRenderer;
import org.jfree.chart.renderer.category.StackedBarRenderer;
import org.jfree.chart.renderer.xy.StackedXYAreaRenderer2;
import org.jfree.chart.renderer.xy.StackedXYBarRenderer;
import org.jfree.chart.renderer.xy.XYAreaRenderer;
import org.jfree.chart.renderer.xy.XYBarRenderer;
import org.jfree.chart.renderer.xy.XYItemRenderer;
import org.jfree.chart.renderer.xy.XYLineAndShapeRenderer;
import org.jfree.chart.renderer.xy.XYStepAreaRenderer;
import org.jfree.chart.renderer.xy.XYStepRenderer;
import org.jfree.chart.title.LegendTitle;
import org.jfree.chart.title.Title;
import org.jfree.chart.urls.StandardXYURLGenerator;
import org.jfree.chart.urls.XYURLGenerator;
import org.jfree.data.function.Function2D;
import org.jfree.data.function.LineFunction2D;
import org.jfree.data.general.Dataset;
import org.jfree.data.general.DatasetUtilities;
import org.jfree.data.statistics.Regression;
import org.jfree.data.xy.XYDataset;
import org.jfree.data.xy.XYSeriesCollection;

import schmitzm.jfree.chart.renderer.SelectionRenderer;
import schmitzm.jfree.chart.selection.DatasetSelectionModel;
import schmitzm.jfree.chart.style.BasicChartStyle;
import schmitzm.jfree.chart.style.ChartStyle;
import schmitzm.jfree.chart.style.ChartStyleXMLFactory;
import schmitzm.jfree.chart.style.ChartType;
import schmitzm.lang.LangUtil;
import schmitzm.lang.ResourceProvider;

/**
 * In dieser Klasse sind Hilfsmethoden fuer JFreeChart hinterlegt.
 * @author <a href="mailto:Martin.Schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class JFreeChartUtil {
  private static Logger LOGGER = Logger.getLogger(JFreeChartUtil.class.getName());
  
  /** {@link ResourceProvider}, der die Lokalisation fuer GUI-Komponenten
   *  zur Verfuegung stellt. Diese sind in properties-Datein unter
   *  {@code schmitzm.jfree.resource.locales.JFreeResourceBundle_XXX.properties}
   *  hinterlegt. */
  public static ResourceProvider RESOURCE = new ResourceProvider( LangUtil.extendPackagePath(JFreeChartUtil.class,"resource.locales.JFreeResourceBundle"), Locale.ENGLISH );

  /**
   * Convenience method to access the {@link ResourceProvider}.
   */
  public static String R(String key, Object... values) {
    return RESOURCE.getString(key, values);
  }

  /** Instance of {@link ChartStyleXMLFactory}. */
  public static final ChartStyleXMLFactory<ChartStyle> CHART_STYLE_FACTORY = new ChartStyleXMLFactory<ChartStyle>();

  /**
   * Sets all chart legends (in)visible.
   * @param chart   a chart to apply
   * @param visible indicates the visible property
   */
  public static void setLegendVisible(JFreeChart chart, boolean visible) {
    for (Object title : chart.getSubtitles())
      if ( title instanceof LegendTitle )
        ((Title)title).setVisible(visible);
  }
  
  /**
   * Fires a change event to all {@link PlotChangeListener} connected
   * to the plot. 
   * @param plot a plot
   */
  public static void fireChangeEvent(Plot plot) {
    plot.notifyListeners( new PlotChangeEvent(plot) );
  }
  
  /**
   * Fires a change event to all {@link RendererChangeListener} connected
   * to the renderer. 
   * @param renderer a renderer
   */
  public static void fireChangeEvent(AbstractRenderer renderer) {
    renderer.notifyListeners( new RendererChangeEvent(renderer,true) );
  }

  /**
   * Creates a {@link XYItemRenderer} for the given stepped and stacked
   * properties.
   * @param type    chart type
   * @param stepped stepped property
   * @param stacked stacked property
   */
  public static XYItemRenderer createXYRenderer(BasicChartStyle style) {
    return createXYRenderer(
        style.getType(),
        style.isStepped(),
        style.isStacked()
    );
  }
  
  /**
   * Creates a {@link XYItemRenderer} for the given stepped and stacked
   * properties.
   * @param type    chart type
   * @param stepped stepped property
   * @param stacked stacked property
   */
  public static XYItemRenderer createXYRenderer(ChartType type, boolean stepped, boolean stacked) {
    switch ( type ) {
      case LINE: // "stacked" is not available for line charts
                 if ( stacked )
                   LOGGER.warn("Stacked property can not be applied to line charts. Property ignored!");
                 // Create renderer according to "stepped"
                 if ( stepped )
                   return new XYStepRenderer();
                 // normal line renderer
                 return new XYLineAndShapeRenderer(true, false);
      case AREA: // Apply stacked
                 if ( stacked ) {
                   // "stacked" and "stepped" can not be applied the same time
                   if ( stepped )
                     LOGGER.warn("A stacked area chart can not be stepped the same time (use a stacked bar chart instead). Stepped propery ignored!");
                   // Stacked area renderer
                   return new StackedXYAreaRenderer2();
                 }
                 // Stepped area renderer
                 if ( stepped )
                   return new XYStepAreaRenderer(XYStepAreaRenderer.AREA);
                 // Neither stepped nor stepped
                 return new XYAreaRenderer(XYAreaRenderer.AREA);
      case BAR: // bar charts are automatically stepped,
                // so only "stacked" is applied
                if ( stacked )
                  return new StackedXYBarRenderer();
                // Normal bar renderer
                return new XYBarRenderer();
    }
    throw new UnsupportedOperationException("XYItemRenderer could not be created for chart type: "+type); 
  }

  
  /**
   * Creates a {@link CategoryItemRenderer} for the given stepped and stacked
   * properties.
   * @param type    chart type
   * @param stepped stepped property
   * @param stacked stacked property
   */
  public static CategoryItemRenderer createCategoryRenderer(BasicChartStyle style) {
    return createCategoryRenderer(
        style.getType(),
        style.isStepped(),
        style.isStacked()
    );
  }
  
  /**
   * Creates a {@link CategoryItemRenderer} for the given stepped and stacked
   * properties.
   * @param type    chart type
   * @param stepped stepped property
   * @param stacked stacked property
   */
  public static CategoryItemRenderer createCategoryRenderer(ChartType type, boolean stepped, boolean stacked) {
    switch ( type ) {
      case LINE: // "stacked" is not available for line charts
                 if ( stacked )
                   LOGGER.warn("Stacked property can not be applied to line charts. Property ignored!");
                 // Create renderer according to "stepped"
                 if ( stepped )
                   return new CategoryStepRenderer();
                 // normal line renderer
                 return new LineAndShapeRenderer(true, false);
      case AREA: // category area chart is automatically stepped,
                 // so only "stacked" is applied
                 if ( stacked )
                   return new StackedAreaRenderer(false);
                 // Normal area renderer
                 return new AreaRenderer();
      case BAR: // bar charts are automatically stepped,
                // so only "stacked" is applied
                if ( stacked )
                  return new StackedBarRenderer(false);
                // Normal bar renderer
                return new BarRenderer();
    }
    throw new UnsupportedOperationException("CategoryItemRenderer could not be created for chart type: "+type); 
  }

  /**
   * Returns the renderer for the {@code i}-th dataset of a plot.
   * @param plot  a plot
   * @param index the dataset index
   * @return the primary renderer if no specific renderer is
   *         set for dataset {@code i}
   */
  public static Object getRendererForDataset(Plot plot, int i) {
    Object renderer = null;
    if ( plot instanceof XYPlot ) {
      renderer = ((XYPlot)plot).getRenderer(i);
      if ( renderer == null )
        renderer = ((XYPlot)plot).getRenderer();
    }

    if ( plot instanceof CategoryPlot ) {
      renderer = ((CategoryPlot)plot).getRenderer(i);
      if ( renderer == null )
        renderer = ((CategoryPlot)plot).getRenderer();
    }
    
    return renderer;
  }
  
  /**
   * Searchs the renderers of a plot for {@link SelectionRenderer} for
   * a {@link Dataset}.
   * @param plot    the plot to check for renderer  
   * @param dataset the dataset a {@link SelectionRenderer} is
   *                searched for
   * @return {@code null} if there is no {@link SelectionRenderer} defined
   *         for the dataset
   */
  public static SelectionRenderer<?> getSelectionRendererForDataset(Plot plot, Dataset dataset) {
    if ( plot instanceof XYPlot ) {
      XYPlot xyPlot = (XYPlot)plot;
      for (int i=0; i<xyPlot.getDatasetCount(); i++) {
        Dataset d = xyPlot.getDataset(i);
        if ( d != dataset )
          continue;
        // determine renderer for dataset (if no renderer is set,
        // take the primary renderer)
        XYItemRenderer r = xyPlot.getRenderer(i);
        if ( r == null )
          r = xyPlot.getRenderer();
        // stop search, if renderer is a SelectionRenderer
        if ( r instanceof SelectionRenderer<?> )
          return (SelectionRenderer<?>)r;
      }
    }

    if ( plot instanceof CategoryPlot ) {
      CategoryPlot catPlot = (CategoryPlot)plot;
      for (int i=0; i<catPlot.getDatasetCount(); i++) {
        Dataset d = catPlot.getDataset(i);
        if ( d != dataset )
          continue;
        // determine renderer for dataset (if no renderer is set,
        // take the primary renderer)
        CategoryItemRenderer r = catPlot.getRenderer(i);
        if ( r == null )
          r = catPlot.getRenderer();
        // stop search, if renderer is a SelectionRenderer
        if ( r instanceof SelectionRenderer<?> )
          return (SelectionRenderer<?>)r;
      }
    }
    
    return null;
  }
 
  /**
   * Searchs the renderers of a plot for {@link SelectionRenderer} for
   * a {@link Dataset} and returns its {@link DatasetSelectionModel}.
   * @param plot    the plot to check for renderer  
   * @param dataset the dataset a {@link SelectionRenderer} is
   *                searched for
   * @return {@code null} if there is no {@link SelectionRenderer} defined
   *         for the dataset
   */
  public static DatasetSelectionModel<?,?,?> getSelectionModelForDataset(Plot plot, Dataset dataset) {
    SelectionRenderer<?> selRenderer = getSelectionRendererForDataset(plot, dataset);
    if ( selRenderer != null )
      return selRenderer.getSelectionModel();
    return null;
  }

  /**
   * Returns all {@link SelectionRenderer} for a {@link Plot}.
   * @param plot    the plot to check for renderer  
   * @return never {@code null}
   */
  public static Map<Integer,SelectionRenderer<?>> getSelectionRenderer(Plot plot) {
    Map<Integer,SelectionRenderer<?>> selRenderer = new HashMap<Integer,SelectionRenderer<?>>();
    
    if ( plot instanceof XYPlot ) {
      XYPlot xyPlot = (XYPlot)plot;
      for (int i=0; i<xyPlot.getRendererCount(); i++) {
        XYItemRenderer r = xyPlot.getRenderer(i);
        if ( r instanceof SelectionRenderer<?> )
          selRenderer.put(i, (SelectionRenderer<?>)r);
      }
    }

    if ( plot instanceof CategoryPlot ) {
      CategoryPlot catPlot = (CategoryPlot)plot;
      for (int i=0; i<catPlot.getRendererCount(); i++) {
        CategoryItemRenderer r = catPlot.getRenderer(i);
        if ( r instanceof SelectionRenderer<?> )
          selRenderer.put(i, (SelectionRenderer<?>)r);
      }
    }
    
    return selRenderer;
  }
  
  /**
   * Searchs the renderers of a plot for all {@link SelectionRenderer}
   * and returns their {@link DatasetSelectionModel DatasetSelectionModels}.
   * @param plot    the plot to check for renderer  
   * @return never {@code null}
   */
  public static Map<Integer,DatasetSelectionModel<?,?,?>> getSelectionModels(Plot plot) {
    Map<Integer,SelectionRenderer<?>>         selRenderer = getSelectionRenderer(plot);
    Map<Integer,DatasetSelectionModel<?,?,?>> selModels   = new HashMap<Integer, DatasetSelectionModel<?,?,?>>();
    for (int dataset : selRenderer.keySet())
      selModels.put(dataset, selRenderer.get(dataset).getSelectionModel());
    return selModels;
  }
  
  /**
   * Returns the (first) dataset of a {@link Plot} which is rendererd
   * by a renderer.
   * @param plot     the plot which holds the datasets and renderer
   * @param renderer the renderer a dataset is seached for
   * @return {@code null} if there is no dataset which is rendered
   *         by the given renderer 
   */
  public static Dataset getDatasetForRenderer(Plot plot, Object renderer) {
    if ( plot instanceof XYPlot ) {
      XYPlot xyPlot = (XYPlot)plot;
      for (int i=0; i<xyPlot.getRendererCount(); i++)
        if ( xyPlot.getRenderer(i) == renderer )
          return xyPlot.getDataset(i);
    }

    if ( plot instanceof CategoryPlot ) {
      CategoryPlot catPlot = (CategoryPlot)plot;
      for (int i=0; i<catPlot.getRendererCount(); i++)
        if ( catPlot.getRenderer(i) == renderer )
          return catPlot.getDataset(i);
    }
    
    return null;
  }

  /**
   * Applies general dataset properties to a {@link Dataset}.
   * @param dataset dataset the properties are applied to
   * @param sample  dataset the properties are taken from
   */
  public static void applyDatasetProperties(Dataset dataset, Dataset sample) {
    dataset.setGroup( sample.getGroup() );
  }
  
  /**
   * Creates a {@link JFreeChart} which shows a point for each data iten and
   * a the appropriate regression line.
   * @param dataset        the data
   * @param title          title for the chart
   * @param xAxisTitle     title for the X-Axis
   * @param yAxisTitle     title for the Y-Axis
   * @param regressionLine indicates whether the regression line is shown
   */
  public static JFreeChart createRegressionChart(XYDataset dataset, String title, String xAxisTitle, String yAxisTitle, boolean regressionLine) {
    if ( regressionLine && !(dataset instanceof XYSeriesCollection) )
      throw new UnsupportedOperationException("Regression line can only be created for XYSeriesCollection!");
    
    NumberAxis xAxis   = new NumberAxis(xAxisTitle);
    xAxis.setAutoRangeIncludesZero(false);
    NumberAxis yAxis   = new NumberAxis(yAxisTitle);
    yAxis.setAutoRangeIncludesZero(false);

    XYPlot             plot             = new XYPlot(dataset, xAxis, yAxis, null);
    XYToolTipGenerator toolTipGenerator = new StandardXYToolTipGenerator();
    XYURLGenerator     urlGenerator     = new StandardXYURLGenerator();
    XYItemRenderer     renderer         = new XYLineAndShapeRenderer(false, true);
    
    renderer.setBaseToolTipGenerator(toolTipGenerator);
    renderer.setURLGenerator(urlGenerator);
    plot.setRenderer(renderer);
    plot.setOrientation(PlotOrientation.VERTICAL);

    // Create Regression line 
    if ( regressionLine) {

    	// only if there are at least 2 items, otherwise we would get a IllegalArgumentException
    	if (((XYSeriesCollection)dataset).getItemCount(0)>=2) {
    		// create sample data for curve (plot function directly is not yet available)
    		XYDataset regressionData = createRegressionLineDataset((XYSeriesCollection)dataset,0,title+" (RegressionLine)",2);
    		addRegressionLineToPlot(plot, regressionData, Color.blue);
    	} else {
    		LOGGER.info("Not enought items in the series to create a regression line.");
    	}
    }
    JFreeChart chart = new JFreeChart(title, JFreeChart.DEFAULT_TITLE_FONT, plot, true);

    return chart; //ChartFactory.createScatterPlot(title, xAttr, yAttr, dataset, PlotOrientation.HORIZONTAL, true, true, true);
  }

  /**
   * Creates sample data for a regression line of an {@link XYDataset}.
   * @param dataset    data the regression line is created for
   * @param series     series of {@code dataset} the regression line is created for
   * @param seriesKey  the key for the created dataset
   * @param sampleCnt  count of created samples
   */
  public static XYDataset createRegressionLineDataset(XYSeriesCollection dataset, int series, Comparable<?> seriesKey, int sampleCount) {
    double minX = dataset.getSeries(0).getMinX();
    double maxX = dataset.getSeries(0).getMaxX();
    return createRegressionLineDataset(dataset,series,seriesKey,minX,maxX,sampleCount);
  }

  /**
   * Creates sample data for a regression line of an {@link XYDataset}.
   * @param dataset    data the regression line is created for
   * @param series     series of {@code dataset} the regression line is created for
   * @param seriesKey  the key for the created dataset
   * @param startX     the first value the sample data is created for
   * @param endX       the last value the sample data is created for
   * @param sampleCnt  count of created samples
   */
  public static XYDataset createRegressionLineDataset(XYDataset dataset, int series, Comparable<?> seriesKey, double startX, double endX, int sampleCount) {
    double[]   coefficients = Regression.getOLSRegression(dataset, series);
    Function2D curve        = new LineFunction2D(coefficients[0], coefficients[1]);
    return DatasetUtilities.sampleFunction2D(curve, startX, endX, 2, seriesKey);
  }

  /**
   * Adds a line plot for regression data to a plot. 
   * @param plot           the plot the line is added to (at the end)
   * @param regressionData the regression data
   * @param lineColor      the color for the regression line
   */
  public static void addRegressionLineToPlot(XYPlot plot, XYDataset regressionData, Color lineColor) {
    int datasetCount = plot.getDatasetCount();
    // set the regression data
    plot.setDataset(datasetCount, regressionData);
    // create and set the renderer for the regression line
    XYLineAndShapeRenderer renderer = new XYLineAndShapeRenderer(true,false);
    renderer.setSeriesPaint(0, Color.blue);
    plot.setRenderer(datasetCount, renderer);
  }



//###################
//  BISHER NUR WILDE TESTS
//###################
//  /**
//   * Converts a Excel file to CVS format.
//   * @param xlsInput  stream to Excel source
//   * @param sheetName name of the sheet in the Excel source (if {@code null} the
//   *                  first sheet is used)
//   * @param headRow   indicates whether the first row contains the column names
//   * @param csvHeader indicates whether the column names are also taken to the CVS (note:
//   *                  if {@code true} and {@code headRow = false} dummy column
//   *                  names "A", "B", ... are generates in CVS!)
//   * @param delimChr  character takes as delimiter in CSV  
//   * @param strEncChr character String values are encapsulated with (if {@code null}, Strings
//   *                  are not encapsulated)                  
//   */
//  public static InputStream createCSVFromXLS(InputStream xlsInput, String sheetName, boolean headRow, boolean csvHeader, String delimChr, String strEncChar) throws IOException {
//    PipedOutputStream csvOutput = new PipedOutputStream();
//   
//    // read XLS and create a CSV
//    HSSFWorkbook wb    = new HSSFWorkbook( new POIFSFileSystem(xlsInput) );
//    HSSFSheet    sheet = ( sheetName != null ) ? wb.getSheet(sheetName) : wb.getSheetAt(0);
//    
//    int firstRow = sheet.getFirstRowNum();
//    int lastRow  = sheet.getLastRowNum();
////    int rowCount = sheet.getPhysicalNumberOfRows();
//    for (int y=firstRow; y <= lastRow; y++) {
//      HSSFRow row = sheet.getRow(y);
//      int firstCol = row.getFirstCellNum();
//      int lastCol  = row.getLastCellNum();
//      for (int x=firstCol; x <= lastCol; x++) {
//        HSSFCell cell = row.getCell(x);
////        String csvValue = convertCellValueToString(cell,???);
//      }
//    }
//   
//    return new PipedInputStream(csvOutput);
//  }
//  
//  public static String convertCellValueToString(HSSFCell cell, HSSFFormulaEvaluator formulaEvaluator) {
//    String strValue = "";
//    switch( cell.getCellType() ) {
//      case HSSFCell.CELL_TYPE_BOOLEAN: strValue = ""+cell.getBooleanCellValue();
//                                       break;
//      case HSSFCell.CELL_TYPE_NUMERIC: strValue = ""+cell.getNumericCellValue();
//                                       break;
//      case HSSFCell.CELL_TYPE_FORMULA: strValue = ""+cell.getNumericCellValue();
//    }
//    return strValue;
//    
//  }
//
//  /**
//   * Creates a Dataset from Excel file.
//   * @param xlsInput  stream to Excel source
//   * @param sheetName name of the sheet in the Excel source (if {@code null} the
//   *                  first sheet is used)
//   */
//  public static CategoryDataset createDatasetFromXLS(InputStream xlsInput, String sheetName) throws IOException {
//    PipedOutputStream csvOutput = new PipedOutputStream();
//    InputStreamReader csvInput  = new InputStreamReader( new PipedInputStream(csvOutput) ); 
//    
//
////    return new CSV(';','"').readCategoryDataset(csvInput);
//
//    HSSFWorkbook wb = new HSSFWorkbook(new POIFSFileSystem(xlsInput));
//    ExcelExtractor extractor = new ExcelExtractor(wb);
//    
//    extractor.setFormulasNotResults(false);
//    extractor.setIncludeSheetNames(false);
//    String text = extractor.getText();
//    System.out.println(text);
//    return null;
//  }
}
