/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package skrueger;

import java.util.HashMap;
import java.util.Map;

import org.geotools.feature.AttributeType;

import schmitzm.geotools.feature.AttributeTypeFilter;

/**
 * Implements an {@link AttributeTypeFilter} using the
 * {@linkplain AttributeMetaData#isVisible() visible}-property of an
 * {@link AttributeMetaData} map (or array).<br>
 * If this filter is created from a {@code null} map or {@code null} array,
 * the filter accepts ALL attributes except geometries.
 * @see AttributeTypeFilter#NO_GEOMETRY
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class AttributeMetaDataAttributeTypeFilter implements AttributeTypeFilter {

  /** The meta data of a set of attributes */
  protected Map<Integer,AttributeMetaData> attrMetaDataMap = null;

  /**
   * Creates a new filter.
   * @param attrMetaData the meta data of some attributes
   */
  public AttributeMetaDataAttributeTypeFilter(AttributeMetaData[] attrMetaData) {
    this.attrMetaDataMap = new HashMap<Integer,AttributeMetaData>();
    for (int i=0; attrMetaData!=null && i<attrMetaData.length; i++)
      this.attrMetaDataMap.put(i, attrMetaData[i]);
  }

  /**
   * Creates a new filter.
   * @param attrMetaData the meta data of some attributes
   */
  public AttributeMetaDataAttributeTypeFilter(Map<Integer,AttributeMetaData> attrMetaData) {
    this.attrMetaDataMap = attrMetaData;
  }

  /**
   * Returns {@code true} if the attribute meta data at index {@code idx} is
   * visible and the attribute is no geometry at all.
   */
  public boolean accept(AttributeType type, int idx) {
    // if no meta data is given, all attributes (except Geometry)
    // are treaten as visible
    if ( attrMetaDataMap == null )
      return NO_GEOMETRY.accept(type, idx);

    AttributeMetaData metaData = attrMetaDataMap.get(idx);
    return NO_GEOMETRY.accept(type, idx)  // no geometry attributes at all
        && metaData != null               // meta data must be present for column
        && metaData.isVisible();          // attribute must be visible
  }
}
