/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package skrueger.geotools;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.net.URL;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;
import java.util.Random;

import javax.swing.ImageIcon;
import javax.swing.JPanel;

import org.apache.log4j.Logger;
import org.geotools.data.FeatureSource;
import org.geotools.feature.AttributeType;
import org.geotools.feature.FeatureCollection;
import org.geotools.styling.Style;
import org.opengis.referencing.crs.CoordinateReferenceSystem;

import schmitzm.geotools.styling.StylingUtil;
import skrueger.AttributeMetaData;
import skrueger.i8n.Translation;

import com.vividsolutions.jts.geom.Envelope;

/**
 * This class enables a non Atlas context to use the Atlas LayerPanel
 * {@link JPanel} as a {@link MapContextManagerInterface}
 * 
 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons Kr&uuml;ger</a>
 */
public class StyledFS implements StyledFeatureSourceInterface {
	private static final Logger LOGGER = Logger.getLogger(StyledFS.class);

	private final FeatureSource fs;

	/**
	 * A unique ID which identifies the Layer in the Atlas. It's more important
	 * than it should be ;-)
	 */
	private String id;

	private Style style;

	private Translation title;

	private Translation desc;

	private File sldFile;

	private HashMap<Integer, AttributeMetaData> map;

	/**
	 * This class enables a non Atlas context to use the Atlas LayerPanel
	 * {@link JPanel} as a {@link MapContextManagerInterface}
	 * 
	 * @param fs
	 *            {@link FeatureSource} that is beeing styled.
	 * 
	 * @param sldFile
	 *            may be <code>null</code>. Otherwise the SLD {@link File} to
	 *            import and associate with this {@link StyledFS}
	 */
	public StyledFS(FeatureSource fs, File sldFile) {

		this.fs = fs;
		id = StyledFS.class.getSimpleName()
				+ new Random(new Date().getTime()).nextInt(10000000);

		this.sldFile = sldFile;

		if ((sldFile != null) && (sldFile.exists())) {
			try {
				style = StylingUtil.loadSLD(sldFile)[0];
			} catch (FileNotFoundException e) {
				LOGGER
						.debug("The SLD file passed was empty. Leaving the Style untouched. (We are in the constructor.. so its null");
			} catch (Exception e) {
				LOGGER.warn("Reading SLD failed: " + sldFile, e);
			}
		}
		title = new Translation();
		title.fromOneLine(sldFile.getName());

		desc = new Translation();
		desc.fromOneLine(sldFile.getAbsolutePath());
	}

	public void dispose() {
	}

	/**
	 * Returnes human readable {@link String} of the CRS natively used by this
	 * {@link DpLayer}
	 * 
	 * If crs == null, it will call {@link #getGeoObject()}
	 * 
	 */
	public String getCRSString() {
		if (getCrs() == null)
			return "CRS?";

		return getCrs().getName().getCode();
	}

	public CoordinateReferenceSystem getCrs() {
		return fs.getSchema().getDefaultGeometry().getCoordinateSystem();
	}

	public Translation getDesc() {
		return desc;
	}

	public Envelope getEnvelope() {
		try {
			return fs.getBounds();
		} catch (IOException e) {
			e.printStackTrace();
			return null;
		}
	}

	public FeatureSource getGeoObject() {
		return fs;
	}

	public String getId() {
		return id;
	}

	public ImageIcon getImageIcon() {
		return null;
	}

	public URL getInfoURL() {
		return null;
	}

	public Translation getKeywords() {
		return null;
	}

	public Style getStyle() {
		return style;
	}

	public Translation getTitle() {
		return title;
	}

	public boolean isDisposed() {
		return false;
	}

	/**
	 * If true, this layer will not be shown in the legend. Default = false
	 */
	/**
	 * 
	 * Killed by SK: 6. April 09: Ein Layer soll nicht generell auf
	 * verstecken/nicht verstecken gestellt werden können. Das sind
	 * Eigenschaften der Karte/MapContext, ebenso wie die Reihenfolge der Layer.
	 * Im Atlas verwaltet deshalb nun die Klasse skrueger.atlas.Map welche Layer
	 * nicht in der Legende auftauchen sollen. Meines Wissens hat keiner bisher
	 * die Funktion genutzt.
	 * 
	 * // public boolean isHideInLegend() { // return false; // }
	 */

	public void setDesc(Translation dec) {
		this.desc = dec;
	}

	public void setImageIcon(ImageIcon icon) {
	}

	public void setKeywords(Translation keywords) {
	}

	public void setStyle(Style style) {
		this.style = style;

	}

	public void setTitle(Translation title) {
		this.title = title;

	}

	public void uncache() {
	}

	/**
	 * 
	 */
	public Map<Integer, AttributeMetaData> getAttributeMetaDataMap() {
		if (map == null) {

			map = new HashMap<Integer, AttributeMetaData>();

			// Leaving out the first one, it will be the_geom
			for (int i = 1; i < fs.getSchema().getAttributeCount(); i++) {
				AttributeType att = fs.getSchema().getAttributeType(i);

				AttributeMetaData attMetaData = new AttributeMetaData(i, att
						.getLocalName());
				map.put(i, attMetaData);
			}
		}
		return map;
	}

	/**
	 * @return The {@link File} where the SLD was loaded from or
	 *         <code>null</code> if there didn't exist a {@link File}. 
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public File getSldFile() {
		return sldFile;
	}

	public void setSldFile(File sldFile) {
		this.sldFile = sldFile;
	}

	/**
	 * Returns the features of the {@link FeatureSource}.
	 * 
	 * @see {@link StyledFeaturesInterface}
	 */
	@Override
	public FeatureCollection getFeatureCollection() {
		FeatureCollection features;
		try {
			features = getGeoObject().getFeatures();
		} catch (IOException e) {
			throw new RuntimeException(
					"Error getting the features of the  FeatureSource");
		}
		return features;
	}

	/**
	 * Same as {@link #getGeoObject()} method, but complies to the
	 * {@link StyledFeaturesInterface}
	 * 
	 * @see {@link StyledFeaturesInterface}
	 */
	@Override
	public FeatureSource getFeatureSource() {
		return getGeoObject();
	}

}
