/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package skrueger.geotools;

import java.util.Iterator;
import java.util.Map;
import java.util.Vector;

import org.apache.log4j.Logger;
import org.geotools.data.DefaultQuery;
import org.geotools.data.FeatureSource;
import org.geotools.data.Query;
import org.geotools.data.memory.MemoryDataStore;
import org.geotools.feature.AttributeType;
import org.geotools.feature.FeatureCollection;
import org.opengis.filter.Filter;

import schmitzm.geotools.gui.FeatureCollectionTableModel;
import skrueger.AttributeMetaData;
import skrueger.i8n.I8NUtil;
import skrueger.i8n.Translation;

import com.vividsolutions.jts.geom.Envelope;

/**
 * This class extends the the {@link FeatureCollectionTableModel} with the
 * functionalities of the {@link AttributeMetaData} of
 * {@linkplain StyledLayerInterface styled objects}.
 * <ul>
 * <li>column names are translated according to
 * {@link AttributeMetaData#getTitle()}</li>
 * <li>columns are hidden according to {@link AttributeMetaData#isVisible()()}</li>
 * </ul>
 * 
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
 *         (University of Bonn/Germany)
 */
public class StyledFeatureCollectionTableModel extends
		FeatureCollectionTableModel {
	final static private Logger LOGGER = Logger
			.getLogger(StyledFeatureCollectionTableModel.class);
	/** Holds the data source as styled layer. */
	protected StyledLayerInterface<?> layer = null;
	/** Contains only the visible elements of the {@link AttributeMetaData}-Map */
	protected Map<Integer, AttributeMetaData> visibleAMD = null;
	/** Holds the data source for the table as {@code FeatureSource}. */
	protected FeatureSource featureSource = null;
	/** Contains the complete {@link AttributeMetaData}-Map of the styled layer. */
	protected Map<Integer, AttributeMetaData> origAMD = null;
	/** Holds the current filter on the table */
	protected Filter filter = null;
	/** Holds the Bounds for all features. Only set once during the constructor **/
	protected Envelope bounds;

	

	/**
	 * Creates a new table model for a styled layer.
	 * 
	 * @param layer
	 *            the styled layer
	 */
	public StyledFeatureCollectionTableModel(
			StyledFeatureCollectionInterface layer) {
		this(layer, Filter.INCLUDE);
	}

	/**
	 * Creates a new table model for a styled layer.
	 * 
	 * @param layer
	 *            the styled layer
	 * @param filter
	 *            filter applied to the table
	 */
	public StyledFeatureCollectionTableModel(
			StyledFeatureCollectionInterface layer, Filter filter) {
		super();
		setFeatureCollection(layer, filter);
	}


	/**
	 * Creates a new table model for a styled layer.
	 * 
	 * @param layer
	 *            the styled layer
	 */
	public StyledFeatureCollectionTableModel(StyledFeaturesInterface layer) {
		this(layer, Filter.INCLUDE);
	}

	/**
	 * Creates a new table model for a styled layer.
	 * 
	 * @param layer
	 *            the styled layer
	 * @param filter
	 *            filter applied to the table
	 */
	public StyledFeatureCollectionTableModel(StyledFeaturesInterface layer,
			Filter filter) {
		super();
		setFeatureCollection(layer, filter);
	}

	/**
	 * Sets a new data source for the table.
	 * 
	 * @param fs
	 *            the feature source
	 * @param amd
	 *            {@link AttributeMetaData}-Map to define the visible attributes
	 *            and translation
	 */
	protected void setFeatureSource(FeatureSource fs,
			Map<Integer, AttributeMetaData> amd, Filter filter)
			throws Exception {
		if (filter == null)
			filter = Filter.INCLUDE;

		this.featureSource = fs;
		this.filter = filter;
		this.origAMD = amd;
		this.visibleAMD = null;

		FeatureCollection fc = null;
		if (fs != null) {

			bounds = fs.getBounds();

			Query query = new DefaultQuery(fs.getSchema().getTypeName(), filter);
			if (amd != null) {
				// determine the names of the visible Attributes
				this.visibleAMD = StyledLayerUtil.getVisibleAttributeMetaData(
						amd, true);
				Vector<String> visibleAttrNames = new Vector<String>();
				// Add the column with the geometry (usually "the_geom")
				visibleAttrNames.add(fs.getSchema().getDefaultGeometry()
						.getLocalName());
				for (int attrIdx : visibleAMD.keySet()) {

					/**
					 * If the user removed columns from the schema of the DBF
					 * file, there might exist AttributeMetaData for columns
					 * that don't exists. We check here to avoid an
					 * ArrayOutOfIndex.
					 */
					if (attrIdx < fs.getSchema().getAttributeCount()) {
						final AttributeType attributeTypeAtIdx = fs.getSchema()
								.getAttributeType(attrIdx);
						visibleAttrNames.add(attributeTypeAtIdx.getLocalName());
					} else {
						LOGGER.warn("AttributeMetaData has been found for columnIdx="+attrIdx+", but fs.getSchema().getAttributeCount() = "+fs.getSchema().getAttributeCount()+". Ignored.");
					}
				}

				// create a query for the visible attributes
				String[] properties = visibleAttrNames.toArray(new String[0]);

				LOGGER.debug("Query contains the following attributes: "
						+ visibleAttrNames);

				query = new DefaultQuery(fs.getSchema().getTypeName(), filter,
						properties);
			}
			fc = fs.getFeatures(query);

			// FAILS:!!!, even with query = new
			// DefaultQuery(fs.getSchema().getTypeName(), filter);
			// java.lang.UnsupportedOperationException: Unknown feature
			// attribute: PQM_MOD
			// at
			// schmitzm.geotools.feature.FeatureOperationTree.evaluate(FeatureOperationTree.java:93)
			// bounds = fc.getBounds();
			// SK, 17.4.2009
			//      
			// System.out.println("Filter = "+filter);
			// System.out.println("Size of FC = "+fc.size());
			// System.out.println("anz att= "+fc.getNumberOfAttributes());
		}
		setFeatureCollection(fc);
	}

	/**
	 * Converts the {@code StyledFeatureCollection} to a {@code FeatureSource}
	 * and sets this as the new data source for the table.
	 * 
	 * @param fs
	 *            the feature source
	 * @param amd
	 *            {@link AttributeMetaData}-Map to define the visible attributes
	 *            and translation
	 */
	public void setFeatureCollection(StyledFeaturesInterface layer,
			Filter filter) {
		this.layer = layer;
		try {
			if (layer == null)
				setFeatureSource(null, null, null);
			else {
				FeatureCollection fc = layer.getFeatureCollection();
				String fcName = fc.getSchema().getTypeName();
				FeatureSource fs = new MemoryDataStore(fc)
						.getFeatureSource(fcName);
				setFeatureSource(fs, layer.getAttributeMetaDataMap(), filter);
			}
		} catch (Exception err) {
			throw new RuntimeException(err);
		}
	}

	/**
	 * Sets the {@code StyledFeatureCollection} as new data source for the
	 * table.
	 * 
	 * @param fs
	 *            the feature source
	 * @param amd
	 *            {@link AttributeMetaData}-Map to define the visible attributes
	 *            and translation
	 */
	public void setFeatureCollection(StyledFeatureSourceInterface layer,
			Filter filter) {
		this.layer = layer;
		try {
			if (layer == null)
				setFeatureSource(null, null, null);
			else
				setFeatureSource(layer.getGeoObject(), layer
						.getAttributeMetaDataMap(), filter);
		} catch (Exception err) {
			throw new RuntimeException(err);
		}
	}

	/**
	 * Resets the filter for the table.
	 * 
	 * @param filter
	 *            a filter
	 */
	public void setFilter(Filter filter) {
		try {
			setFeatureSource(this.featureSource, this.origAMD, filter);
		} catch (Exception err) {
			LOGGER.error("Setting the filter of the table model", err);
			throw new RuntimeException(err);
		}
	}

	/**
	 * @return <code>Filter.INCLUDE</code> or the {@link Filter} applied to the
	 *         Features
	 */
	public Filter getFilter() {
		return this.filter;
	}

	/**
	 * After calling {@code super.reorganize(.)} this method replaced the column
	 * descriptions with the titles of the {@code AttributeMetaData}.
	 * 
	 * @param fireTableStructureChanged
	 *            indicates whether a table event is initiated after reorganize
	 */
	@Override
	protected void reorganize(boolean fireTableStructureChanged) {
		super.reorganize(false);
		// translate the column names
		if (visibleAMD != null) {
			Iterator<Integer> keys = visibleAMD.keySet().iterator();
			for (int i = 0; i < colNames.length && keys.hasNext(); i++) {
				Translation title = visibleAMD.get(keys.next()).getTitle();
				if (!I8NUtil.isEmpty(title)) {
					// System.out.println("set colname " + i + " to " +
					// title.toString());
					colNames[i] = title.toString();
				}
			}
		}
		if (fireTableStructureChanged)
			fireTableStructureChanged();
	}

	/**
	 * @return Cached bounds for the whole dataset (without applying the filter)
	 *         or <code>null</code>
	 */
	public Envelope getBounds() {
		return bounds;
	}
}
