/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package skrueger.geotools.labelsearch;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;

import org.apache.log4j.Logger;
import org.geotools.data.DefaultQuery;
import org.geotools.feature.AttributeType;
import org.geotools.feature.Feature;
import org.geotools.feature.FeatureCollection;
import org.geotools.feature.FeatureType;
import org.geotools.map.MapLayer;
import org.geotools.styling.Style;
import org.geotools.styling.TextSymbolizer;
import org.opengis.filter.Filter;
import org.opengis.filter.expression.Expression;
import org.opengis.filter.expression.PropertyName;

import schmitzm.geotools.styling.StylingUtil;
import schmitzm.lang.LangUtil;
import schmitzm.lang.ResourceProvider;

/**
 * This class allows to search for a {@link String} in a map. The algorithm will
 * analyze the {@link Style} of every visible(?) layer and determine the label
 * attribute. This field is then searched for every feature.
 * 
 * @author Stefan A. Krüger
 * 
 */
public class LabelSearch {
	final static private Logger LOGGER = Logger.getLogger(LabelSearch.class);

	public static ResourceProvider RESOURCE = new ResourceProvider(LangUtil
			.extendPackagePath(LabelSearch.class, "labelsearch"),
			Locale.ENGLISH);

	public static String R(String key, Object... values) {
		return RESOURCE.getString(key, values);
	}

	protected final schmitzm.geotools.gui.JMapPane mapPane;

	public LabelSearch(final schmitzm.geotools.gui.JMapPane mapPane) {
		this.mapPane = mapPane;
	}

	private AttributeType getLabelAttribute(final TextSymbolizer ts,
			final FeatureType schema) {
		if (ts == null) {
			// This layer has no labels
			return null;
		}
		final Expression labelExp = ts.getLabel();
		if (labelExp instanceof PropertyName) {
			final PropertyName pn = (PropertyName) labelExp;
			final String propertyName = pn.getPropertyName();
			return schema.getAttributeType(propertyName);
		} else {
			// When does this happen
		}

		return null;
	}

	public List<SearchResult> search(final String string) {

		final String searchMe = string.toLowerCase();

		final ArrayList<SearchResult> hits = new ArrayList<SearchResult>();

		for (final MapLayer ml : mapPane.getContext().getLayers()) {
			try {

				// System.out.println("layer = "+ml.getTitle());

				if (!ml.isVisible())
					continue;

				final List<TextSymbolizer> allTS = StylingUtil
						.getTextSymbolizers(ml.getStyle());
				if (allTS.size() == 0) {
					// A layer without any TextSymbolizer doesn't have to be
					// searched any more.
					continue;
				}

				final String typeName = ml.getFeatureSource().getSchema()
						.getTypeName();

				// Expression labelExp = ts.getLabel();
				// ff.like(labelExp, "*"+searchMe+"*");
				// FeatureCollection features =
				// ml.getFeatureSource().getFeatures(
				// new DefaultQuery(typeName, ff.like(labelExp,
				// "*"+searchMe+"*"), properties));

				final FeatureCollection features = ml
						.getFeatureSource()
						.getFeatures(new DefaultQuery(typeName, Filter.INCLUDE));

				// new MemoryDataStore().getFeatureSource(typeName)

				/**
				 * We do the comparison NOT with a ff.like, because that doesn't
				 * support case insensitivity and i don't find a lower or UPPER
				 * function.
				 */
				final Iterator<Feature> fi = features.iterator();
				try {
					while (fi.hasNext()) {
						final Feature f = fi.next();

						final TextSymbolizer ts = StylingUtil
								.getTextSymbolizer(ml.getStyle(), f);
						if (ts == null)
							continue;

						final AttributeType labelAttribute = getLabelAttribute(
								ts, ml.getFeatureSource().getSchema());

						if (labelAttribute == null) {
							continue;
						}

						// System.out.println("labelAttrib local name" +
						// labelAttribute.getLocalName());

						final Object value = f.getAttribute(labelAttribute
								.getLocalName());

						// System.out.println("labelAttrib value " + value);

						if (value == null) {
							LOGGER
									.info("Skipping f: getLocalName() is null for feature="
											+ f);
							continue;
						}

						/**
						 * LabelString ist z.B. "IMPETUS pluviograph". Suchwort
						 * "plu" soll treffen. Also wird nach spaces zerlegt und
						 * dann gesucht
						 */
						final String labelString = value.toString()
								.toLowerCase();
						if (labelString.startsWith(searchMe)) {
							hits.add(createSearchResult(f, value.toString(), ml
									.getTitle(), ml));
						} else {
							final String[] parts = labelString.trim()
									.split(" ");
							for (final String part : parts) {
								if (part.startsWith(searchMe)) {
									hits.add(createSearchResult(f, value
											.toString(), ml.getTitle(), ml));
									break;
								}
							}
						}

					}
				} finally {
					features.close(fi);
				}

			} catch (final IOException e) {
				// Searching this layer failed
				LOGGER.error(e);
			}
		} // next layer

		// Hits from the top-most layer should appear first.
		Collections.reverse(hits);

		return hits;
	}

	protected SearchResult createSearchResult(final Feature f,
			final String title, final String inTitle, MapLayer ml) {
		return new SearchResultFeature(f, title, inTitle, mapPane, ml);
	}

}
