/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
/** 
 Copyright 2008 Stefan Alfons Krüger 
 
 atlas-framework - This file is part of the Atlas Framework

 This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
 This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

 Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
 Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
 Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/
package skrueger.geotools.selection;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashSet;
import java.util.Set;
import java.util.Vector;

import javax.swing.JTable;
import javax.swing.ListSelectionModel;
import javax.swing.event.ListSelectionListener;

import org.geotools.feature.Feature;
import org.geotools.map.MapLayer;
import org.geotools.styling.FeatureTypeStyle;
import org.geotools.styling.Style;
import org.opengis.filter.identity.FeatureId;

import schmitzm.geotools.FilterUtil;
import schmitzm.geotools.gui.JMapPane;
import schmitzm.geotools.map.event.FeatureSelectedEvent;
import schmitzm.geotools.map.event.JMapPaneEvent;
import schmitzm.geotools.map.event.JMapPaneListener;
import schmitzm.geotools.styling.StylingUtil;
import schmitzm.lang.LangUtil;
import skrueger.geotools.MapPaneToolBar;
import skrueger.geotools.StyledLayerInterface;

/**
 * This class keeps the selection of a (feature) {@link JTable} synchronized
 * with the {@link StyledLayerSelectionModel} of a layer. This is done by
 * implementing:
 * <ul>
 * <li>a {@link PropertyChangeListener} which listens to the
 * {@link StyledLayerSelectionModel} and accordingly changes the {@link JTable}
 * selection</li>
 * <li>a {@link ListSelectionListener} which listens to the {@link JTable} and
 * accordingly changes the {@link StyledLayerSelectionModel} selection</li>
 * </ul>
 * After creating, the instance of this synchronizer must be added as listener
 * to both, the {@link StyledLayerSelectionModel} and the table's
 * {@link ListSelectionModel}.
 * 
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
 *         (University of Bonn/Germany)
 */
public class FeatureMapLayerSelectionSynchronizer extends
		StyledLayerSelectionModelSynchronizer<StyledFeatureLayerSelectionModel>
		implements JMapPaneListener {
	public static final String SELECTION_STYLING = "SELECTION";
	/**
	 * Holds the {@link MapLayer} to keep synchronized with the layer selection
	 * model.
	 */
	protected final MapLayer mapLayer;
	protected final StyledLayerInterface<?> styledLayer;
	protected final JMapPane mapPane;
	private final MapPaneToolBar toolBar;

	/**
	 * Creates a new synchronizer
	 * 
	 * @param layerSelModel
	 *            layer selection model to keep synchronized with the
	 *            {@link MapLayer}
	 * 
	 * @param mapLayer
	 *            {@link MapLayer} to keep synchronized with.
	 */
	public FeatureMapLayerSelectionSynchronizer(
			StyledFeatureLayerSelectionModel layerSelModel,
			StyledLayerInterface<?> styledLayer, MapLayer mapLayer,
			JMapPane mapPane, MapPaneToolBar toolBar) {

		super(layerSelModel);
		this.styledLayer = styledLayer;

		this.mapLayer = mapLayer;
		this.mapPane = mapPane;
		this.toolBar = toolBar;
	}

	/**
	 * Called by {@link StyledLayerSelectionModel} when a the selection on other
	 * selection components (map, chart, ...) has changed. When calling this
	 * 
	 * method changes the {@link MapLayer} selection according to the
	 * {@link StyledLayerSelectionModel} selection.
	 * 
	 * @param evt
	 *            an event
	 */
	@Override
	public void propertyChange(PropertyChangeEvent evt) {

		if (!isEnabled())
			return;

		if (!(evt instanceof StyledLayerSelectionEvent))
			return;
		StyledLayerSelectionEvent selEvt = (StyledLayerSelectionEvent) evt;
		// Only react on own layer and ignore events invoked by
		// this component itself
		if (selEvt.getEmitter() != layerSelModel || selectionChangeCausedByMe)
			return;
		// Apply new selection on table (except this event is one of
		// some more events)
		Vector<String> newSelection = layerSelModel.getSelection();
		if (newSelection == null)
			return;

		// Avoid event circles in valueChanged(..)
		selectionChangeCausedByMe = true;

		changeLayerStyle(newSelection);

		// Danger of event circles in valueChanged(..) banned
		selectionChangeCausedByMe = false;
	}

	/**
	 * Changes the Style of the {@link MapLayer} to reflect changes of the
	 * selection.
	 * 
	 * @param newSelection
	 *            A {@link Vector} of Feature-IDs that are selected.
	 */
	private void changeLayerStyle(final Vector<String> newSelection) {
		Style selectionMapStyle = null;
		try {
			if (newSelection.isEmpty()) {

				selectionMapStyle = styledLayer.getStyle();
				// LOGGER.debug("NO SELECTION .. set to original style directly");

			} else {
				LOGGER.debug("SELECTION .. change style");

				// We take Style from the MapLayer that is displayed at the
				// moment. We do not use the styledLayer.getStyle, because in
				// the atlas, this always return the default style, but
				// additional styles might be selected.
				// Taking the style from the mapLayer indicated, that we have to
				// remove any selection rules first.
				Style originalStyle = mapLayer.getStyle();

				// TODO The default style is not good here. We need
				// .createSelectionStyle(normalStyle, geoObj);
				selectionMapStyle = StylingUtil
						.createSelectionStyle(styledLayer.getGeoObject());
				// selectionMapStyle = StylingUtil
				// .createDefaultStyle(styledLayer.getGeoObject());

				selectionMapStyle.getFeatureTypeStyles()[0]
						.setName(SELECTION_STYLING);

				/**
				 * 
				 * Add a Filter to the selectionMapStyle, so that it is only
				 * used on objects that are selected. <br/>
				 * 
				 * Note 1:<br/>
				 * It is NEVER allowed to GeoTools extend Filter () { .. } (and write
				 * tests into the evaluate block). Especially for the
				 * ShapeFileRenderer, we may only use a geotools Filter.<br/>
				 * 
				 * Note 2:<br/>
				 * The FilterUtil.FILTER_FAC2.id(fids) wants a set of FeatureId-Objects!
				 */

				Set<FeatureId> fids = new HashSet<FeatureId>();
				for (String fid : newSelection) {
					fids.add(FilterUtil.FILTER_FAC2.featureId(fid));
				}

				selectionMapStyle.getFeatureTypeStyles()[0].getRules()[0]
						.setFilter(FilterUtil.FILTER_FAC2.id(fids));

				FeatureTypeStyle[] originalFeatureTypeStyles = originalStyle
						.getFeatureTypeStyles();
				FeatureTypeStyle[] newFeatureTypes;
				if (originalFeatureTypeStyles[originalFeatureTypeStyles.length - 1]
						.getName() != null
						&& originalFeatureTypeStyles[originalFeatureTypeStyles.length - 1]
								.getName().equals(SELECTION_STYLING)) {
					newFeatureTypes = Arrays.copyOf(originalFeatureTypeStyles,
							originalFeatureTypeStyles.length - 1);
				} else {
					newFeatureTypes = originalFeatureTypeStyles;
				}

				// The last FTS is the selection FTS
				newFeatureTypes = LangUtil.extendArray(newFeatureTypes,
						selectionMapStyle.getFeatureTypeStyles());

				selectionMapStyle.setFeatureTypeStyles(newFeatureTypes);

			}

			mapLayer.setStyle(selectionMapStyle);
			mapPane.refresh();
		} catch (Exception e) {
			LOGGER.error("Error while trying to create a selection style", e);
		}
	}

	/**
	 * Used to synchronize {@link FeatureSelectedEvent}s with the
	 * {@link StyledFeatureLayerSelectionModel}
	 */
	@Override
	public void performMapPaneEvent(JMapPaneEvent e) {

		// Ignore event if it is caused by us or the synchronizer is disabled.
		if (!isEnabled() || selectionChangeCausedByMe)
			return;

		if (!(e instanceof FeatureSelectedEvent)) {
			// LOGGER.debug("Ignoring event " + e);
			return;
		}

		/**
		 * Only listen to FeatureSelectedEvents if an appropriate tool is
		 * selected.
		 */
		final int selectedTool = toolBar.getSelectedTool();
		if (selectedTool != MapPaneToolBar.TOOL_SELECTION_ADD
				&& selectedTool != MapPaneToolBar.TOOL_SELECTION_REMOVE
				&& selectedTool != MapPaneToolBar.TOOL_SELECTION_SET) {
			return;
		}

		// only listen to events directly coming from JMapPane
		// selection (ignore selections from FilterDialog)
		if (e.getSourceObject() != this.mapPane)
			return;

		FeatureSelectedEvent fse = (FeatureSelectedEvent) e;

		/**
		 * Checking, that the FeatureSelectedEvent actually contains features
		 * from this layer
		 */
		final String sourceID = fse.getSourceLayer().getTitle();
		final String syncForID = mapLayer.getTitle();
		if (sourceID != null && syncForID != null
				&& !sourceID.equals(syncForID)) {
			LOGGER.debug("Ignoring a FeatureSelectedEvent from " + sourceID);
			return;
		}

		LOGGER.debug("do event " + fse);

		// Avoid event circles in propertyChange(..)
		selectionChangeCausedByMe = true;

		Collection<Feature> selectionResult = (Collection<Feature>) fse
				.getSelectionResult();

		// reset the selection of the DpLayerSelectionModel
		// layerSelModel.setValueIsAdjusting(true);

		if (selectedTool == MapPaneToolBar.TOOL_SELECTION_ADD) {
			layerSelModel.setValueIsAdjusting(true);
			for (Feature f : selectionResult) {
				layerSelModel.addSelection(f.getID());
			}
			layerSelModel.setValueIsAdjusting(false);
		} else if (selectedTool == MapPaneToolBar.TOOL_SELECTION_SET) {
			layerSelModel.setValueIsAdjusting(true);
			layerSelModel.clearSelection();

			for (Feature f : selectionResult) {
				layerSelModel.addSelection(f.getID());
			}

			LOGGER.debug("Setting selection to " + selectionResult.size());
			layerSelModel.setValueIsAdjusting(false);
		} else if (selectedTool == MapPaneToolBar.TOOL_SELECTION_REMOVE) {
			layerSelModel.setValueIsAdjusting(true);
			for (Feature f : selectionResult) {
				layerSelModel.removeSelection(f.getID());
			}
			layerSelModel.setValueIsAdjusting(false);
		}

		// Show selected features in the map, which is not automatically done by
		// the origin: FeatureSelectedEvent
		changeLayerStyle(layerSelModel.getSelection());

		// Danger of event circles in propertyChange(..) banned
		selectionChangeCausedByMe = false;
	}

}
