/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package skrueger.i8n;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Random;

import javax.swing.JComponent;

import org.apache.log4j.Logger;

/**
 * Represents a {@link HashMap} of translations. toString() returns the
 * appropriate translation
 * 
 * @author @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
 *         Kr&uuml;ger</a>
 */

public class Translation extends HashMap<String, String> {
	public static final String LOCALECHANGE_PROPERTY = "localechange";
	public static final String NO_TRANSLATION = "NO TRANSLATION";
	public static final String DEFAULT_KEY = "default";
	static final Logger log = Logger.getLogger(Translation.class);
	static String activeLang = Locale.getDefault().getLanguage();

	static protected List<PropertyChangeListener> listeners = new ArrayList<PropertyChangeListener>();

	static {

		// TODO default aus Locale auslesen und mit möglichen vergleichen...
		// mmm.. vor laden von atlasml immer DEFAULT_KEY, also hier nicht

		// Get default locale
		Locale locale = Locale.getDefault();
		setActiveLang(locale.getLanguage());
	}

	private List<ActionListener> actionListeners = new ArrayList<ActionListener>();

	@Override
	/*
	 * @comment To make a copy of a translation see methods toOneLine() and
	 * fromOneLine()
	 */
	public Translation clone() {
		return (Translation) super.clone();
	}

	/**
	 * Get the two-letter language sting that is active
	 */
	public static String getActiveLang() {
		return activeLang;
	}

	/**
	 * Set up the {@link Translation}-system to use language. If a change is
	 * performed, events are fired to listeners. Nothing is done if the new
	 * language equals the old language. The system's default locale is changed.
	 * 
	 * @param newLang
	 *            The ISO Code of the new active language
	 */
	public static void setActiveLang(String newLang) {
		setActiveLang(newLang, true);
	}

	/**
	 * Set up the {@link Translation}-system to use language. If a change is
	 * performed, events are fired to listeners. Nothing is done if the new
	 * language equals the old language.
	 * 
	 * @param newLang
	 *            The ISO Code of the new active language
	 * 
	 * @param setDefaultLocale
	 *            Shall the system's default locale be changed?
	 */
	public static void setActiveLang(String newLang, boolean setDefaultLocale) {
		if (getActiveLang().equals(newLang)) {
			return;
		}

		if (!I8NUtil.isValidISOLangCode(newLang)) {
			throw new IllegalArgumentException("'" + newLang
					+ "' is not a valid ISO language code.");
		}

		Locale newLocale = new Locale(newLang);
		if (setDefaultLocale)
			Locale.setDefault(newLocale);

		/**
		 * Setting default locale for Swing JComponents to work around bug
		 * http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=4884480
		 */
		JComponent.setDefaultLocale(newLocale);

		Translation.activeLang = newLang;

		fireLocaleChangeEvents();

		log.info("skrueger.i8n.Translation switched ActiveLang to " + newLang);
	}

	/**
	 * Initializes a new {@link Translation} with a default translation if a
	 * simple text is passed. If a "oneLine" text is parsed, it is interpreted.
	 * Other translations may be added later - this is a HashMap<br/>
	 * 
	 * @param defaultTranslation
	 * 
	 * @see public Translation(List<String> languages, String
	 *      defaultTranslation) {
	 * 
	 */
	public Translation(String defaultTranslation) {

		fromOneLine(defaultTranslation);

	}

	/**
	 * Initializes a new {@link Translation}, an uses the given String to
	 * initialize the {@link Translation} for all languages codes passed.
	 * 
	 * The translations can be changed later
	 */
	public Translation(List<String> languages, String defaultTranslation) {
		// put(DEFAULT_KEY, defaultTranslation);
		if (languages == null) {
			put(DEFAULT_KEY, defaultTranslation);
		} else
			for (String code : languages) {
				if (code.equals(getActiveLang())) {
					put(code, defaultTranslation);
				}
			}
	}

	/**
	 * Sometimes Translations are optional, like for keywords.
	 */
	public Translation() {
	}

	/**
	 * Fills the {@link Translation} with the values coded into the String
	 * Format of {@link String} is: "de{Baum}en{tree}"
	 * <p>
	 * <ul>
	 * <li>If <code>oneLineCoded</code> is empty or null, NO TRANSLATION is set.
	 * <li>If format can't be recognized, the {@link String} is interpreted as
	 * the translation in the <code>{@value #DEFAULT_KEY}</code> language
	 * 
	 * @author Stefan Alfons Krüger
	 */
	public void fromOneLine(final String oneLineCoded) {
		clear();
		if ((oneLineCoded == null) || (oneLineCoded.equals(""))) {
			put(DEFAULT_KEY, "");
			return;
		}

		if (oneLineCoded.indexOf("}") == -1) {
			// log.warn("The String '"+oneLineCoded+"' is not in oneLine coded => put(DEFAULT_KEY,oneLineCoded);");
			put(DEFAULT_KEY, oneLineCoded);
		}

		String eatUp = oneLineCoded;
		while (eatUp.indexOf("}") != -1) {
			String substring = eatUp.substring(0, eatUp.indexOf("}"));

			// log.debug("substring = "+substring);
			String key = substring.substring(0, substring.indexOf("{"));
			String value = substring.substring(substring.indexOf("{") + 1,
					substring.length());
			// log.debug("key="+key);
			// log.debug("value="+value);
			put(key, value);
			eatUp = eatUp.substring(eatUp.indexOf("}") + 1);
		}
	}

	/**
	 * Exports the Translations to a String of the Format: "de{Baum}en{tree}"
	 * 
	 * @author Stefan Alfons Krüger
	 */
	public String toOneLine() {
		return I8NUtil.toOneLine(this);
	}

	/**
	 * Returns the right translation by using the {@link #activeLang} field. If
	 * no translation is set, an ugly String {@link #NO_TRANSLATION} will re
	 * returned. This might be changed for the final release. If the correct
	 * language was not found, any entry in the {@link Translation}
	 * {@link HashMap} will be returned, that contains more than an empty
	 * string.
	 */
	@Override
	public String toString() {
		if (get(activeLang) != null) {
			return get(activeLang);
		}
		// ****************************************************************************
		// MS: The ISDSS needs the concept of the default lang!! So I took the
		// following in again!!
		// ****************************************************************************
		// else return "";
		// //****************************************************************************
		// // The following is commented out.. the concept of the default lang
		// seems to be bad....
		// //****************************************************************************
		// MS:
		else {
			if (get(DEFAULT_KEY) != null) {
				// log.debug("default lang returned, cuz the translation to "+activeLang+" was not found. Schmeiss raus martin, wenn du das mit der default trans geklärt hast.");
				return get(DEFAULT_KEY);
			}

			// log.debug("return first best <> '' ");
			if (size() > 0)
				for (String s : values()) {
					if ((s != null) && (s.trim().length() > 0))
						return s;
				}
		}
//		log.warn("No translation found!");
		return NO_TRANSLATION;
	}

	/**
	 * Copy this {@link Translation} to another {@link Translation} e.g. for
	 * editing
	 * 
	 * @return the destination {@link Translation}
	 */
	public Translation copy(Translation backup) {
		if (backup == null)
			throw new IllegalArgumentException(
					"Target translation may not be null.");
		for (String s : keySet()) {
			backup.put(s, get(s));
		}
		return backup;
	}

	/**
	 * {@link PropertyChangeListener} can be registered to be informed when the
	 * {@link Locale} changed.
	 * 
	 * @param propertyChangeListener
	 */
	public static void addLocaleChangeListener(
			PropertyChangeListener propertyChangeListener) {
		listeners.add(propertyChangeListener);
	}

	/**
	 * Informs all registered {@link PropertyChangeListener}s about a change of
	 * the the {@link Locale}.
	 */
	public static void fireLocaleChangeEvents() {
		PropertyChangeEvent pce = new PropertyChangeEvent(new Translation(
				new ArrayList<String>(), "fakeSource"), LOCALECHANGE_PROPERTY,
				null, getActiveLang());
		for (PropertyChangeListener pcl : listeners) {
			if (pcl != null)
				pcl.propertyChange(pce);
		}
	}

	public void addTranslationChangeListener(ActionListener actionListener) {
		actionListeners.add(actionListener);
	}
	
	public boolean removeTranslationChangeListener(ActionListener actionListener) {
		return actionListeners.remove(actionListener);
	}

	public void fireTranslationChangedEvents(String lang) {
		ActionEvent ae = new ActionEvent(this, new Random().nextInt(), lang);
		
		for (ActionListener al : actionListeners) {
			al.actionPerformed( ae);
		}
	}
	
	@Override
	public String put(String lang, String value) {
		String result = super.put(lang, value);
		fireTranslationChangedEvents(lang);
		return result;
	}
	
	
}
