/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package skrueger.swing;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.FlowLayout;
import java.awt.Window;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JRootPane;
import javax.swing.KeyStroke;

import schmitzm.swing.SwingUtil;
import skrueger.i8n.Translation;

public class TranslationAskJDialog extends JDialog {

	private String[] backup = new String[50]; // Maximum 50 languages ;-)
	private OkButton okButton;
	private CancelButton cancelButton;

	public static final String PROPERTY_CANCEL_AND_CLOSE = "CANCEL";
	public static final String PROPERTY_APPLY_AND_CLOSE = "APPLY";

	private JComponent[] translationEditJPanelsOrJustComponents;

	private boolean hasBeenCanceled;

	private JButton[] optionalButtons;

	/**
	 * Since the registerKeyboardAction() method is part of the JComponent class
	 * definition, you must define the Escape keystroke and register the
	 * keyboard action with a JComponent, not with a JDialog. The JRootPane for
	 * the JDialog serves as an excellent choice to associate the registration,
	 * as this will always be visible. If you override the protected
	 * createRootPane() method of JDialog, you can return your custom JRootPane
	 * with the keystroke enabled:
	 */
	@Override
	protected JRootPane createRootPane() {
		KeyStroke stroke = KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0);
		JRootPane rootPane = new JRootPane();
		rootPane.registerKeyboardAction(new ActionListener() {

			public void actionPerformed(ActionEvent e) {
				cancel();
			}

		}, stroke, JComponent.WHEN_IN_FOCUSED_WINDOW);

		return rootPane;
	}

	/**
	 * The {@link TranslationAskJDialog} fills its content pane with an
	 * arbitrary number of components. If these {@link Component}s are
	 * {@link TranslationEditJPanel}s, the {@link JDialog} manages to backup the
	 * values and restore them if the dialog is canceled. Other
	 * {@link JComponent}s are just displayed.<br/>
	 * This class handles the cancel button itself. You may still want to listen
	 * to PROPERTY_APPLY_AND_CLOSE events. This dialog is modal. The dialog has
	 * to be set visible afterwards.<br/>
	 * 
	 * @param owner
	 *            A component of the GUI that this dialog is related to. If no
	 *            {@link Window} is passed, SwingUtil.getParentWindow(owner) is
	 *            called.
	 */
	public TranslationAskJDialog(Component owner,
			final JComponent... translationEditJPanels) {
		super(SwingUtil.getParentWindow(owner));
		setComponents(translationEditJPanels);
	}

	/**
	 * The {@link TranslationAskJDialog} fills its content pane with an
	 * arbitrary number of components. If these {@link Component}s are
	 * {@link TranslationEditJPanel}s, the {@link JDialog} manages to backup the
	 * values and restore them if the dialog is canceled. Other
	 * {@link JComponent}s are just displayed.<br/>
	 * This class handles the cancel button itself. You may still want to listen
	 * to PROPERTY_APPLY_AND_CLOSE events. This dialog is modal. The dialog has
	 * to be set visible afterwards.<br/>
	 * Using this constructor, you have to call setComponents afterwards.
	 */
	public TranslationAskJDialog(Component owner) {
		super(SwingUtil.getParentWindow(owner));
	}

	/**
	 * The {@link TranslationAskJDialog} fills its content pane with an
	 * arbitrary number of components. If these {@link Component}s are
	 * {@link TranslationEditJPanel}s, the {@link JDialog} manages to backup the
	 * values and restore them if the dialog is canceled. Other
	 * {@link JComponent}s are just displayed.
	 * 
	 * @param translationEditJPanels
	 *            Arbitrary list of {@link JComponent}s and
	 *            {@link TranslationEditJPanel}s.
	 */
	public void setComponents(final JComponent... translationEditJPanels) {
		this.translationEditJPanelsOrJustComponents = translationEditJPanels;

		backup();

		init();
	}

	/**
	 * Stores the original values of all {@link TranslationEditJPanel}s so
	 * cancel works.
	 */
	protected void backup() {
		// Remember backups for all the TranslationEditJPanel
		int count = 0;
		for (JComponent component : translationEditJPanelsOrJustComponents) {
			if (component instanceof TranslationEditJPanel) {
				TranslationEditJPanel tep = (TranslationEditJPanel) component;
				Translation orig = tep.getTranslation();

				// We don't want to overwrite the Translation object on
				// restore(). We just want to change its value.
				backup[count++] = orig.toOneLine();
			}
		}
	}

	private void init() {
		setDefaultCloseOperation(JDialog.DO_NOTHING_ON_CLOSE);
		addWindowListener(new WindowAdapter() {

			public void windowClosing(WindowEvent e) {
				cancel();
			}

		});
		SwingUtil.centerFrameOnScreen(this);
		Box box = Box.createVerticalBox();
		for (JComponent panel : translationEditJPanelsOrJustComponents) {
			panel.setAlignmentX(java.awt.Component.LEFT_ALIGNMENT);
			panel.setBorder(BorderFactory.createEmptyBorder(5, 6, 5, 6));
			box.add(panel);

		}
		JPanel cp = new JPanel(new BorderLayout());
		cp.add(box, BorderLayout.WEST);
		cp.add(getButtons(), BorderLayout.SOUTH);
		setContentPane(cp);

		setTitle(SwingUtil.R("TranslationAskJDialog.Title"));
		setModal(true);
		pack();
	}

	public void setButtons(JButton... optionalButtons) {
		this.optionalButtons = optionalButtons;
		init();
	}

	/**
	 * Called when the dilaog is closed using the cancel button.
	 */
	protected void cancel() {
		restore();
		firePropertyChange(PROPERTY_CANCEL_AND_CLOSE, null, null);
		setCancelled(true);
		setVisible(false);
		dispose();
	}

	/**
	 * Used to restore all the values when cancel has been pressed.
	 */
	protected void restore() {
		int count = 0;
		for (JComponent component : translationEditJPanelsOrJustComponents) {
			if (component instanceof TranslationEditJPanel) {
				TranslationEditJPanel tep = (TranslationEditJPanel) component;
				tep.getTranslation().fromOneLine(backup[count++]);
			}
		}
	}

	private JComponent getButtons() {
		JPanel jPanel = new JPanel(new FlowLayout(FlowLayout.RIGHT));

		if (optionalButtons != null)
			for (JButton b : optionalButtons) {
				jPanel.add(b);
			}

		if (okButton == null) {
			okButton = new OkButton(new AbstractAction() {
				{
					// Set a mnemonic character. In most look and feels, this
					// causes the
					// specified character to be underlined This indicates that
					// if the component
					// using this action has the focus and In some look and
					// feels, this causes
					// the specified character in the label to be underlined and
					putValue(Action.MNEMONIC_KEY, new Integer(
							java.awt.event.KeyEvent.VK_E));

					// // Set tool tip text
					// putValue(Action.SHORT_DESCRIPTION,
					// "Accept the changes made to the translation."); //i8n

				}

				public void actionPerformed(ActionEvent evt) {
					TranslationAskJDialog.this.firePropertyChange(
							PROPERTY_APPLY_AND_CLOSE, null, null);

					if (!checkValidInputs())
						return;

					okClose();

				}

			});

		}
		jPanel.add(okButton);

		if (cancelButton == null) {
			cancelButton = new CancelButton(new AbstractAction("") {
				public void actionPerformed(ActionEvent evt) {
//					restore();
//					TranslationAskJDialog.this.firePropertyChange(
//							PROPERTY_CANCEL_AND_CLOSE, null, null);
//					setVisible(false);
//					setCancelled(true);
//					dispose();
					cancel();
				}
			});
		}
		jPanel.add(cancelButton);

		return jPanel;
	}

	/**
	 * This method is only called when the dialog is closed and not cancelled.
	 * Can be overwritten to do anything when the dialog has been accepted.
	 */
	protected void okClose() {
		setVisible(false);
		dispose();
	}

	/**
	 * @return <code>true</code> if none of the translations contains illegal
	 *         characters.
	 */
	protected boolean checkValidInputs() {

		for (JComponent component : translationEditJPanelsOrJustComponents) {
			if (component instanceof TranslationEditJPanel) {
				TranslationEditJPanel tep = (TranslationEditJPanel) component;

				for (String l : tep.getTranslation().values()) {
					if (l.contains("{") || l.contains("}")) {
						JOptionPane
								.showMessageDialog(
										this,
										SwingUtil
												.R("TranslationAskJDialog.ErrorMsg.InvalidCharacterInTranslation"));
						return false;
					}
				}

			}
		}

		return true;
	}

	private void setCancelled(boolean hasBeenCanceled) {
		this.hasBeenCanceled = hasBeenCanceled;
	}

	/**
	 * After the modal dialog has been closed, this allows to find out, whether
	 * the dialog has been canceled.
	 * 
	 * @return <code>true</code> if the {@link JDialog} has been canceled.
	 */
	public boolean isCancelled() {
		return hasBeenCanceled;
	}

}
